/*
** $Id: calc.c, initiated June 20, 2008 $
** Calculus library
** See Copyright Notice in agena.h
*/


#include <stdlib.h>
#include <math.h>
#include <string.h>

#define calc_c
#define LUA_LIB

#include "agena.h"

#include "agnxlib.h"
#include "agenalib.h"
#include "agncmpt.h"
#include "agnhlps.h"

#include "cephes.h"
#include "interp.h"
#include "rlhmath.h"

#if !(defined(LUA_DOS) || defined(__OS2__) || defined(LUA_ANSI))
#define AGENA_CALCLIBNAME "calc"
LUALIB_API int (luaopen_calc) (lua_State *L);
#endif


#define checkarray(L, n) (NumArray *)luaL_checkudata(L, n, "calc")

typedef struct NumArray {
  int size;
  lua_Number *values;  /* 2.9.0 change */
} NumArray;

static int aux_newnumarray (lua_State *L, size_t n) {  /* creates numeric C array userdata object, rewritten 2.9.0 */
  NumArray *a;
  a = (NumArray *)lua_newuserdata(L, sizeof(NumArray));  /* push userdata on stack */
  a->values = calloc(n, sizeof(lua_Number));
  a->size = n;
  /* attach proper metatable with __gc method to userdata; name of mt name must be different from the one of
     the numarray package to avoid crashes during gc. */
  luaL_getmetatable(L, "calc");
  lua_setmetatable(L, -2);
  return 1;  /* leave userdata on the top of the stack */
}

static int arraygc (lua_State *L) {  /* __gc method */
  NumArray *a;
  (void)L;
  a = checkarray(L, 1);  /* 2.9.0 */
  xfree(a->values);
  return 0;
}

static const struct luaL_Reg numarray_arraylib [] = {
  {"__gc", arraygc},
  {NULL, NULL}
};

#define getarray(x, n)     (x)->values[(n)]
#define setarray(x, n, v)  (x)->values[(n)] = (v)


static int calc_fsum (lua_State *L) {
  /* computes the sum of f(a), ..., f(b); 0.10.0 - April 20, 2008;
     12 % faster than an Agena implementation;
     extended 0.30.0, December 30, 2009: the procedure now uses a modified Kahan summation algorithm
     to avoid roundoff errors; the modified Kahan algorithm has been developed by Kazufumi Ozawa
     and has been published in his paper `Analysis and Improvement of Kahan's Summation Algorithm`,
     available at http://ci.nii.ac.jp/naid/110002673355/en */
  lua_Number a, b;
  volatile lua_Number q, s, sold, u, v, w, x, t;
  int nargs;
  s = q = 0;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  nargs = lua_gettop(L);  /* 2.1.4 */
  luaL_checkstack(L, nargs, "too many arguments");
  a = agn_checknumber(L, 2);  /* start value */
  b = agn_checknumber(L, 3);  /* stop value */
  while (a <= b) {
    x = agnL_fncall(L, 1, a, 4, nargs, 1);
    v = x - q;
    sold = s;
    s = s + v;
    if (fabs(x) < fabs(q)) {
      t = x;
      x = -q;
      q = t;
    }
    u = (v - x) + q;
    if (fabs(sold) < fabs(v)) {
      t = sold;
      sold = v;
      v = t;
    }
    w = (s - sold) - v;
    q = u + w;
    a++;
  }
  lua_pushnumber(L, s);
  return 1;
}


static int calc_fprod (lua_State *L) {
  /* computes the product of f(a), ..., f(b); 0.29.0 - Nov 23, 2009 */
  lua_Number a, b, r;
  int error;
  r = 1;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  while (a <= b) {
    lua_pushvalue(L, 1);
    lua_pushnumber(L, a);
    r *= agn_ncall(L, 1, 1, &error, 1);
    a++;
  }
  lua_pushnumber(L, r);
  return 1;
}


/* DE-Quadrature, Numerical Automatic Integrator for Improper Integral
      method    : Double Exponential (DE) Transformation
      dimension : one
      table     : use
   functions
      intde  : integrator of f(x) over (a,b).

   written by Takuya Ooura in C and available at http://www.kurims.kyoto-u.ac.jp/~ooura/intde.html.
   See intde2.c file.
   For an explanation of the algorithm used, see: http://en.wikipedia.org/wiki/Tanh-sinh_quadrature

   Copyright(C) 1996 Takuya OOURA.
   You may use, copy, modify this code for any purpose and without fee.
   You may distribute this ORIGINAL package. */

void intdeini (int lenaw, double tiny, double eps, double *aw) {
  /* ---- adjustable parameter ---- */
  double efs = 0.1, hoff = 8.5;
  /* ------------------------------ */
  int noff, nk, k, j;
  double pi2, tinyln, epsln, h0, ehp, ehm, h, t, ep, em, xw, wg;

  pi2 = 2 * atan(1.0);
  tinyln = -log(tiny);
  epsln = 1 - log(efs * eps);
  h0 = hoff / epsln;
  ehp = sun_exp(h0);
  ehm = 1 / ehp;
  aw[2] = eps;
  aw[3] = sun_exp(-ehm * epsln);
  aw[4] = sqrt(efs * eps);
  noff = 5;
  aw[noff] = 0.5;
  aw[noff + 1] = h0;
  aw[noff + 2] = pi2 * h0 * 0.5;
  h = 2;
  nk = 0;
  k = noff + 3;
  do {
    t = h * 0.5;
    do {
      em = sun_exp(h0 * t);
      ep = pi2 * em;
      em = pi2 / em;
      j = k;
      do {
        xw = 1 / (1 + sun_exp(ep - em));
        wg = xw * (1 - xw) * h0;
        aw[j] = xw;
        aw[j + 1] = wg * 4;
        aw[j + 2] = wg * (ep + em);
        ep *= ehp;
        em *= ehm;
        j += 3;
      } while (ep < tinyln && j <= lenaw - 3);
      t += h;
      k += nk;
    } while (t < 1);
    h *= 0.5;
    if (nk == 0) {
      if (j > lenaw - 6) j -= 3;
      nk = j - noff;
      k += nk;
      aw[1] = nk;
    }
  } while (2 * k - noff - 3 <= lenaw);
  aw[0] = k - 3;
}


static int calc_intde (lua_State *L) {
  int noff, lenawm, nk, k, j, jtmp, jm, m, klim;
  double aw[8000], r, a, b, eps, err, epsh, ba, ir, xa, fa, fb, errt, errh, errd, h, iback, irback;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agnL_checknumber(L, 2);
  b = agnL_checknumber(L, 3);
  eps = agnL_optnonnegative(L, 4, 1.0e-15);  /* 2.13.0 */
  intdeini(8000, 1.0e-307, eps, aw);
  noff = 5;
  lenawm = (int)(aw[0] + 0.5);
  nk = (int)(aw[1] + 0.5);
  epsh = aw[4];
  ba = b - a;
  r = agnL_fncall(L, 1, (a + b) * aw[noff], 10, 9, 1);
  ir = r * aw[noff + 1];
  r *= aw[noff + 2];
  err = fabs(r);
  k = nk + noff;
  j = noff;
  do {
    j += 3;
    xa = ba * aw[j];
    fa = agnL_fncall(L, 1, a + xa, 10, 9, 1);
    fb = agnL_fncall(L, 1, b - xa, 10, 9, 1);
    ir += (fa + fb) * aw[j + 1];
    fa *= aw[j + 2];
    fb *= aw[j + 2];
    r += fa + fb;
    err += fabs(fa) + fabs(fb);
  } while (aw[j] > epsh && j < k);
  errt = err * aw[3];
  errh = err * epsh;
  errd = 1 + 2 * errh;
  jtmp = j;
  while (fabs(fa) > errt && j < k) {
    j += 3;
    fa = agnL_fncall(L, 1, a + ba * aw[j], 10, 9, 1);
    ir += fa * aw[j + 1];
    fa *= aw[j + 2];
    r += fa;
  }
  jm = j;
  j = jtmp;
  while (fabs(fb) > errt && j < k) {
    j += 3;
    fb = agnL_fncall(L, 1, b - ba * aw[j], 10, 9, 1);
    ir += fb * aw[j + 1];
    fb *= aw[j + 2];
    r += fb;
  }
  if (j < jm) jm = j;
  jm -= noff + 3;
  h = 1;
  m = 1;
  klim = k + nk;
  while (errd > errh && klim <= lenawm) {
    iback = r;
    irback = ir;
    do {
      jtmp = k + jm;
      for (j = k + 3; j <= jtmp; j += 3) {
        xa = ba * aw[j];
        fa = agnL_fncall(L, 1, a + xa, 10, 9, 1);
        fb = agnL_fncall(L, 1, b - xa, 10, 9, 1);
        ir += (fa + fb) * aw[j + 1];
        r += (fa + fb) * aw[j + 2];
      }
      k += nk;
      j = jtmp;
      do {
        j += 3;
        fa = agnL_fncall(L, 1, a + ba * aw[j], 10, 9, 1);
        ir += fa * aw[j + 1];
        fa *= aw[j + 2];
        r += fa;
      } while (fabs(fa) > errt && j < k);
      j = jtmp;
      do {
        j += 3;
        fb = agnL_fncall(L, 1, b - ba * aw[j], 10, 9, 1);
        ir += fb * aw[j + 1];
        fb *= aw[j + 2];
        r += fb;
      } while (fabs(fb) > errt && j < k);
    } while (k < klim);
    errd = h * (fabs(r - 2 * iback) + fabs(ir - 2 * irback));
    h *= 0.5;
    m *= 2;
    klim = 2 * klim - noff;
  }
  r *= h * ba;
  if (errd > errh) {
    err = -errd * (m * fabs(ba));
  } else {
    err = err * aw[2] * (m * fabs(ba));
  }
  if (err < 0)
    lua_pushfail(L);
  else
    lua_pushnumber(L, r);
  lua_pushnumber(L, err);
  return 2;
}


void intdeiini (int lenaw, double tiny, double eps, double *aw) {
  /* ---- adjustable parameter ---- */
  double efs = 0.1, hoff = 11.0;
  /* ------------------------------ */
  int noff, nk, k, j;
  double pi4, tinyln, epsln, h0, ehp, ehm, h, t, ep, em, xp, xm,
    wp, wm;
  pi4 = atan(1.0);
  tinyln = -log(tiny);
  epsln = 1 - log(efs * eps);
  h0 = hoff / epsln;
  ehp = sun_exp(h0);
  ehm = 1 / ehp;
  aw[2] = eps;
  aw[3] = sun_exp(-ehm * epsln);
  aw[4] = sqrt(efs * eps);
  noff = 5;
  aw[noff] = 1;
  aw[noff + 1] = 4 * h0;
  aw[noff + 2] = 2 * pi4 * h0;
  h = 2;
  nk = 0;
  k = noff + 6;
  do {
    t = h * 0.5;
    do {
      em = sun_exp(h0 * t);
      ep = pi4 * em;
      em = pi4 / em;
      j = k;
      do {
        xp = sun_exp(ep - em);
        xm = 1 / xp;
        wp = xp * ((ep + em) * h0);
        wm = xm * ((ep + em) * h0);
        aw[j] = xm;
        aw[j + 1] = xp;
        aw[j + 2] = xm * (4 * h0);
        aw[j + 3] = xp * (4 * h0);
        aw[j + 4] = wm;
        aw[j + 5] = wp;
        ep *= ehp;
        em *= ehm;
        j += 6;
      } while (ep < tinyln && j <= lenaw - 6);
      t += h;
      k += nk;
    } while (t < 1);
    h *= 0.5;
    if (nk == 0) {
      if (j > lenaw - 12) j -= 6;
      nk = j - noff;
      k += nk;
      aw[1] = nk;
    }
  } while (2 * k - noff - 6 <= lenaw);
  aw[0] = k - 6;
}


static int calc_intdei (lua_State *L) {
  int noff, lenawm, nk, k, j, jtmp, jm, m, klim;
  double aw[8000], err, eps, r, a, epsh, ir, fp, fm, errt, errh, errd, h, iback, irback;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agnL_checknumber(L, 2);
  eps = agnL_optnonnegative(L, 3, 1.0e-15);  /* 2.13.0 */
  intdeiini(8000, 1.0e-307, eps, aw);
  noff = 5;
  lenawm = (int)(aw[0] + 0.5);
  nk = (int)(aw[1] + 0.5);
  epsh = aw[4];
  r = agnL_fncall(L, 1, a + aw[noff], 10, 9, 1);
  ir = r * aw[noff + 1];
  r *= aw[noff + 2];
  err = fabs(r);
  k = nk + noff;
  j = noff;
  do {
    j += 6;
    fm = agnL_fncall(L, 1, a + aw[j], 10, 9, 1);
    fp = agnL_fncall(L, 1, a + aw[j + 1], 10, 9, 1);
    ir += fm * aw[j + 2] + fp * aw[j + 3];
    fm *= aw[j + 4];
    fp *= aw[j + 5];
    r += fm + fp;
    err += fabs(fm) + fabs(fp);
  } while (aw[j] > epsh && j < k);
  errt = err * aw[3];
  errh = err * epsh;
  errd = 1 + 2 * errh;
  jtmp = j;
  while (fabs(fm) > errt && j < k) {
    j += 6;
    fm = agnL_fncall(L, 1, a + aw[j], 10, 9, 1);
    ir += fm * aw[j + 2];
    fm *= aw[j + 4];
    r += fm;
  }
  jm = j;
  j = jtmp;
  while (fabs(fp) > errt && j < k) {
    j += 6;
    fp = agnL_fncall(L, 1, a + aw[j + 1], 10, 9, 1);
    ir += fp * aw[j + 3];
    fp *= aw[j + 5];
    r += fp;
  }
  if (j < jm) jm = j;
  jm -= noff + 6;
  h = 1;
  m = 1;
  klim = k + nk;
  while (errd > errh && klim <= lenawm) {
    iback = r;
    irback = ir;
    do {
      jtmp = k + jm;
      for (j = k + 6; j <= jtmp; j += 6) {
        fm = agnL_fncall(L, 1, a + aw[j], 10, 9, 1);
        fp = agnL_fncall(L, 1, a + aw[j + 1], 10, 9, 1);
        ir += fm * aw[j + 2] + fp * aw[j + 3];
        r += fm * aw[j + 4] + fp * aw[j + 5];
      }
      k += nk;
      j = jtmp;
      do {
        j += 6;
        fm = agnL_fncall(L, 1, a + aw[j], 10, 9, 1);
        ir += fm * aw[j + 2];
        fm *= aw[j + 4];
        r += fm;
      } while (fabs(fm) > errt && j < k);
      j = jtmp;
      do {
        j += 6;
        fp = agnL_fncall(L, 1, a + aw[j + 1], 10, 9, 1);
        ir += fp * aw[j + 3];
        fp *= aw[j + 5];
        r += fp;
      } while (fabs(fp) > errt && j < k);
    } while (k < klim);
    errd = h * (fabs(r - 2 * iback) + fabs(ir - 2 * irback));
    h *= 0.5;
    m *= 2;
    klim = 2 * klim - noff;
  }
  r *= h;
  if (errd > errh) {
    err = -errd * m;
  } else {
    err *= aw[2] * m;
  }
  if (err < 0)
    lua_pushfail(L);
  else
    lua_pushnumber(L, r);
  lua_pushnumber(L, err);
  return 2;
}


void intdeoini (int lenaw, double tiny, double eps, double *aw) {
  /* ---- adjustable parameter ---- */
  int lmax = 5;
  double efs = 0.1, enoff = 0.40, pqoff = 2.9, ppoff = -0.72;
  /* ------------------------------ */
  int noff0, nk0, noff, k, nk, j;
  double pi4, tinyln, epsln, frq4, per2, pp, pq, ehp, ehm, h, t,
    ep, em, tk, xw, wg, xa;

  pi4 = atan(1.0);
  tinyln = -log(tiny);
  epsln = 1 - log(efs * eps);
  frq4 = 1 / (2 * pi4);
  per2 = 4 * pi4;
  pq = pqoff / epsln;
  pp = ppoff - log(pq * pq * frq4);
  ehp = sun_exp(2 * pq);
  ehm = 1 / ehp;
  aw[3] = lmax;
  aw[4] = eps;
  aw[5] = sqrt(efs * eps);
  noff0 = 6;
  nk0 = 1 + (int)(enoff * epsln);
  aw[1] = nk0;
  noff = 2 * nk0 + noff0;
  wg = 0;
  xw = 1;
  for (k = 1; k <= nk0; k++) {
    wg += xw;
    aw[noff - 2 * k] = wg;
    aw[noff - 2 * k + 1] = xw;
    xw = xw * (nk0 - k) / k;
  }
  wg = per2 / wg;
  for (k = noff0; k <= noff - 2; k += 2) {
    aw[k] *= wg;
    aw[k + 1] *= wg;
  }
  xw = sun_exp(pp - 2 * pi4);
  aw[noff] = sqrt(xw * (per2 * 0.5));
  aw[noff + 1] = xw * pq;
  aw[noff + 2] = per2 * 0.5;
  h = 2;
  nk = 0;
  k = noff + 3;
  do {
    t = h * 0.5;
    do {
      em = sun_exp(2 * pq * t);
      ep = pi4 * em;
      em = pi4 / em;
      tk = t;
      j = k;
      do {
        xw = sun_exp(pp - ep - em);
        wg = sqrt(frq4 * xw + tk * tk);
        xa = xw / (tk + wg);
        wg = (pq * xw * (ep - em) + xa) / wg;
        aw[j] = xa;
        aw[j + 1] = xw * pq;
        aw[j + 2] = wg;
        ep *= ehp;
        em *= ehm;
        tk += 1;
        j += 3;
      } while (ep < tinyln && j <= lenaw - 3);
      t += h;
      k += nk;
    } while (t < 1);
    h *= 0.5;
    if (nk == 0) {
      if (j > lenaw - 6) j -= 3;
      nk = j - noff;
      k += nk;
      aw[2] = nk;
    }
  } while (2 * k - noff - 3 <= lenaw);
  aw[0] = k - 3;
}


static int calc_intdeo (lua_State *L) {
  int lenawm, nk0, noff0, nk, noff, lmax, m, k, j, jm, l;
  double aw[8000], r, a, omega, err, eps, per, perw, w02, ir, h, iback, irback, t, tk,
    xa, fm, fp, errh, s0, s1, s2, errd;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agnL_checknumber(L, 2);
  omega = agnL_optnumber(L, 3, 1);
  eps = agnL_optnonnegative(L, 4, 1.0e-15);  /* 2.13.0 */
  intdeoini(8000, 1.0e-307, eps, aw);
  lenawm = (int)(aw[0] + 0.5);
  nk0 = (int)(aw[1] + 0.5);
  noff0 = 6;
  nk = (int)(aw[2] + 0.5);
  noff = 2 * nk0 + noff0;
  lmax = (int)(aw[3] + 0.5);
  eps = aw[4];
  per = 1 / fabs(omega);
  w02 = 2 * aw[noff + 2];
  perw = per * w02;
  r = agnL_fncall(L, 1, a + aw[noff] * per, 10, 9, 1);
  ir = r * aw[noff + 1];
  r *= aw[noff + 2];
  err = fabs(r);
  h = 2;
  m = 1;
  k = noff;
  jm = fm = fp = errh = 0;  /* to prevent compiler warnings */
  do {
    iback = r;
    irback = ir;
    t = h * 0.5;
    do {
      if (k == noff) {
        tk = 1;
        k += nk;
        j = noff;
        do {
          j += 3;
          xa = per * aw[j];
          fm = agnL_fncall(L, 1, a + xa, 10, 9, 1);
          fp = agnL_fncall(L, 1, a + xa + perw * tk, 10, 9, 1);
          ir += (fm + fp) * aw[j + 1];
          fm *= aw[j + 2];
          fp *= w02 - aw[j + 2];
          r += fm + fp;
          err += fabs(fm) + fabs(fp);
          tk += 1;
        } while (aw[j] > eps && j < k);
        errh = err * aw[5];
        err *= eps;
        jm = j - noff;
      } else {
        tk = t;
        for (j = k + 3; j <= k + jm; j += 3) {
          xa = per * aw[j];
          fm = agnL_fncall(L, 1, a + xa, 10, 9, 1);
          fp = agnL_fncall(L, 1, a + xa + perw * tk, 10, 9, 1);
          ir += (fm + fp) * aw[j + 1];
          fm *= aw[j + 2];
          fp *= w02 - aw[j + 2];
          r += fm + fp;
          tk += 1;
        }
        j = k + jm;
        k += nk;
      }
      while (fabs(fm) > err && j < k) {
        j += 3;
        fm = agnL_fncall(L, 1, a + per * aw[j], 10, 9, 1);
        ir += fm * aw[j + 1];
        fm *= aw[j + 2];
        r += fm;
      }
      fm = agnL_fncall(L, 1, a + perw * tk, 10, 9, 1);
      s2 = w02 * fm;
      r += s2;
      if (fabs(fp) > err || fabs(s2) > err) {
        l = 0;
        for (;;) {
          l++;
          s0 = 0;
          s1 = 0;
          s2 = fm * aw[noff0 + 1];
          for (j = noff0 + 2; j <= noff - 2; j += 2) {
            tk += 1;
            fm = agnL_fncall(L, 1, a + perw * tk, 10, 9, 1);
            s0 += fm;
            s1 += fm * aw[j];
            s2 += fm * aw[j + 1];
          }
          if (s2 <= err || l >= lmax) break;
          r += w02 * s0;
        }
        r += s1;
        if (s2 > err) err = s2;
      }
      t += h;
    } while (t < 1);
    if (m == 1) {
      errd = 1 + 2 * errh;
    } else {
      errd = h * (fabs(r - 2 * iback) + fabs(ir - 2 * irback));
    }
    h *= 0.5;
    m *= 2;
  } while (errd > errh && 2 * k - noff <= lenawm);
  r *= h * per;
  if (errd > errh) {
    err = -errd * per;
  } else {
    err *= per * m * 0.5;
  }
  if (err < 0)
    lua_pushfail(L);
  else
    lua_pushnumber(L, r);
  lua_pushnumber(L, err);
  return 2;
}


static int calc_Si (lua_State *L) {  /* 0.32.2 */
  lua_Number x, si, ci;
  x = agn_checknumber(L, 1);
  if (x == HUGE_VAL)
    lua_pushnumber(L, 0.5*PI);
  else if (-x == HUGE_VAL)
    lua_pushnumber(L, -0.5*PI);
  else {
    sici(x, &si, &ci);
    lua_pushnumber(L, si);
  }
  return 1;
}


static int calc_Ci (lua_State *L) {  /* 0.32.2 */
  lua_Number x, si, ci;
  x = agn_checknumber(L, 1);
  if (x == HUGE_VAL)
    lua_pushnumber(L, 0);
  else if (-x == HUGE_VAL)
    lua_pushfail(L);
  else if (x < 0)
    lua_pushundefined(L);
  else {
    sici(agn_checknumber(L, 1), &si, &ci);
    lua_pushnumber(L, ci);
  }
  return 1;
}


static int calc_Ssi (lua_State *L) {  /* 0.32.2 */
  lua_Number x, si, ci;
  x = agn_checknumber(L, 1);
  if (x == HUGE_VAL)
    lua_pushnumber(L, 0);
  else if (-x == HUGE_VAL)
    lua_pushnumber(L, -PI);
  else {
    sici(x, &si, &ci);
    lua_pushnumber(L, si - 0.5*PI);
  }
  return 1;
}


static int calc_Shi (lua_State *L) {  /* 0.32.2 */
  lua_Number x, shi, chi;
  x = agn_checknumber(L, 1);
  if (-x == HUGE_VAL)
    lua_pushnumber(L, x);
  else {
    shichi(x, &shi, &chi);
    lua_pushnumber(L, shi);
  }
  return 1;
}


static int calc_Chi (lua_State *L) {  /* 0.32.2 */
  lua_Number x, shi, chi;
  x = agn_checknumber(L, 1);
  if (-x == HUGE_VAL)
    lua_pushfail(L);
  else if (x < 0)
    lua_pushundefined(L);
  else {
    shichi(agn_checknumber(L, 1), &shi, &chi);
    lua_pushnumber(L, chi);
  }
  return 1;
}


static int calc_dawson (lua_State *L) {  /* 0.32.2 */
  lua_Number x;
  x = agn_checknumber(L, 1);
  lua_pushnumber(L, dawsn(x));
  return 1;
}


static int calc_Ei (lua_State *L) {  /* 0.32.2, extended 2.1.5 for negative arguments */
  lua_Number x;
  x = agn_checknumber(L, 1);
  if (x == HUGE_VAL)
    lua_pushnumber(L, HUGE_VAL);
  else if (-x == HUGE_VAL)
    lua_pushnumber(L, 0);
  else
    lua_pushnumber(L, (x > 0) ? ei(x) : Ei(x));
  return 1;
}


static int calc_Psi (lua_State *L) {  /* 0.32.2 */
  lua_Number x;
  x = agn_checknumber(L, 1);
  lua_pushnumber(L, psi(x));
  return 1;
}


static int calc_dilog (lua_State *L) {  /* 0.32.2 */
  lua_Number x;
  x = agn_checknumber(L, 1);
  if (x == HUGE_VAL)
    lua_pushnumber(L, -HUGE_VAL);
  else
    lua_pushnumber(L, spence(x));
  return 1;
}


static int calc_fresnelc (lua_State *L) {  /* 0.32.3 */
  lua_Number x, s, c;
  x = agn_checknumber(L, 1);
  fresnl(x, &s, &c);
  lua_pushnumber(L, c);
  return 1;
}


static int calc_fresnels (lua_State *L) {  /* 0.32.3 */
  lua_Number x, s, c;
  x = agn_checknumber(L, 1);
  fresnl(x, &s, &c);
  lua_pushnumber(L, s);
  return 1;
}


/* calc.sections (f, a, b, step)

   returns all intervals where a function has a change in sign. f must be a function, a the left border of the
   main interval, b its right border, and step the step size. The return is a sequence of pairs denoting the
   found subintervals.

   Maple versions: 1997, 1998; tuned Agena version January 21, 2008;
   C Version of Agena version 29.05.2010, 0.32.3 */

static int calc_sections (lua_State *L) {
  lua_Number xleft, xright, step, fl, fr, total;
  volatile lua_Number c, y, t, i;
  size_t seqcounter, counter;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  xleft = agn_checknumber(L, 2);
  xright = agn_checknumber(L, 3);
  step = agn_checknumber(L, 4);
  agn_createseq(L, 4);  /* result */
  fl = agnL_fncall(L, 1, xleft, 10, 9, 1);
  i = xleft;
  c = 0.0;
  seqcounter = counter = 0;
  total = trunc(fabs(xright - xleft)/step + 1);  /* total number of iterations, do not use floor, (int), etc ! */
  while (i <= xright || total > counter) {
    fr = agnL_fncall(L, 1, i + step, 10, 9, 1);
    if (fl*fr <= 0) {
      lua_pushnumber(L, i);
      lua_pushnumber(L, i + step);
      agn_createpair(L, -2, -1);
      lua_remove(L, -2);
      lua_remove(L, -2);
      lua_seqseti(L, -2, ++seqcounter);
    }
    counter++;
    /* since right border in the curent iteration is equal to the left border in the next
      iteration, avoid calling f twice. */
    fl = fr;
    /* apply Kahan summation algorithm to avoid roundoff errors, 0.30.0 */
    y = step - c;
    t = i + y;
    c = (t - i) - y;
    i = t;
  }
  return 1;  /* return result */
}


/*
  Modified Regula Falsi method; Maple versions: 1996, 1997; Agena port January 20, 2008/July 02, 2008
  This algorithm, taken from an unknown FORTRAN book in the 1980s, is the most accurate and the
  fastest one, applicable to a large range of function types. All other methods tested
  including checking for a change of sign are slower and more inaccurate.
*/

static int calc_regulafalsi (lua_State *L) {
  lua_Number x1, x2, u, v, z, a, b, eps;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  eps = agn_checknonnegative(L, 4);  /* 2.13.0 */
  x1 = a;
  x2 = b;
  z = AGN_NAN;
  u = agnL_fncall(L, 1, x1, 10, 9, 1);
  v = agnL_fncall(L, 1, x2, 10, 9, 1);
  if (fabs(u) < eps) {
    lua_pushnumber(L, x1);
    return 1;
  }
  if (fabs(v) < eps) {
    lua_pushnumber(L, x2);
    return 1;
  }
  while (fabs(u - v) > eps) {
    z = x2 - v*(x2 - x1)/(v - u);
    u = v;
    x1 = x2;
    x2 = z;
    v = agnL_fncall(L, 1, x2, 10, 9, 1);
  }
  if (z >= a && z <= b)
    lua_pushnumber(L, z);
  else
    lua_pushnil(L);
  return 1;
}


#define aux_geteps(L, x, nargs, procname) { \
  eps = 0; \
  delta = 0; \
  sided = 0; \
  deriv = 1; \
  while (lua_ispair(L, nargs)) { \
    agn_pairgeti(L, nargs, 1); \
    if (!agn_isstring(L, -1)) { \
      agn_poptop(L); \
      luaL_error(L, "Error in " LUA_QS ": left-hand side of option must be a string.", (procname)); \
    } \
    option = lua_tostring(L, -1); \
    if (tools_streq(option, "eps")) { \
      agn_pairgeti(L, nargs, 2); \
      if (agn_isnumber(L, -1)) { \
        eps = lua_tonumber(L, -1); \
        agn_poptop(L); \
        if (eps <= 0) { \
          agn_poptoptwo(L); \
          luaL_error(L, "Error in " LUA_QS ": epsilon must be positive.", (procname)); \
        }; \
      } else  { \
        agn_poptoptwo(L); \
        luaL_error(L, "Error in " LUA_QS ": right-hand side of epsilon option must be a number.", (procname)); \
      } \
    } else if (tools_streq(option, "delta")) { \
      agn_pairgeti(L, nargs, 2); \
      if (agn_isnumber(L, -1)) { \
        delta = lua_tonumber(L, -1); \
        agn_poptop(L); \
        if (delta <= 0) { \
          agn_poptoptwo(L); \
          luaL_error(L, "Error in " LUA_QS ": delta must be positive.", (procname)); \
        }; \
      } else  { \
        agn_poptoptwo(L); \
        luaL_error(L, "Error in " LUA_QS ": right-hand side of delta option must be a number.", (procname)); \
      } \
    } else if (tools_streq(option, "side")) { \
      agn_pairgeti(L, nargs, 2); \
      if (!agn_isstring(L, -1)) { \
        agn_poptoptwo(L); \
        luaL_error(L, "Error in " LUA_QS ": right-hand side of side option must be a string.", (procname)); \
      } \
      option = lua_tostring(L, -1); \
      if (tools_streq(option, "left")) sided = 1; \
      else if (tools_streq(option, "right")) sided = 2; \
      else if (tools_streq(option, "both")) sided = 3; \
      else if (tools_streq(option, "all")) sided = 4; \
      else { \
        agn_poptop(L); \
        luaL_error(L, "Error in " LUA_QS ": unknown right-hand side value for side option.", (procname)); \
      } \
      agn_poptop(L); \
    } else if (tools_streq(option, "deriv")) { \
      agn_pairgeti(L, nargs, 2); \
      if (agn_isnumber(L, -1)) { \
        deriv = lua_tointeger(L, -1); \
        agn_poptop(L); \
        if (deriv < 0 || deriv > 3) { \
          agn_poptoptwo(L); \
          luaL_error(L, "Error in " LUA_QS ": deriv must be 0, 1, 2 or 3.", (procname)); \
        }; \
      } else  { \
        agn_poptoptwo(L); \
        luaL_error(L, "Error in " LUA_QS ": right-hand side of deriv option must be a number.", (procname)); \
      } \
    } else { \
      agn_poptop(L); \
      luaL_error(L, "Error in " LUA_QS ": unknown option.", (procname)); \
    } \
    agn_poptop(L); \
    nargs--; \
  } \
  if (eps == 0) eps = tools_matheps((x)); \
  if (delta == 0) delta = tools_matheps((x)); \
}


/* Compute the symmetric derivative of a function f at a point x.
   The algorithm is based on Conte and de Boor's `Coefficients of Newton form of polynomial of degree 3`;
   written June 09, 2008, 0.11.2. The function must be at index 1. */
lua_Number aux_diff (lua_State *L, lua_Number x, lua_Number eps, int deriv, int nargs, lua_Number *abserr) {  /* 2.10.2 */
  lua_Number h, q, fph, fmh, a[4], b[4];
  int i, j;
  if (deriv == 0) {  /* 2.14.1 */
    *abserr = 0;
    return agnL_fncall(L, 1, x, 3, nargs, 1);
  }
  q = 0;
  h = eps;
  for (i=0; i < 4; i++) {  /* determine appropriate epsilon step size */
    a[i] = x + (i - 2)*h;  /* patched 2.10.1 */
    b[i] = agnL_fncall(L, 1, a[i], 3, nargs, 1);  /* 2.10.1 */
  }
  for (j=0; j < 4; j++) {  /* compute divided difference table */
    for (i=0; i < 3 - j; i++) {
      b[i] = (b[i + 1] - b[i])/(a[i + j + 1] - a[i]);
    }
  }
  for (i=0; i < 4; i++) q += b[i];
  q = fabs(q);
  if (deriv < 2 && q < 100*eps) q = 100*eps;  /* for second and higher derivative, results would be imprecise */
  h = cbrt(eps/(2*q));
  if (deriv < 2 && h > 100*eps) h = 100*eps;  /* for second and higher derivative, results would be imprecise */
  *abserr = fabs(h*h*q);  /* absolute error, previously: fabs(h*h*q*100) */
  fph = agnL_fncall(L, 1, x + h, 3, nargs, 1);  /* 2.10.1 */
  fmh = agnL_fncall(L, 1, x - h, 3, nargs, 1);  /* 2.10.1 */
  if (deriv == 1) {
    return (fph - fmh)/(2*h);  /* symmetric difference quotient */
  } else if (deriv == 2) {  /* second derivative, new 2.10.2 */
    lua_Number fx = agnL_fncall(L, 1, x, 3, nargs, 1);
    return (fph - 2*fx + fmh)/(h*h);  /* new 2.10.2 */
  } else {  /* third derivative, new 2.10.2 */
    lua_Number f2ph, f2mh, f3ph, f3mh;
    f2ph = agnL_fncall(L, 1, x + 2*h, 3, nargs, 1);
    f2mh = agnL_fncall(L, 1, x - 2*h, 3, nargs, 1);
    f3ph = agnL_fncall(L, 1, x + 3*h, 3, nargs, 1);
    f3mh = agnL_fncall(L, 1, x - 3*h, 3, nargs, 1);
    return (f3mh - 8*f2mh + 13*fmh - 13*fph + 8*f2ph - f3ph)/(8*h*h*h);
  }
}

static int calc_diff (lua_State *L) {
  lua_Number x, eps, delta, abserr;
  int sided, deriv, nargs;
  const char *option;
  (void)sided;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  option = NULL;
  x = agn_checknumber(L, 2);
  nargs = lua_gettop(L);
  aux_geteps(L, x, nargs, "calc.diff");  /* changed 2.9.8 */
  lua_pushnumber(L, aux_diff(L, x, eps, deriv, nargs, &abserr));  /* externalised 2.10.2 */
  lua_pushnumber(L, abserr);  /* absolute error */
  return 2;
}


/*
Differentiation using Richardson`s extrapolation, taken from http://netlib.org/textbook/mathews/chap6.f

NUMERICAL METHODS: FORTRAN Programs, (c) John H. Mathews 1994, used by kind permission.

Author of the original FORTRAN routine:
Prof. John  H.  Mathews
Department of Mathematics
California State University Fullerton
Fullerton, CA  92634

The FORTRAN routine accompanies the book `NUMERICAL METHODS for Mathematics, Science and Engineering`,
2nd Ed, 1992, Prentice Hall, Englewood Cliffs, New Jersey, 07632, U.S.A.

Algorithm 6.2 (Differentiation Using Extrapolation).
Section 6.1, Approximating the Derivative, Page 327
Agena port January 01, 2010, 0.30.1 */

#define MAX_N 2*15  /* original value has been 15 */
static int calc_xpdiff (lua_State *L) {  /* 0.32.3 */
  lua_Number fx, fph, fmh, f2ph, f2mh, f3ph, f3mh, x, d[MAX_N + 1][MAX_N + 1],
    eps, delta, deriv, err, h, relerr, small, abserr;
  int j, k, n, sided, nargs, stencil7;
  const char *option = NULL;
  (void)sided;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  x = agnL_checknumber(L, 2);
  nargs = lua_gettop(L);
  aux_geteps(L, x, nargs, "calc.xpdiff");  /* changed 2.10.1 */
  fx = agnL_fncall(L, 1, x, 3, nargs, 1);
  if (deriv == 0) {  /* 2.14.1 */
    lua_pushnumber(L, fx);
    return 1;
  } else if (!isfinite(fx)) {  /* infinity or undefined ? */
    lua_pushundefined(L);
    return 1;
  }
  stencil7 = 1;  /* default method for third derivative */
  small = 1e-7;
  h = 1.0;
  n = 0;
  err = 1.0;
  relerr = 1.0;
  for (j=0; j < 10; j++) {  /* 2.10.2, avoid `undefined` */
    fph = agnL_fncall(L, 1, x + h, 3, nargs, 1);  /* 2.10.1; h and not eps */
    fmh = agnL_fncall(L, 1, x - h, 3, nargs, 1);  /* 2.10.1; h and not eps */
    if (isfinite(fph) && isfinite(fmh)) break;
    h /= 2;
  }
  if ( (deriv != 3) && (!isfinite(fph) || !isfinite(fmh)) ) {  /* changes 2.16.1 due to GCC warning */
    /* fallback, as call to calc.diff often computes a reliable result */
    lua_pushnumber(L, aux_diff(L, x, eps, deriv, nargs, &abserr));
    lua_pushnumber(L, abserr);
    return 2;
  }
  if (deriv == 1)
    d[0][0] = 0.5*(fph - fmh)/h;
  else if (deriv == 2)
    d[0][0] = (fph - 2*fx + fmh)/(h*h);  /* new 2.10.2, formula taken from Wikipedia, 'Symmetric derivative' */
  else {
    f2ph = agnL_fncall(L, 1, x + 2*h, 3, nargs, 1);
    f2mh = agnL_fncall(L, 1, x - 2*h, 3, nargs, 1);
    f3ph = agnL_fncall(L, 1, x + 3*h, 3, nargs, 1);
    f3mh = agnL_fncall(L, 1, x - 3*h, 3, nargs, 1);
    if (isfinite(f2ph) && isfinite(f2mh) && (!isfinite(f3ph) || !isfinite(f3mh)))
      stencil7 = 0;
    else if (!isfinite(f2ph) || !isfinite(f2mh) || !isfinite(f3ph) || !isfinite(f3mh)) {
      /* fallback, as call to calc.diff often computes a reliable result */
      lua_pushnumber(L, aux_diff(L, x, eps, deriv, nargs, &abserr));
      lua_pushnumber(L, abserr);
      return 2;
    }
    /* new 2.10.2, 7-stencil, www.holoborodko.com/pavel/numerical-methods/numerical-derivative/central-differences/
       by Pavel Holoborodko */
    if (stencil7)
      d[0][0] = (f3mh - 8*f2mh + 13*fmh - 13*fph + 8*f2ph - f3ph)/(8*h*h*h);
    else  /* use 5-stencil method, also published on Pavel Holoborodko's site */
      d[0][0] = (-f2mh + 2*fmh - 2*fph + 2*f2ph)/(2*h*h*h);
  }
  for (j=1; (j <= MAX_N) && (relerr > eps) && (err > delta); j++) {
    h /= 2;
    fph = agnL_fncall(L, 1, x + h, 3, nargs, 1);  /* 2.10.1 */
    fmh = agnL_fncall(L, 1, x - h, 3, nargs, 1);  /* 2.10.1 */
    if (deriv == 1)
      d[j][0] = 0.5*(fph - fmh)/h;
    else if (deriv == 2)
      d[j][0] = (fph - 2*fx + fmh)/(h*h);  /* new 2.10.2 */
    else {  /* third derivative, new 2.10.2 */
      f2ph = agnL_fncall(L, 1, x + 2*h, 3, nargs, 1);
      f2mh = agnL_fncall(L, 1, x - 2*h, 3, nargs, 1);
      if (stencil7) {
        f3ph = agnL_fncall(L, 1, x + 3*h, 3, nargs, 1);
        f3mh = agnL_fncall(L, 1, x - 3*h, 3, nargs, 1);
        d[j][0] = (f3mh - 8*f2mh + 13*fmh - 13*fph + 8*f2ph - f3ph)/(8*h*h*h);
      } else {
        d[j][0] = (-f2mh + 2*fmh - 2*fph + 2*f2ph)/(2*h*h*h);
      }
    }
    for (k=1; k <= j; k++) {
      d[j][k] = d[j][k - 1] + (d[j][k - 1] - d[j - 1][k - 1])/(sun_pow(4, k) - 1);
    }
    err = fabs(d[j][j] - d[j - 1][j - 1]);
    relerr = 2*err/(fabs(d[j][j]) + fabs(d[j - 1][j - 1]) + small);
    n = j;
  }
  lua_pushnumber(L, d[n][n]);
  lua_pushnumber(L, err);
  return 2;
}


/* Computes a mathematical epsilon value that takes into account the magnitude of the value of function
   f at point x. By default, r = 3 points around (and including) the centre x are used to calculate
   the result. You may choose any other positive integer for r with r < 100.

   The optional fourth argument `adjust` - a Boolean - controls whether the function shall adjust the result if the
   internal epsilon value computed has become too large to be probably of any use. By default no adjustment takes place.

   The second return, a Boolean, indicates whether the function adjusted the result (only if `adjust` has been set
   to true).

   The function is useful to analyse functions whose values are very close to zero and where floating point arithmetic
   may lead to fatal round-off errors for it returns an epsilon value that is larger than the one returned by
   `math.eps`. */

static int calc_eps (lua_State *L) {  /* 2.9.8 */
  lua_Number eps, h, q, x, a[100], b[100];
  int i, j, r, n, adjust, adjustment, error;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  x = agn_checknumber(L, 2);
  r = agnL_optposint(L, 3, 3);
  if (r > 99)
    luaL_error(L, "Error in " LUA_QS ": third argument %d too big.", "calc.eps", r);
  adjust = agnL_optboolean(L, 4, 0);
  adjustment = 0;
  eps = tools_matheps(x);  /* get mathematical epsilon */
  h = eps;
  q = 0;
  n = r + 1;
  for (i=0; i < n; i++) {
    a[i] = x + (i - (int)(r/2))*h;  /* compute n points around the centre */
    lua_pushvalue(L, 1);  /* can't use aux_getfunc* here due to special optional arguments r and adjust */
    lua_pushnumber(L, a[i]);
    b[i] = agn_ncall(L, 1, 1, &error, 1);
  }
  for (j=0; j < n; j++) {
    for (i=0; i < r - j; i++) {
      b[i] = (b[i + 1] - b[i])/(a[i + j + 1] - a[i]);
    }
  }
  for (i=0; i < n; i++) q += b[i];
  q = fabs(q);
  if ((adjustment = adjust && q < 100*eps)) q = 100*eps;
  h = sun_pow(eps/((r - 1)*q), 1.0/r);
  if ((adjustment = adjust && h > 100*eps)) h = 100*eps;
  lua_pushnumber(L, h);  /* `ordinate` epsilon */
  lua_pushboolean(L, adjustment);
  return 2;
}


/* An estimate to the min location
   f:	    function under investigation
   a:		left border of the range the min is seeked
   b:  	right border of the range the min is seeked
   tol:	(optional) Acceptable tolerance */

static int calc_fminbr (lua_State *L) {
  lua_Number x, v, w, fv, fx, fw, r, a, b, tol, Eps;  /* Abscissae, descr. see above, f(x), f(v), f(w), r */
  int error;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  tol = agnL_optnumber(L, 4, AGN_EPSILON);
  r = (3 - sqrt(5))/2;	  /* Gold section ratio */
  if (!(tol > 0 && b > a))
    luaL_error(L, "Er ror in " LUA_QS ": tolerance <= 0 or a <= b.", "calc.fminbr");
  v = a + r*(b - a);
  lua_pushvalue(L, 1);
  lua_pushnumber(L, v);
  fv = agn_ncall(L, 1, 1, &error, 1);  /* First step - always gold section */
  x = v; w = v;
  fx = fv; fw = fv;
  Eps = agn_geteps(L);  /* 2.1.3 optimisation, a and b might be changed in this loop */
  while (1) {  /* Main iteration loop */
    lua_Number range, middle_range, tol_act, new_step, t, fa, fb, ft;
    range = b - a;           /* Range over which the minimum is seeked for */
    middle_range = (a+b)/2;
    tol_act = Eps*fabs(x) + tol/3;  /* Actual tolerance; new_step: Step at this iteration */
    if (fabs(x - middle_range) + range/2 <= 2*tol_act) {
      lua_pushvalue(L, 1);
      lua_pushnumber(L, a);
      fa = agn_ncall(L, 1, 1, &error, 1);
      lua_pushvalue(L, 1);
      lua_pushnumber(L, b);
      fb = agn_ncall(L, 1, 1, &error, 1);
      lua_pushvalue(L, 1);
      lua_pushnumber(L, x);
      fx = agn_ncall(L, 1, 1, &error, 1);
      if (fa < fx) {  /* the function is not that good if the minimum is exactly at the left or right border */
        lua_pushnumber(L, a);   /* 0.29.4 patch, 28.12.2009 */
      } else if (fb < fx) {
        lua_pushnumber(L, b);
      } else {
        lua_pushnumber(L, x);  /* Acceptable approx. is found */
      }
      return 1;
    }
    /* Obtain the gold section step */
    new_step = r * (x < middle_range ? b - x : a - x);
    /*Decide if the interpolation can be tried */
    if (fabs(x-w) >= tol_act) {	/* If x and w are distinct interpolatiom may be tried */
      lua_Number p, q, t;       /* Interpolation step is calculated as p/q; division operation is delayed */
                                /*  until last moment */
      t = (x - w) * (fx - fv);
      q = (x - v) * (fx - fw);
      p = (x - v)*q - (x - w)*t;
      q = 2*(q - t);
      if (q > 0)        /*  q was calculated with the op-  */
        p = -p;         /*  posite sign; make q positive   */
      else              /*  and assign possible minus to   */
        q = -q;         /*  p                              */
      if (fabs(p) < fabs(new_step*q) &&  /*  If x+p/q falls in [a,b]  */
        p > q*(a - x + 2*tol_act) &&         /*  not too close to a and   */
        p < q*(b - x - 2*tol_act)) {         /*  b, and isn't too large   */
        new_step = p/q;                  /*  it is accepted  */
      }
         /* If p/q is too large then the gold section procedure can reduce [a,b] range to more extent  */
    }
    if (fabs(new_step) < tol_act) {   /* Adjust the step to be not less */
      if (new_step > 0)               /* than tolerance */
        new_step = tol_act;
      else
        new_step = -tol_act;
    }
    /* Obtain the next approximation to min and reduce the enveloping range */
    t = x + new_step;	/* Tentative point for the min */
    lua_pushvalue(L, 1);
    lua_pushnumber(L, t);
    ft = agn_ncall(L, 1, 1, &error, 1);
    if (ft <= fx) {    /* t is a better approximation */
      if (t < x)       /* Reduce the range so that */
        b = x;         /* t would fall within it */
      else
        a = x;
      v = w; w = x; x = t;  /* Assign the best approx to x */
      fv = fw; fw = fx; fx = ft;
    } else { /* x remains the better approx */
      if (t < x)  /* Reduce the range enclosing x */
        a = t;
      else
        b = t;
      if (ft <= fw || w == x) {
        v = w; w = t;
        fv = fw; fw = ft;
      } else if (ft <= fv || v == x || v == w) {
        v = t;
        fv = ft;
      }
    }
  }
}


/* takes a structure of pairs of numbers and stores them to the arrays x and y */

static void agn_toarray (lua_State *L, int idx, const char *procname, size_t n, lua_Number *x, lua_Number *y) {
  size_t i;
  int type;
  type = lua_type(L, idx);
  luaL_argcheck(L, type == LUA_TTABLE || type == LUA_TSEQ, 1, "table or sequence of pairs expected");
  if (type == LUA_TTABLE) {
    i = 0;
    lua_pushnil(L);
    while (lua_next(L, idx)) {
      agn_pairgetinumbers(L, procname, -1, &(x[i]), &(y[i])); i++;
    }
  }
  else {
    for (i=0; i < n; i++) {
      lua_seqgeti(L, idx, i + 1);  /* push object */
      agn_pairgetinumbers(L, procname, -1, &(x[i]), &(y[i]));
    }
  }
}


static void agn_seqtoarray (lua_State *L, lua_Number *x, int idx, size_t n) {  /* converts sequence to an array */
  size_t i;
  for (i=0; i < n; i++) {
    x[i] = lua_seqgetinumber(L, lua_upvalueindex(idx), i + 1);
  }
}


int nevillef (lua_State *L) {  /* 2.2.0 RC 3 */
  int n;
  lua_Number t, *x, *y;
  t = agn_checknumber(L, 1);
  n = agn_tonumber(L, lua_upvalueindex(3));
  agn_createarray(x, n, "calc.neville");
  agn_createarray(y, n, "calc.neville");
  agn_seqtoarray(L, x, 1, n);
  agn_seqtoarray(L, y, 2, n);
  lua_pushnumber(L, neville(n, x, y, t));
  xfreeall(x, y);  /* 2.9.8 */
  return 1;
}

static int calc_neville (lua_State *L) {
  size_t n;
  lua_Number t, *x, *y;
  n = agn_nops(L, 1);  /* returns 0 if object at idx is not a structure */
  agn_createarray(x, n, "calc.neville");
  agn_createarray(y, n, "calc.neville");
  agn_toarray(L, 1, "calc.neville", n, x, y);
  if (lua_gettop(L) > 1) {
    t = agn_checknumber(L, 2);
    lua_pushnumber(L, neville(n, x, y, t));
  } else {  /* 2.2.0 RC 3 */
    agn_arraytoseq(L, x, n);  /* push x as a sequence */
    agn_arraytoseq(L, y, n);  /* push y as a sequence */
    lua_pushnumber(L, n);     /* push number of elements */
    lua_pushcclosure(L, &nevillef, 3);
  }
  xfreeall(x, y);  /* 1.12.9 fix, 2.9.8 */
  return 1;
}


/* calc.polyfit(obj, n);

   Returns a sequence of coefficients of an nth-degree polynomial of a sample, in order of descending degree fitting the
   input sequence or sequence obj of pairs x[k]:y[k], with x[k] and y[k] being numbers, and using polynomial regression.
   The degree n must be a positive integer.

   The return may be passed to `calc.polygen` to generate a polynomial function (use `unpack` when passing the coefficient
   vector), e.g. calc.polygen( unpack( calc.polyfit(seq( 1:0, 1:3, 2:1 ), 2 ) ) ).

   For the original C source file see: http://programbank4u.blogspot.de/2013/04/polynomial-regression-c-program.html,
   posted 11th April 2013 by Harika. */

static int calc_polyfit (lua_State *L) {  /* 2.1.3 */
  size_t n, c;
  int m, i, j, k;
  lua_Number *x, *y, *a, *b, sum, temp, q;
  n = agn_nops(L, 1);  /* returns 0 if object at idx is not a structure */
  m = agn_checkinteger(L, 2);
  if (m >= n)
    luaL_error(L, "Error in " LUA_QS ": too few items in structure for given degree %d.", "calc.polyfit", m);
  else if (m < 1)
    luaL_error(L, "Error in " LUA_QS ": degree must be positive.", "calc.polyfit");
  agn_createarray(x, n, "calc.polyfit");
  agn_createarray(y, n, "calc.polyfit");
  agn_toarray(L, 1, "calc.polyfit", n, x, y);
  agn_createarray(a, n * n, "calc.polyfit");
  agn_createarray(b, n, "calc.polyfit");
  q = 0;
  for (i=0; i <= m; i++) {
    for (j=0; j <= m; j++) {
      sum = 0;
      for (k=0; k < n; k++)
        /* sum = sum + sun_pow(x[k], i + j); */
        sum = tools_koadd(sum, sun_pow(x[k], i + j), &q);  /* 2.13.0 */
      a[i + n*j] = sum;
    }
  }
  q = 0;
  for (i=0; i <= m; i++) {
    sum = 0;
    for (k=0; k < n; k++)
      /* sum = sum + y[k] * sun_pow(x[k], i); */
      sum = tools_koadd(sum, y[k] * sun_pow(x[k], i), &q);  /* 2.13.0 */
    b[i] = sum;
  }
  for (k=0; k <= m; k++) {
    for (i=0; i <= m; i++) {
      if (i == k) continue;
      temp = a[i + n*k] / a[k + n*k];
      q = 0;
      for (j=k; j <= m; j++)
        /* a[i + n*j] = a[i + n*j] - temp*a[k + n*j]; */
        a[i + n*j] = tools_koadd(a[i + n*j], -temp*a[k + n*j], &q);  /* 2.13.0 */
      /* b[i] = b[i] - temp*b[k]; */
      q = 0;
      b[i] = tools_koadd(b[i], -temp*b[k], &q);  /* 2.13.0 */
    }
  }
  agn_createseq(L, m + 1);
  c = 1;
  for (i=m; i >= 0; i--)
    agn_seqsetinumber(L, -1, c++, b[i] / a[i + n*i]);
  xfreeall(x, y, a, b);  /* 2.9.8 */
  return 1;
}


int newton (lua_State *L) {
  int i, n;
  lua_Number t, r, *x, *nf;
  t = agn_checknumber(L, 1);
  n = agn_tonumber(L, lua_upvalueindex(3));
  agn_createarray(x, n, "calc.interp");
  agn_createarray(nf, n, "calc.interp");
  agn_seqtoarray(L, x, 1, n);
  agn_seqtoarray(L, nf, 2, n);
  r = nf[n - 1];
  for (i=n - 2; i >= 0; i--)
    r = r * (t - x[i]) + nf[i];
  xfreeall(x, nf);  /* 2.9.8 */
  lua_pushnumber(L, r);
  return 1;
}

static int calc_interp (lua_State *L) {
  size_t n, nargs;
  lua_Number *x, *y, *coeff;
  n = agn_nops(L, 1);  /* returns 0 if object at idx is not a structure */
  nargs = lua_gettop(L);
  agn_createarray(x, n, "calc.interp");
  agn_createarray(y, n, "calc.interp");
  agn_toarray(L, 1, "calc.interp", n, x, y);
  if (nargs == 3 && lua_isseq(L, 3)) {
    size_t i;
    coeff = NULL;
    if (agn_seqsize(L, 3) != n) {
      xfreeall(x, y);  /* 2.8.2 patch, 2.9.8 */
      luaL_error(L, "Error in " LUA_QS ": number of coefficients must be equal to number of points.", "calc.interp");
    }
    agn_createarray(coeff, n, "calc.interp");
    for (i=0; i < n; i++)
      coeff[i] = lua_seqgetinumber(L, 3, i + 1);
  } else
    coeff = divdiff(n, x, y);  /* compute the coefficients of the Newton form */
  if (coeff == NULL) {
    xfreeall(x, y);  /* 2.2.0 RC 3 patch, 2.9.8 */
    luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", "calc.interp");
  }
  if (nargs > 1)
    lua_pushnumber(L, nf_eval(n, x, coeff, agn_checknumber(L, 2)));
  else {
    agn_arraytoseq(L, x, n);  /* push x as a sequence */
    agn_arraytoseq(L, coeff, n);  /* push coefficients as a sequence */
    lua_pushnumber(L, n);     /* push number of elements */
    lua_pushcclosure(L, &newton, 3);
  }
  xfreeall(x, y, coeff);  /* 1.12.9 fix, 2.9.8 */
  return 1;
}


static int calc_newtoncoeffs (lua_State *L) {
  size_t i, n;
  lua_Number *x, *y, *coeff;
  n = agn_nops(L, 1);  /* returns 0 if object at idx is not a structure */
  agn_createarray(x, n, "calc.newtoncoeffs");
  agn_createarray(y, n, "calc.newtoncoeffs");
  agn_toarray(L, 1, "calc.newtoncoeffs", n, x, y);
  coeff = divdiff(n, x, y);
  if (coeff == NULL) {
    xfreeall(x, y);  /* 2.8.2 patch, 2.9.8 */
    luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", "calc.newtoncoeffs");
  }
  agn_createseq(L, n);
  for (i=0; i < n; i++) {
    agn_seqsetinumber(L, -1, i + 1, coeff[i]);
  }
  xfreeall(x, y, coeff);  /* 1.12.9 fix, 2.9.8 */
  return 1;
}


static void agn_nakcoeffs (lua_State *L, int mainidx, int idx, lua_Number *x, size_t n) {  /* copies numbers in a sequence to an array */
  size_t i;
  lua_seqgeti(L, mainidx, idx);
  if (lua_type(L, -1) != LUA_TSEQ)
    luaL_error(L, "Error in " LUA_QS ": expected a sequence, got %s.", "calc.nakspline", luaL_typename(L, -1));
  if (agn_seqsize(L, -1) != n)
    luaL_error(L, "Error in " LUA_QS ": number of coefficients not equal to number of points, got .", "calc.nakspline");
  for (i=0; i < n; i++) {
    x[i] = lua_seqgetinumber(L, -1, i + 1);
  }
  agn_poptop(L);
}


static int calc_naksplinecoeffs (lua_State *L) {
  size_t n;
  lua_Number *x, *y, *b, *c, *d;
  n = agn_nops(L, 1);  /* returns 0 if object at idx is not a structure */
  if (n < 4) {  /* 1.12.9 fix to prevent invalid array access in cubic_nak */
    lua_pushfail(L);
    return 1;
  }
  agn_createarray(x, n, "calc.naksplinecoeffs");
  agn_createarray(y, n, "calc.naksplinecoeffs");
  agn_createarray(b, n, "calc.naksplinecoeffs");
  agn_createarray(c, n, "calc.naksplinecoeffs");
  agn_createarray(d, n, "calc.naksplinecoeffs");
  agn_toarray(L, 1, "calc.naksplinecoeffs", n, x, y);
  if (cubic_nak(n, x, y, b, c, d) == -1) {
    xfreeall(x, y, b, c, d);  /* 1.12.9 / 2.2.0 RC 3 fixes, 2.9.8 */
    luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", "calc.naksplinecoeffs");
  }
  agn_createseq(L, 3);
  agn_arraytoseq(L, b, n);
  lua_seqseti(L, -2, 1);
  agn_arraytoseq(L, c, n);
  lua_seqseti(L, -2, 2);
  agn_arraytoseq(L, d, n);
  lua_seqseti(L, -2, 3);
  xfreeall(x, y, b, c, d);  /* 1.12.9 fix, 2.9.8 */
  return 1;
}


static int calc_clampedsplinecoeffs (lua_State *L) {
  size_t n;
  lua_Number *x, *y, *b, *c, *d, da, db;
  n = agn_nops(L, 1);  /* returns 0 if object at idx is not a structure */
  if (n < 2) {  /* 1.12.9 fix to prevent invalid array access in cubic_nak */
    lua_pushfail(L);
    return 1;
  }
  lua_pushvalue(L, 2);
  agn_pairgetinumbers(L, "calc.clampedsplinecoeffs", -1, &da, &db);  /* removes pair pushed on the stack */
  agn_createarray(x, n, "calc.clampedsplinecoeffs");
  agn_createarray(y, n, "calc.clampedsplinecoeffs");
  agn_createarray(b, n, "calc.clampedsplinecoeffs");
  agn_createarray(c, n, "calc.clampedsplinecoeffs");
  agn_createarray(d, n, "calc.clampedsplinecoeffs");
  agn_toarray(L, 1, "calc.clampedsplinecoeffs", n, x, y);
  if (cubic_clamped(n, x, y, b, c, d, da, db) == -1) {
    xfreeall(x, y, b, c, d);  /* 1.12.9 / 2.2.0 RC 3 fix, 2.9.8 */
    luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", "calc.clampedsplinecoeffs");
  }
  agn_createseq(L, 3);
  agn_arraytoseq(L, b, n);
  lua_seqseti(L, -2, 1);
  agn_arraytoseq(L, c, n);
  lua_seqseti(L, -2, 2);
  agn_arraytoseq(L, d, n);
  lua_seqseti(L, -2, 3);
  xfreeall(x, y, b, c, d);  /* 1.12.9 fix, 2.9.8 */
  return 1;
}


int nakspline (lua_State *L) {
  size_t i, found, n;
  lua_Number t, *x, *y, *b, *c, *d;
  i = 1;
  found = 0;
  t = agn_checknumber(L, 1);
  n = agn_tonumber(L, lua_upvalueindex(6));
  agn_createarray(x, n, "calc.nakspline");
  agn_createarray(y, n, "calc.nakspline");
  agn_createarray(b, n, "calc.nakspline");
  agn_createarray(c, n, "calc.nakspline");
  agn_createarray(d, n, "calc.nakspline");
  agn_seqtoarray(L, x, 1, n);
  agn_seqtoarray(L, y, 2, n);
  agn_seqtoarray(L, b, 3, n);
  agn_seqtoarray(L, c, 4, n);
  agn_seqtoarray(L, d, 5, n);
  while (!found && (i < n-1)) {
    if (t < x[i])
      found = 1;
    else
      i++;
  }
  t = y[i - 1] + (t - x[i - 1]) * (b[i - 1] + (t - x[i - 1]) * (c[i - 1] + (t - x[i - 1]) * d[i - 1]));
  xfreeall(x, y, b, c, d);  /* 2.9.8 */
  lua_pushnumber(L, t);
  return 1;
}


static int calc_nakspline (lua_State *L) {
  size_t n, nargs;
  lua_Number t, *x, *y, *b, *c, *d;
  n = agn_nops(L, 1);  /* returns 0 if object at idx is not a structure */
  if (n < 4) {  /* 1.12.9 fix to prevent invalid array access in cubic_nak */
    lua_pushfail(L);
    return 1;
  }
  nargs = lua_gettop(L);
  agn_createarray(x, n, "calc.nakspline");
  agn_createarray(y, n, "calc.nakspline");
  agn_createarray(b, n, "calc.nakspline");
  agn_createarray(c, n, "calc.nakspline");
  agn_createarray(d, n, "calc.nakspline");
  agn_toarray(L, 1, "calc.nakspline", n, x, y);
  if (nargs == 3 && lua_type(L, 3) == LUA_TSEQ) {
    if (agn_seqsize(L, 3) != 3) {
      xfreeall(x, y, b, c, d);  /* 2.2.0 RC 3 fix, 2.9.8 */
      luaL_error(L, "Error in " LUA_QS ": expected a sequence of three sequences.", "calc.nakspline");
    }
    agn_nakcoeffs(L, 3, 1, b, n);
    agn_nakcoeffs(L, 3, 2, c, n);
    agn_nakcoeffs(L, 3, 3, d, n);
  } else {
    if (cubic_nak(n, x, y, b, c, d) == -1) {
      xfreeall(x, y, b, c, d);  /* 1.12.9 / 2.2.0 RC 3 fix, 2.9.8 */
      luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", "calc.nakspline");
    }
  }
  if (nargs > 1) {
    t = agn_checknumber(L, 2);
    lua_pushnumber(L, spline_eval(n, x, y, b, c, d, t));
  } else {
    agn_arraytoseq(L, x, n);  /* push x as a sequence */
    agn_arraytoseq(L, y, n);  /* push y as a sequence */
    agn_arraytoseq(L, b, n);  /* push b as a sequence */
    agn_arraytoseq(L, c, n);  /* push c as a sequence */
    agn_arraytoseq(L, d, n);  /* push d as a sequence */
    lua_pushnumber(L, n);     /* push number of elements */
    lua_pushcclosure(L, &nakspline, 6);
  }
  xfreeall(x, y, b, c, d);  /* 1.12.9 fix, 2.9.8 */
  return 1;
}


static int calc_clampedspline (lua_State *L) {
  size_t n, nargs;
  lua_Number t, *x, *y, *b, *c, *d, da, db;
  n = agn_nops(L, 1);  /* returns 0 if object at idx is not a structure */
  if (n < 2) {  /* 1.12.9 fix to prevent invalid array access in cubic_nak */
    lua_pushfail(L);
    return 1;
  }
  nargs = lua_gettop(L);
  if (nargs < 2) {
    luaL_error(L, "Error in " LUA_QS ": at least two arguments expected.", "calc.clampedspline");
  }
  lua_pushvalue(L, 2);
  agn_pairgetinumbers(L, "calc.clampedsplinecoeffs", -1, &da, &db);  /* removes pair pushed on the stack */
  agn_createarray(x, n, "calc.clampedspline");
  agn_createarray(y, n, "calc.clampedspline");
  agn_createarray(b, n, "calc.clampedspline");
  agn_createarray(c, n, "calc.clampedspline");
  agn_createarray(d, n, "calc.clampedspline");
  agn_toarray(L, 1, "calc.clampedspline", n, x, y);
  if (nargs == 4 && lua_type(L, 4) == LUA_TSEQ) {
    if (agn_seqsize(L, 4) != 3) {
      xfreeall(x, y, b, c, d);  /* 2.2.0 RC 3 fix, 2.9.8 */
      luaL_error(L, "Error in " LUA_QS ": expected a sequence of three sequences.", "calc.clampedspline");
    }
    agn_nakcoeffs(L, 4, 1, b, n);
    agn_nakcoeffs(L, 4, 2, c, n);
    agn_nakcoeffs(L, 4, 3, d, n);
  } else {
    if (cubic_clamped(n, x, y, b, c, d, da, db) == -1) {
      xfreeall(x, y, b, c, d);  /* 2.2.0 RC 3 fix, 2.9.8 */
      luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", "calc.clampedspline");
    }
  }
  if (nargs > 2) {
    t = agn_checknumber(L, 3);
    lua_pushnumber(L, spline_eval(n, x, y, b, c, d, t));
  } else {
    agn_arraytoseq(L, x, n);  /* push x as a sequence */
    agn_arraytoseq(L, y, n);  /* push y as a sequence */
    agn_arraytoseq(L, b, n);  /* push b as a sequence */
    agn_arraytoseq(L, c, n);  /* push c as a sequence */
    agn_arraytoseq(L, d, n);  /* push d as a sequence */
    lua_pushnumber(L, n);     /* push number of elements */
    lua_pushcclosure(L, &nakspline, 6);
  }
  xfreeall(x, y, b, c, d);  /* 1.12.9 fix, 2.9.8 */
  return 1;
}


static int polygen_generator (lua_State *L) {
  size_t i;
  lua_Number r, x, q;
  NumArray *a;
  x = agn_checknumber(L, 1);  /* get argument */
  a = (NumArray *)lua_touserdata(L, lua_upvalueindex(1));
  r = getarray(a, 0);  /* coefficient C[n] */
  q = 0;
  for (i=1; i < a->size; i++)
    r = tools_koadd(r*x, getarray(a, i), &q);  /* 2.13.0 */
  lua_pushnumber(L, r);
  return 1;
}

static int calc_polygen (lua_State *L) {
  size_t i, nops;
  NumArray *a;
  nops = lua_gettop(L);
  luaL_checkstack(L, nops, "too many arguments");
  if (nops == 0)
    luaL_error(L, "Error in " LUA_QS ": expected at least one number.", "calc.polygen");
  else if (nops > 256)  /* 2.8.2, there allegedly is a bug in Lua when more than 256 arguments are pushed into the userdata stack */
    luaL_error(L, "Error in " LUA_QS ": received too many arguments.", "calc.polygen");
  aux_newnumarray(L, nops);  /* push userdata a */
  a = (NumArray *)lua_touserdata(L, -1);
  for (i=0; i < nops; i++)  /* register coefficients in *a */
    setarray(a, i, agn_checknumber(L, i + 1));
  lua_pushcclosure(L, &polygen_generator, 1);  /* register userdata in C closure */
  return 1;
}


/* source: http://programbank4u.blogspot.de/2013/04/lagrange-interpolation-c-program.html */

static int linterp (lua_State *L) {  /* 2.1.3 */
  size_t n;
  int i, j;
  lua_Number x0, sum, prod;
  NumArray *x, *y;
  x0 = agn_checknumber(L, 1);
  x = (NumArray *)lua_touserdata(L, lua_upvalueindex(1));
  y = (NumArray *)lua_touserdata(L, lua_upvalueindex(2));
  n = x->size;
  sum = 0;
  for (i=0; i < n; i++) {
    prod = 1;
    for (j=0; j < n; j++) {
      if (j == i) continue;
      prod *= (x0 - getarray(x, j)) / (getarray(x, i) - getarray(x, j));
    }
    sum += getarray(y, i) * prod;
  }
  lua_pushnumber(L, sum);
  return 1;
}


/* calc.linterp

   returns a function that conducts a Lagrange interpolation for a given sequence or table o of numeric pairs
   x:y where x and y denote a point in the plane. It is often said that Lagrange interpolation is suited for
   theoretical purposes only, since it is also very slow. */

static int calc_linterp (lua_State *L) {
  size_t n, i;
  int type;
  lua_Number x, y;
  NumArray *a, *b;
  type = lua_type(L, 1);
  if ((type != LUA_TSEQ && type != LUA_TTABLE) || ( n = agn_nops(L, 1) ) == 0) {
    luaL_error(L, "Error in " LUA_QS ": expected a sequence or table of one or more pairs.", "calc.linterp");
    return 1;
  }
  aux_newnumarray(L, n);  /* push userdata a */
  aux_newnumarray(L, n);  /* push userdata b */
  a = (NumArray *)lua_touserdata(L, -2);
  b = (NumArray *)lua_touserdata(L, -1);
  if (type == LUA_TSEQ) {
    for (i=1; i <= n; i++) {
      lua_seqgeti(L, 1, i);
      agn_pairgetinumbers(L, "calc.linterp", -1, &x, &y);  /* pops the pair after evaluation */
      setarray(a, i - 1, x);
      setarray(b, i - 1, y);
    }
  } else {
    for (i=1; i <= n; i++) {
      lua_rawgeti(L, 1, i);
      agn_pairgetinumbers(L, "calc.linterp", -1, &x, &y);  /* pops the pair after evaluation */
      setarray(a, i - 1, x);
      setarray(b, i - 1, y);
    }
  }
  lua_pushcclosure(L, &linterp, 2);
  return 1;
}


/* 2.1.5, Simpson-Simpson Adaptive Quadrature, taken from:
   http://www.mymathlib.com/quadrature/adaptive.html, by RLH, Copyright  2004 RLH. All rights reserved.
   Adapted for Agena. */

struct Subinterval {
  lua_Number upper_limit;
  lua_Number lower_limit;
  lua_Number function[5];
  struct Subinterval *interval;
};

/* Description:

   Starting at the left-end point, a, find the min power of 2, m, so that the difference between using
   Simpson's rule and the composite Simpson's rule on the interval [a, a+(b-a)/2^m] is less than twice
   the tolerance * (length of the subinterval)/(b-a).  Then repeat the process for integrating over the
   interval [(a+(b-a)/2^m, b] until the right end point, b, is finally reached.

   The integral is then the sum of the integrals of each subinterval. If at any time, the length of the
   subinterval for which the estimates based on Simpson's rule and the composite Simpson's rule is less
   than min_h the process is terminated.

  Arguments:
     double a          The lower limit of the integration interval.
     double b          The upper limit of integration.
     double tolerance  The acceptable error estimate of the integral. The integral of a subinterval is
                       accepted when the magnitude of the estimated error falls below the pro-rated tolerance
                       of the subinterval.
     double min_h      The minimum subinterval length. If no subinterval of length > min_h is found for which
                       the estimated error falls below the pro-rated tolerance, the process terminates.

  Return:
     The integral of f(x) from a to a +  h, or `fail' if no subinterval of length > min_h was found for which
     the estimated error was less that the pro-rated error.


  Source:
     http://www.mymathlib.com/c_source/quadrature/adaptive/simpson_simpson.c, adapted by alexander walz */

static void Simpsons_Rule_Update (lua_State *L, struct Subinterval *pinterval, lua_Number *s1, lua_Number *s2) {
  lua_Number h, h4;
  int error;
  h = pinterval->upper_limit - pinterval->lower_limit;
  h4 = 0.25 * h;
  lua_pushvalue(L, 1);
  lua_pushnumber(L, pinterval->lower_limit + h4);
  pinterval->function[1] = agn_ncall(L, 1, 1, &error, 1);
  lua_pushvalue(L, 1);
  lua_pushnumber(L, pinterval->upper_limit - h4);
  pinterval->function[3] = agn_ncall(L, 1, 1, &error, 1);
  *s1 = pinterval->function[0] + 4.0 * pinterval->function[2]
       + pinterval->function[4];
  *s1 *= 0.166666666666666666666667 * h;
  *s2 = pinterval->function[0] + 4.0 * pinterval->function[1]
       + 2.0 * pinterval->function[2] + 4.0 * pinterval->function[3]
       + pinterval->function[4];
  *s2 *= 0.0833333333333333333333333 * h;
}

static int calc_simaptive (lua_State *L) {
  struct Subinterval *pinterval, *qinterval;
  int err, error;
  lua_Number integral, epsilon_density, epsilon, s1, s2, a, b, tolerance, min_h;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  min_h = agnL_optnumber(L, 4, 1e-7);
  tolerance = agnL_optnumber(L, 5, 0.5*agn_geteps(L));
  integral = 0.0;
  epsilon_density = 2.0 * tolerance / (b - a);
  /* Create the initial level, with lower_limit = a, upper_limit = b, and f(x) evaluated at a, b, and (a + b) / 2. */
  pinterval = (struct Subinterval*)agn_malloc(L, sizeof(struct Subinterval), "calc.simaptive", NULL);
  pinterval->interval = NULL;
  pinterval->upper_limit = b;
  pinterval->lower_limit = a;
  lua_pushvalue(L, 1);  /* push function */
  lua_pushnumber(L, pinterval->lower_limit);
  pinterval->function[0] = agn_ncall(L, 1, 1, &error, 1);
  lua_pushvalue(L, 1);  /* push function */
  lua_pushnumber(L, 0.5 * (pinterval->upper_limit + pinterval->lower_limit));
  pinterval->function[2] = agn_ncall(L, 1, 1, &error, 1);
  lua_pushvalue(L, 1);  /* push function */
  lua_pushnumber(L, pinterval->upper_limit);
  pinterval->function[4] = agn_ncall(L, 1, 1, &error, 1);
  /* Calculate the tolerance for the current interval. Calculate the single subinterval Simpson rule,
     and the two subintervals composite Simpson rule. */
  err = 1;
  epsilon = epsilon_density * (b - a);
  Simpsons_Rule_Update(L, pinterval, &s1, &s2);
  while (pinterval->upper_limit - pinterval->lower_limit > min_h) {
    if (fabs(s1 - s2) < epsilon) {
      /* If the two estimates are close, then increment the integral and if we are not at the right end,
         set the left end of the new interval to the right end of the old interval and the right end of
         the new interval remains the same (as the previous right end for this interval. */
      integral += s2;
      if (pinterval->interval == NULL) {
        err = 0;
        break;
      }
      qinterval = pinterval->interval;
      qinterval->lower_limit = pinterval->upper_limit;
      qinterval->function[0] = qinterval->function[2];
      qinterval->function[2] = qinterval->function[3];
      xfree(pinterval);
      pinterval = qinterval;
    }
    else {
      /* If the two estimates are not close, then create a new interval with same left end point and right
         end point at the midpoint of the current interval. */
      qinterval = (struct Subinterval*)agn_malloc(L, sizeof(struct Subinterval), "calc.simaptive", NULL);
      qinterval->interval = pinterval;
      qinterval->lower_limit = pinterval->lower_limit;
      qinterval->upper_limit = 0.5 * (pinterval->upper_limit + pinterval->lower_limit);
      qinterval->function[0] = pinterval->function[0];
      qinterval->function[2] = pinterval->function[1];
      qinterval->function[4] = pinterval->function[2];
      pinterval = qinterval;
    }
    Simpsons_Rule_Update(L, pinterval, &s1, &s2);
    epsilon = epsilon_density * (pinterval->upper_limit - pinterval->lower_limit);
  }
  while (pinterval != NULL) {
    qinterval = pinterval->interval;
    xfree(pinterval);
    pinterval = qinterval;
  }
  lua_pushnumber(L, (err) ? AGN_NAN : integral);
  return 1;
}

/* Added January 09, 2008; 0.9.0 / July 12, 2008; 0.12.0; C version: 2.1.5, 04.03.2014.
   Integration of a function on an interval using a modified bisection method. Algorithm taken from
   `Fortran 90 Kurs` by G. Schmitt, Oldenbourg Verlag, 1996, p. 119 */

static int calc_gtrap (lua_State *L) {
  int k, n, error;
  lua_Number delta, dold, h, sum, fnew, fold, a, b, eps, fa, fb, fah;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  eps = agnL_optnonnegative(L, 4, agn_geteps(L));  /* 2.13.0 */
  dold = 1;
  n = 2;
  h = (b - a)/n;
  lua_pushvalue(L, 1);
  lua_pushnumber(L, a);
  fa = agn_ncall(L, 1, 1, &error, 1);
  lua_pushvalue(L, 1);
  lua_pushnumber(L, a + h);
  fah = agn_ncall(L, 1, 1, &error, 1);
  lua_pushvalue(L, 1);
  lua_pushnumber(L, b);
  fb = agn_ncall(L, 1, 1, &error, 1);
  sum = fa + 2*fah + fb;
  fold = h * sum/2;
  do {
    n = 2*n;
    h = h/2;
    for (k=1; k < n; k = k + 2) {
      lua_pushvalue(L, 1);
      lua_pushnumber(L, a + (lua_Number)k*h);
      sum += 2 * agn_ncall(L, 1, 1, &error, 1);
    }
    fnew = h * sum/2;
    delta = fabs((fnew - fold)/fnew);
    fold = fnew;
    dold = delta;
  } while (delta > eps || delta < dold);
  lua_pushnumber(L, fnew);
  return 1;
}


/* The Airy wave function returns both the first independent solution to the differential equation y"(x) = x*y and its first derivative, for any real x. */
static int calc_Ai (lua_State *L) {  /* 2.8.1 */
  lua_Number x, ai, bi, dai, dbi;
  int r;
  x = agn_checknumber(L, 1);
  r = airy(x, &ai, &dai, &bi, &dbi);
  if (r == -1)
    lua_pushfail(L);
  else {
    lua_pushnumber(L, ai);
    lua_pushnumber(L, dai);
  }
  return 1 + (r == 0);
}


/* The Airy wave function returns both the second independent solution to the differential equation y"(x) = x*y and its first derivative, for any real x. */
static int calc_Bi (lua_State *L) {  /* 2.8.1 */
  lua_Number x, ai, bi, dai, dbi;
  int r;
  x = agn_checknumber(L, 1);
  r = airy(x, &ai, &dai, &bi, &dbi);
  if (r == -1)
    lua_pushfail(L);
  else {
    lua_pushnumber(L, bi);
    lua_pushnumber(L, dbi);
  }
  return 1 + (r == 0);
}

/* Computes the Riemann Zeta function in the real domain, using the zetac function.
 *
 *                inf.
 *                 -    -x
 *   zetac(x)  =   >   k   ,   x > 1,
 *                 -
 *                k=2
 *
 * is related to the Riemann zeta function by
 *
 *	Riemann zeta(x) = zetac(x) + 1. */

static int calc_zeta (lua_State *L) { /* 2.8.1 */
  lua_pushnumber(L, zetac(agn_checknumber(L, 1)) + 1);
  return 1;
}


/* Returns the polylogarithm of order n at a real point x. The return is a number, or `fail` if n < -1 for this situation is
   not implemented.

 * The polylogarithm of order n is defined by the series
 *
 *              inf   k
 *               -   x
 *  Li (x)  =    >   ---  .
 *    n          -     n
 *              k=1   k
*/

static int calc_polylog (lua_State *L) {  /* 2.8.1 */
  int flag;
  lua_Number r;
  r = polylog(agn_checkinteger(L, 1), agn_checknumber(L, 2), &flag);
  if (flag)  /* not implemented for n < -1 */
    lua_pushfail(L);
  else
    lua_pushnumber(L, r);
  return 1;
}


/* Evaluates the exponential integral
 *
 *                 inf.
 *                   -
 *                  | |   -xt
 *                  |    e
 *      E (x)  =    |    ----  dt.
 *       n          |      n
 *                | |     t
 *                 -
 *                  1

  for non-negative n and real x. The return is a number. */

static int calc_En (lua_State *L) {  /* 2.8.1 */
  lua_pushnumber(L, expn(agn_checkinteger(L, 1), agn_checknumber(L, 2)));
  return 1;
}


/* Evaluates the incomplete gamma integral defined by
 *
 *                           x
 *                            -
 *                   1       | |  -t  a-1
 *  igam(a,x)  =   -----     |   e   t   dt.
 *                  -      | |
 *                 | (a)    -
 *                           0
 *
 *
 * In this implementation both arguments must be positive. */

static int calc_igamma (lua_State *L) {  /* 2.8.1 */
  lua_Number a, x;
  x = agn_checknumber(L, 1);
  a = agn_checknumber(L, 2);
  if (x <= 0 || a <= 0)
    lua_pushfail(L);
  else
    lua_pushnumber(L, igam(a, x));
  return 1;
}


/* Evaluates the complemented incomplete gamma integral defined by
 *
 *  igamc(a,x)   =   1 - igam(a,x)
 *
 *                            inf.
 *                              -
 *                     1       | |  -t  a-1
 *               =   -----     |   e   t   dt.
 *                    -      | |
 *                   | (a)    -
 *                             x
 *
 * In this implementation both arguments must be positive. */
static int calc_igammac (lua_State *L) {  /* 2.8.1 */
  lua_Number a, x;
  x = agn_checknumber(L, 1);
  a = agn_checknumber(L, 2);
  if (x <= 0 || a <= 0)
    lua_pushfail(L);
  else
    lua_pushnumber(L, igamc(a, x));
  return 1;
}


/* Returns the incomplete beta integral of the arguments, evaluated
 * from zero to x.  The function is defined as
 *
 *                  x
 *     -            -
 *    | (a+b)      | |  a-1     b-1
 *  -----------    |   t   (1-t)   dt.
 *   -     -     | |
 *  | (a) | (b)   -
 *                 0
 *
 * The domain of definition is 0 <= x <= 1.  In this
 * implementation a and b are restricted to positive values. */

static int calc_ibeta (lua_State *L) {  /* 2.8.2 */
  lua_Number a, b, x;
  x = agn_checknumber(L, 1);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  if (a <= 0 || b <= 0)
    lua_pushfail(L);
  else if (x < 0 || x > 1)
    lua_pushundefined(L);
  else
    lua_pushnumber(L, incbet(a, b, x));
  return 1;
}


/* Inverse of incomplete beta integral
 *
 * SYNOPSIS:
 *
 * double a, b, x, y, incbi();
 *
 * x = incbi( a, b, y );
 *
 * DESCRIPTION:
 *
 * Given y, the function finds x such that
 *
 *  incbet( a, b, x ) = y . */

static int calc_invibeta (lua_State *L) {  /* 2.8.2 */
  lua_Number y, a, b;
  y = agn_checknumber(L, 1);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  lua_pushnumber(L, incbi(a, b, y));
  return 1;
}


/* Returns the limit of a function at a given point. If the left or right limit does not exist, `undefined`
   is returned. The function returns `undefined` instead of +/- infinity if the left and the right limit are either
   +infinity or -infinity. 2.7 times faster than the Agena version.

   f := << x -> (x^3+2*x^2-4*x-8)/(x^2+4*x+4) >>, reduced to: f := << x -> x - 2 >>, singularity at x = -2 */

static int calc_limit (lua_State *L) {  /* 2.9.7 */
  lua_Number x, a, b, eps, delta, r;
  int nargs, sided, deriv;
  const char *option = NULL;
  nargs = lua_gettop(L);
  luaL_checktype(L, 1, LUA_TFUNCTION);
  x = agn_checknumber(L, 2);
  (void)delta; (void)deriv;
  aux_geteps(L, x, nargs, "calc.limit");  /* changed 2.9.8 */
  /* sun_nextafter may return undefined or a wrong finite value in function calls within the vicinity of singularities, e.g.
     f := << x -> 1/x^2 >> at x = 0, or f := << x -> (x^2-3)/(x-3) >> at x = 3 */
  a = agnL_fncall(L, 1, x - eps, 3, nargs, 1);  /* 2.9.8 */
  b = agnL_fncall(L, 1, x + eps, 3, nargs, 1);  /* 2.9.8 */
  switch (sided) {  /* 2.9.8 */
    case 1: {  /* left-sided limit */
      lua_pushnumber(L, fabs(a) < eps ? 0 : a);
      return 1;
    }
    case 2: {  /* right-sided limit */
      lua_pushnumber(L, fabs(b) < eps ? 0 : b);
      return 1;
    }
    case 3: {  /* both left and right-sided limits */
      lua_pushnumber(L, fabs(a) < eps ? 0 : a);
      lua_pushnumber(L, fabs(b) < eps ? 0 : b);
      return 2;
    }
  }
  if (isunordered(a, b))  /* at least one function value is undefined ? */
    lua_pushundefined(L);
  else if ( (a == HUGE_VAL && b == HUGE_VAL) || (a == -HUGE_VAL && b == -HUGE_VAL) )
    lua_pushnumber(L, a);
  else {
    if (tools_approx(a, b, 2*eps)) {  /* 2.9.8 fix, lim sin(x), x=0, returnd undefined for check failed */
      r = (a + b)/2;
      lua_pushnumber(L, (fabs(r) < eps) ? 0 : r);  /* apply `math.zeroin` */
    } else
      lua_pushundefined(L);
  }
  if (sided == 4) {  /* 2.9.8, return left and right-sided limit, as well */
    lua_pushnumber(L, fabs(a) < eps ? 0 : a);
    lua_pushnumber(L, fabs(b) < eps ? 0 : b);
  }
  return 1 + 2 * (sided == 4);
}


/* Returns `true` if a real function f is continuous at the given point x (a number), and `false` otherwise.
  Three times faster than the Agena implementation. */
static int calc_iscont (lua_State *L) {  /* 2.9.7 */
  lua_Number x, a, b, fx, eps, delta;
  int nargs, sided, deriv;
  (void)delta; (void)sided; (void)deriv;
  const char *option = NULL;
  nargs = lua_gettop(L);
  luaL_checktype(L, 1, LUA_TFUNCTION);
  x = agn_checknumber(L, 2);
  aux_geteps(L, x, nargs, "calc.iscont");  /* changed 2.9.8 */
  fx = agnL_fncall(L, 1, x, 3, nargs, 1);  /* 2.9.8 */
  if (isnan(fx))  /* 2.10.1, undefined ? */
    lua_pushfalse(L);
  else {
    a = agnL_fncall(L, 1, sun_nextafter(x, -HUGE_VAL), 3, nargs, 1);  /* 2.9.8 */
    b = agnL_fncall(L, 1, sun_nextafter(x, HUGE_VAL), 3, nargs, 1);  /* 2.9.8 */
    lua_pushboolean(L, isfinite(a) && isfinite(b) && tools_approx(a, fx, eps) && tools_approx(fx, b, eps));  /* 2.9.8 */
  }
  return 1;
}


/* Computes the Euclidian distance, i.e. the straight-line distance, of two points (a, f(a)) and
   (b, f(b)) on a curve defined by a function f in one real, in the Euclidean plane. a, b must be numbers.
   Around twice as fast than an Agena implementation. */
static int calc_eucliddist (lua_State *L) {  /* 2.9.7 */
  lua_Number a, b, fa, fb;
  int nargs;
  nargs = lua_gettop(L);
  luaL_checktype(L, 1, LUA_TFUNCTION);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  /* determine f(a) */
  fa = agnL_fncall(L, 1, a, 4, nargs, 1);  /* 2.9.8 */
  /* determine f(b) */
  fb = agnL_fncall(L, 1, b, 4, nargs, 1);  /* 2.9.8 */
  /* splitting the formula into several separate steps is not faster */
  lua_pushnumber(L, sun_hypot(a - b, fa - fb));  /* 2.9.8 improvement */
  return 1;
}


/* The function checks for the differentiability of a function f at a point x. It computes the left and right-sided
   difference quotients of f at x. If both are equal and the limit of f at x is equal to f(x), the function returns
   `true`, and `false` otherwise. */
static int calc_isdiff (lua_State *L) {  /* 2.10.1 */
  lua_Number x, fx, eps, delta, fph, fmh, f1ph, f1mh;
  int nargs, sided, deriv;
  (void)delta; (void)sided; (void)deriv;  /* to prevent compiler warnings */
  const char *option = NULL;
  nargs = lua_gettop(L);
  luaL_checktype(L, 1, LUA_TFUNCTION);
  x = agnL_checknumber(L, 2);
  aux_geteps(L, x, nargs, "calc.isdiff");
  /* check if f is defined at x */
  fx = agnL_fncall(L, 1, x, 3, nargs, 1);
  if (!isfinite(fx)) {  /* infinity or undefined ? */
    lua_pushfalse(L);
    return 1;
  }
  fmh = agnL_fncall(L, 1, x - eps, 3, nargs, 1);
  f1mh = (fx - fmh)/eps;  /* left-sided difference quotient */
  fph = agnL_fncall(L, 1, x + eps, 3, nargs, 1);
  f1ph = (fph - fx)/eps;  /* right-sided difference quotient */
  lua_pushboolean(L, tools_approx(f1mh, f1ph, eps) && tools_approx(fmh, fx, eps) && tools_approx(fph, fx, eps));
  return 1;
}


static lua_Number chebytwobynops[101] = {
  0.0,
  2.0000000000000000000000,
  1.0000000000000000000000,
  0.6666666666666666666667,
  0.5000000000000000000000,
  0.4000000000000000000000,
  0.3333333333333333333333,
  0.2857142857142857142857,
  0.2500000000000000000000,
  0.2222222222222222222222,
  0.2000000000000000000000,
  0.1818181818181818181818,
  0.1666666666666666666667,
  0.1538461538461538461538,
  0.1428571428571428571429,
  0.1333333333333333333333,
  0.1250000000000000000000,
  0.1176470588235294117647,
  0.1111111111111111111111,
  0.1052631578947368421053,
  0.1000000000000000000000,
  0.0952380952380952380952,
  0.0909090909090909090909,
  0.0869565217391304347826,
  0.0833333333333333333333,
  0.0800000000000000000000,
  0.0769230769230769230769,
  0.0740740740740740740741,
  0.0714285714285714285714,
  0.0689655172413793103448,
  0.0666666666666666666667,
  0.0645161290322580645161,
  0.0625000000000000000000,
  0.0606060606060606060606,
  0.0588235294117647058824,
  0.0571428571428571428571,
  0.0555555555555555555556,
  0.0540540540540540540541,
  0.0526315789473684210526,
  0.0512820512820512820513,
  0.0500000000000000000000,
  0.0487804878048780487805,
  0.0476190476190476190476,
  0.0465116279069767441860,
  0.0454545454545454545455,
  0.0444444444444444444444,
  0.0434782608695652173913,
  0.0425531914893617021277,
  0.0416666666666666666667,
  0.0408163265306122448980,
  0.0400000000000000000000,
  0.0392156862745098039216,
  0.0384615384615384615385,
  0.0377358490566037735849,
  0.0370370370370370370370,
  0.0363636363636363636364,
  0.0357142857142857142857,
  0.0350877192982456140351,
  0.0344827586206896551724,
  0.0338983050847457627119,
  0.0333333333333333333333,
  0.0327868852459016393443,
  0.0322580645161290322581,
  0.0317460317460317460317,
  0.0312500000000000000000,
  0.0307692307692307692308,
  0.0303030303030303030303,
  0.0298507462686567164179,
  0.0294117647058823529412,
  0.0289855072463768115942,
  0.0285714285714285714286,
  0.0281690140845070422535,
  0.0277777777777777777778,
  0.0273972602739726027397,
  0.0270270270270270270270,
  0.0266666666666666666667,
  0.0263157894736842105263,
  0.0259740259740259740260,
  0.0256410256410256410256,
  0.0253164556962025316456,
  0.0250000000000000000000,
  0.0246913580246913580247,
  0.0243902439024390243902,
  0.0240963855421686746988,
  0.0238095238095238095238,
  0.0235294117647058823529,
  0.0232558139534883720930,
  0.0229885057471264367816,
  0.0227272727272727272727,
  0.0224719101123595505618,
  0.0222222222222222222222,
  0.0219780219780219780220,
  0.0217391304347826086957,
  0.0215053763440860215054,
  0.0212765957446808510638,
  0.0210526315789473684211,
  0.0208333333333333333333,
  0.0206185567010309278351,
  0.0204081632653061224490,
  0.0202020202020202020202,
  0.0200000000000000000000
};

static int cheby (lua_State *L) {  /* 2.10.4 */
  lua_Number p, q, x, y, y2, d0, d1, d2, e0, e1, e2, f0, f1, f2, g0, g1, g2, h0, h1, h2, i0, i1, i2;
  int i, deriv;
  NumArray *a;
  x = agn_checknumber(L, 1);  /* get argument */
  a = (NumArray *)lua_touserdata(L, lua_upvalueindex(1));
  p = getarray(a, a->size);
  q = getarray(a, a->size + 1);
  deriv = (int)getarray(a, a->size + 2);
  d1 = d2 = e1 = e2 = f1 = f2 = g1 = g2 = h1 = h2 = i1 = i2 = 0;
  d0 = e0 = f0 = g0 = h0 = i0 = getarray(a, a->size - 1);
  y = (2.0*x - p - q)/(q - p);
  for (i=a->size - 2; i >= 0; i--) {
    d2 = d1; d1 = d0;
    y2 = 2.0*y;
    d0 = y2*d1 - d2 + getarray(a, i);  /* term 2*y*d1 with factor 2 necessary for deriv = 0 */
    if (deriv > 0 && i > 0) {  /* first derivative, `fall through` */
      e2 = e1; e1 = e0;
      e0 = d0 - e2 + y2*e1;
    }  /* continue statements slow down the function */
    if (deriv > 1 && i > 1) {  /* second derivative, `fall through` */
      f2 = f1; f1 = f0;
      f0 = e0 - f2 + y2*f1;
    }  /* continue statements slow down the function */
    if (deriv > 2 && i > 2) {  /* third derivative, `fall through` */
      g2 = g1; g1 = g0;
      g0 = f0 - g2 + y2*g1;
    }  /* continue statements slow down the function */
    if (deriv > 3 && i > 3) {  /* fourth derivative, `fall through` */
      h2 = h1; h1 = h0;
      h0 = g0 - h2 + y2*h1;
    }  /* continue statements slow down the function */
    if (deriv > 4 && i > 4) {  /* fifth derivative, `fall through` */
      i2 = i1; i1 = i0;
      i0 = h0 - i2 + y2*i1;
    }
  }
  switch (deriv) {  /* ? factor according to http://oeis.org/A038110 ? */
    /* substituting call to tools_intpow with z*z*.. is not faster */
    case 0:
      lua_pushnumber(L,   0.5*(d0 - d2)); break;  /* fabs(q - p)^0 = 1 */
    case 1:
      lua_pushnumber(L,     2*(e0 - e2)/fabs(q - p)); break;  /* fabs(q - p)^1 = fabs(q - p) */
    case 2:
      lua_pushnumber(L,    16*(f0 - f2)/tools_intpow(fabs(q - p), 2)); break;
    case 3:
      lua_pushnumber(L,   192*(g0 - g2)/tools_intpow(fabs(q - p), 3)); break;
    case 4:
      lua_pushnumber(L,  3072*(h0 - h2)/tools_intpow(fabs(q - p), 4)); break;
    case 5:
      lua_pushnumber(L, 61440*(i0 - i2)/tools_intpow(fabs(q - p), 5)); break;
    default:  /* should not happen */
      luaL_error(L, "Error in " LUA_QS ": cannot compute derivative.", "calc.cheby");
  }
  return 1;
}

/* Returns a function computing the Chebyshev interpolant for a given point. `f` is the function to be interpolated,
   p and q represent the domain of the definition, `n` is the order of the interpolant. As a rule of thumb, the larger
   the domain, the larger n should be.

   You may optionally pass the `deriv=k` option as the very last argument to compute the first (k=1), second (k=2)
   or third (k=3) derivative, where k defaults to 0, i.e. the function itself.

   Using this function may speed up numeric computations significantly if the expression to be evaluated consists of
   many subexpressions - and if accuracy is not of primary concern. See also: calc.chebycoeffs.
   The function is thrice as fast as an Agena implementation. */
static int calc_cheby (lua_State *L) {  /* 2.10.4 */
  lua_Number p, q, *y, x, k, f;
  int i, j, nops, nargs, deriv;
  const char *option;
  NumArray *a;
  deriv = 0;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  p = agn_checknumber(L, 2);
  q = agn_checknumber(L, 3);
  if (p > q)
    luaL_error(L, "Error in " LUA_QS ": invalid domain %lf .. %lf.", "calc.cheby", p, q);
  nops = agn_checkposint(L, 4);
  nargs = lua_gettop(L);
  if (lua_ispair(L, nargs)) {
    agn_pairgeti(L, nargs, 1);
    if (!agn_isstring(L, -1)) {
      agn_poptop(L);
      luaL_error(L, "Error in " LUA_QS ": left-hand side of option must be a string.", "calc.cheby");
    }
    option = lua_tostring(L, -1);
    if (tools_streq(option, "deriv")) {  /* 2.16.12 tweak */
      agn_pairgeti(L, nargs, 2);
      if (agn_isnumber(L, -1)) {
        deriv = lua_tointeger(L, -1);
        agn_poptop(L);
        if (deriv < 0 || deriv > 5) {
          agn_poptoptwo(L); \
          luaL_error(L, "Error in " LUA_QS ": deriv must be in 1 .. 5.", "calc.cheby");
        };
      } else  {
        agn_poptoptwo(L);
        luaL_error(L, "Error in " LUA_QS ": right-hand side of deriv option must be a number.", "calc.cheby");
      }
    } else {
      agn_poptop(L);
      luaL_error(L, "Error in " LUA_QS ": unknown option.", "calc.cheby");
    }
    nargs--;
  }
  aux_newnumarray(L, nops + 3);  /* pushes NumArray userdata of Chebyshev coefficient vector onto stack with three ... */
  a = (NumArray *)lua_touserdata(L, -1);  /* ... additional information, see below */
  a->size = nops;
  if (a->size - deriv < 2)
    luaL_error(L, "Error in " LUA_QS ": vector would have too few coefficients.", "calc.cheby");
  y = malloc(nops * sizeof(lua_Number));
  k = PI/(2.0*nops);
  for (i=0; i < nops; i++) {
    /* x = (p + q)/2 + cos(k*(2*i + 1))*(q - p)/2.0; */
    x = 0.5*(p + q + sun_cos(k*(2*i + 1))*(q - p));  /* changed 2.10.4 */
    y[i] = agnL_fncall(L, 1, x, 5, nargs, 1);
  };
  f = (nops <= 100) ? chebytwobynops[nops] : 2.0/nops;
  for (i=0; i < nops; i++) {
    setarray(a, i, 0.0);
    for (j=0; j < nops; j++) {
      a->values[i] += y[j]*sun_cos(k*(i*(2*j + 1)));
    }
    a->values[i] *= f; /* 2.0/nops; */
  }
  xfree(y);
  setarray(a, nops++, p);  /* pack domain p .. q into NumArray just for speed ... */
  setarray(a, nops++, q);
  setarray(a, nops, deriv);  /* ... and derivative to be computed, as well */
  lua_pushcclosure(L, &cheby, 1);
  return 1;
}


/* Computes a SavitzkyGolay filter for the univariate function f to smooth its data by returning a factory interpolating f
   at a point.

   It fits successive subsets of neighbouring data with a low-degree polynomial using the linear least-square method. By
   default, 15 equally-spaced points to the left of x and 15 equally-spaced points points to the right of x are examined.
   You can change this `window` by passing another odd value with the 'points' option. All adjacent points are separated by
   distance eps which is 1e-5 by default. You can change the distance with the 'eps' option.

   Alternatively it can also compute derivatives of any degree n by passing the option deriv = <n>. The larger the degree n
   of the derivative, however, the least accurate the results will become.

   The degree d of the smoothing least-square polynomial is 3 by default and can be changed by the degree=<d> option. Recommended
   degrees are d = 2 or 4, with d not exceeding 6.

   The function automatically determines the most suitable settings for the window and the spacing eps of its points,
   but you can switch this off by passing the adaptive=false option.

   See also: calc.savgolcoeffs. */
static int savgol (lua_State *L) {  /* 2.10.4 */
  lua_Number x, signal, eps;
  int i, c, fact, nleft, nright, deriv;
  NumArray *a;
  x = agn_checknumber(L, 1);  /* get argument */
  a = (NumArray *)lua_touserdata(L, lua_upvalueindex(1));
  nleft = getarray(a, a->size);
  nright = getarray(a, a->size + 1);
  deriv = getarray(a, a->size + 2);
  eps = getarray(a, a->size + 3);
  signal = 0;
  c = 0;
  for (i=nleft; i <= nright; i++) {
    signal += getarray(a, c++)*agnL_fncall(L, lua_upvalueindex(2), x + i*eps, 10, 9, 1);  /* not a multi-variate function */
  };
  /* for formula below see: www.maths.lth.se/matstat/wafo/documentation/wafodoc/wafo/misc/savgol.html
     compute denominator separately to avoid overflows ((signal * n!) / ...) */
  fact = 1;
  for (i=1; i <= deriv; i++) fact *= i;
  lua_pushnumber(L, signal*(fact/tools_intpow(eps, deriv)));
  return 1;
}

static int calc_savgol (lua_State *L) {  /* 2.10.4 */
  int i, npoints, nleft, nright, polydeg, isadaptive, deriv, nops, nargs;
  lua_Number eps, factor;
  const char *option;
  NumArray *a;
  eps = 1e-5;      /* distance between sample points */
  npoints = 31;    /* number of neighbouring points to be examined */
  deriv = 0;       /* n-th derivative, default: compute smoothed function */
  polydeg = 3;     /* polynomial degree for smoothing */
  isadaptive = 1;  /* use adaptive measures to boost precision depending on the degree of the derivative */
  luaL_checktype(L, 1, LUA_TFUNCTION);
  nargs = lua_gettop(L);
  while (lua_ispair(L, nargs)) {
    agn_pairgeti(L, nargs, 1);
    if (!agn_isstring(L, -1)) {
      agn_poptop(L);
      luaL_error(L, "Error in " LUA_QS ": left-hand side of option must be a string.", "calc.savgol");
    }
    option = lua_tostring(L, -1);
    if (tools_streq(option, "deriv")) {  /* 2.16.12 tweak */
      agn_pairgeti(L, nargs, 2);
      if (agn_isnumber(L, -1)) {
        deriv = lua_tointeger(L, -1);
        agn_poptop(L);
        if (deriv < 0) {
          agn_poptoptwo(L);
          luaL_error(L, "Error in " LUA_QS ": deriv must be non-negative.", "calc.savgol");
        };
      } else {
        agn_poptoptwo(L);
        luaL_error(L, "Error in " LUA_QS ": right-hand side of deriv option must be a number.", "calc.savgol");
      }
    } else if (tools_streq(option, "points")) {  /* 2.16.12 tweak */
       agn_pairgeti(L, nargs, 2);
       if (agn_isnumber(L, -1)) {
         npoints = lua_tointeger(L, -1);
         agn_poptop(L);
         if (npoints < 1 || npoints % 2 == 0) {
           agn_poptoptwo(L);
           luaL_error(L, "Error in " LUA_QS ": deriv must be positive and odd.", "calc.savgol");
         };
       } else {
         agn_poptoptwo(L);
         luaL_error(L, "Error in " LUA_QS ": right-hand side of points option must be a number.", "calc.savgol");
       }
     } else if (tools_streq(option, "eps")) {  /* 2.16.12 tweak */
       agn_pairgeti(L, nargs, 2);
       if (agn_isnumber(L, -1)) {
         eps = agn_tonumber(L, -1);
         agn_poptop(L);
         if (eps <= 0) {
           agn_poptoptwo(L);
           luaL_error(L, "Error in " LUA_QS ": eps must be positive.", "calc.savgol");
         }
       } else {
         agn_poptoptwo(L);
         luaL_error(L, "Error in " LUA_QS ": right-hand side of eps option must be a number.", "calc.savgol");
       }
     } else if (tools_streq(option, "degree")) {  /* 2.16.12 tweak */
       agn_pairgeti(L, nargs, 2);
       if (agn_isnumber(L, -1)) {
         polydeg = lua_tointeger(L, -1);
         agn_poptop(L);
         if (polydeg <= 0) {
           agn_poptoptwo(L); \
           luaL_error(L, "Error in " LUA_QS ": degree must be positive.", "calc.savgol");
         };
       } else {
         agn_poptoptwo(L);
         luaL_error(L, "Error in " LUA_QS ": right-hand side of degree option must be a number.", "calc.savgol");
       }
     } else if (tools_streq(option, "adaptive")) {  /* 2.16.12 tweak */
      agn_pairgeti(L, nargs, 2);
      if (agn_isnumber(L, -1)) {
        isadaptive = lua_toboolean(L, -1);
        agn_poptop(L);
      } else {
        agn_poptoptwo(L);
        luaL_error(L, "Error in " LUA_QS ": right-hand side of adaptive option must be a boolean.", "calc.savgol");
      }
    } else {
      agn_poptop(L);
      luaL_error(L, "Error in " LUA_QS ": unknown option.", "calc.savgol");
    }
    nargs--;
  }
  factor = (isadaptive == 1) ? deriv + 1 : 1;
  eps *= factor;
  nright = npoints/2*factor;
  nleft = -nright;
  if (agnL_gettablefield(L, "calc", "savgolcoeffs", "calc.savgol", 1) != LUA_TFUNCTION)
    luaL_error(L, "Error in " LUA_QS ": could not find `calc.savgolcoeffs`.", "calc.savgol");
  lua_pushinteger(L, -nleft);
  lua_pushinteger(L, nright);
  lua_pushinteger(L, deriv);
  lua_pushinteger(L, polydeg + deriv*isadaptive);
  lua_call(L, 4, 1);
  if (!lua_isreg(L, -1))
    luaL_error(L, "Error in " LUA_QS ": coefficient vector is invalid.", "calc.savgol");
  nops = agn_regsize(L, -1);
  aux_newnumarray(L, nops + 4);  /* pushes NumArray userdata of Savgol coefficient vector onto stack with four ... */
  a = (NumArray *)lua_touserdata(L, -1);  /* ... additional information, see below */
  a->size = nops;
  for (i=0; i < nops; i++) {
    setarray(a, i, agn_reggetinumber(L, -2, i + 1));
  }
  lua_remove(L, -2);
  setarray(a, i++, nleft);
  setarray(a, i++, nright);
  setarray(a, i++, deriv);
  setarray(a, i, eps);
  lua_pushvalue(L, 1);  /* push function */
  lua_pushcclosure(L, &savgol, 2);
  return 1;
}


/* The function receives a non-negative integer n and any number x and returns 0 if x < 0, 1 if x > 1, and smoothly interpolates,
   using an (2*n+1)'th-degree Hermite polynomial, between 0 and 1, otherwise.

   The slope of the smoothstep function is zero at both edges, so the result is differentiable over the whole real domain.

   Wikipedia: `Smoothstep is a family of sigmoid-like interpolation and clamping functions commonly used in computer graphics and
   video game engines`, for example to naturally accelerate or decelerate an object.

   See also: http://sol.gfxile.net/interpolation, https://en.wikipedia.org/wiki/Smoothstep */
static int calc_smoothstep (lua_State *L) {  /* 2.10.4 */
  lua_Number x, r;
  lua_Integer a;
  int n, KenPerlin;
  x = agn_checknumber(L, 1);
  if (agn_isstring(L, 2) && tools_streq(agn_tostring(L, 2), "perlin")) {  /* 2.14.1, 2.16.12 tweak */
    a = 0; KenPerlin = 1;
  } else {
    a = agn_checknonnegint(L, 2); KenPerlin = 0;
  }
  r = 0;
  if (x < 0) r = 0;
  else if (x > 1) r = 1;
  else {
    /* 2.14.1, see: https://en.wikipedia.org/wiki/Smoothstep: "[Kenneth] Perlin suggests an improved version of the smoothstep function,
       which has zero 1st- and 2nd-order derivatives at x = 0 and x = 1" */
    if (KenPerlin) {
      r = x*x*x*(x*(x*6 - 15) + 10);
    } else {
      for (n=0; n < a; n++) {
        /* for formula, see: https://stackoverflow.com/questions/41195063/general-smoothstep-equation */
        r += tools_intpow(-1, n) * tools_intpow(x, a + n) *
             tools_binomial(a + n - 1, n) * tools_binomial(2*a - 1, a - n - 1);
      }
    }
  }
  lua_pushnumber(L, r);
  return 1;
}


/* Computes the sigmoid function, having a characteristic `S`-shaped curve or sigmoid curve, for any number x.
   The result is a number between - but excluding - 0 and 1. See: https://en.wikipedia.org/wiki/Sigmoid_function */
static int calc_sigmoid (lua_State *L) {  /* 2.10.4 */
  lua_Number x, t;
  x = agn_checknumber(L, 1);
  t = sun_exp(x);
  lua_pushnumber(L, t/(t + 1));
  return 1;
}


/* calc.logistic(x [, max [, k [, x0]]])

   Computes the logistic function, having a characteristic `S`-shaped curve or sigmoid curve, for any number x, according
   to the formula

   L(x) = max/(1 + exp(-k*(x - x0))),

   where max is the curve's maximum value, k its steepness, and x0 the x-value of the sigmoid's midpoint. By default,
   max = 1, k = 1, and x0 = 0, computing the sigmoid function. If only x is given, it works like `calc.sigmoid`.

   The result is a number between - but excluding - 0 and max. See: https://en.wikipedia.org/wiki/Logistic_function */
static int calc_logistic (lua_State *L) {  /* 2.10.4 */
  lua_Number x;
  x = agn_checknumber(L, 1);
  if (lua_gettop(L) == 1) {  /* = sigmoid function */
    lua_Number t;
    t = sun_exp(x);
    lua_pushnumber(L, t/(t + 1));
  } else {
    lua_Number x0, l, k;
    l = agnL_optnumber(L, 2, 1);   /* curve's maximum value */
    k = agnL_optnumber(L, 3, 1);   /* steepness of the curve */
    x0 = agnL_optnumber(L, 4, 0);  /* x-value of the sigmoid's midpoint */
    lua_pushnumber(L, l/(1 + sun_exp(-k*(x - x0))));
  }
  return 1;
}


/* Computes the Softsign function f(x) = x/(1 + |x|). 40 percent faster than an Agena implementation */
static int calc_softsign (lua_State *L) {  /* 2.14.1 */
  lua_Number x;
  x = agn_checknumber(L, 1);
  lua_pushnumber(L, x/(1 + fabs(x)));
  return 1;
}


/* Computes the Gaussian function a*exp(-(x-b)^2/(2*c^2)) where at a real or complex point x, with a, b, c
   being (real) numbers. By default, a = 1, b = 0, c = 1/sqrt(2). The return depends on the type of x.

   See also: expx2.

   See: https://en.wikipedia.org/wiki/Gaussian_function; formula computed using pencil, paper and Maple V Release 4. */
static int calc_gaussian (lua_State *L) {  /* 2.10.5 */
  lua_Number a, b, c;
  a = agnL_optnumber(L, 2, 1);
  b = agnL_optnumber(L, 3, 0);
  c = agnL_optnumber(L, 4, M_SQRT1_2);  /* = 1/sqrt(2) = sqrt(2)/2 */
  switch (lua_type(L, 1)) {
    case LUA_TNUMBER: {
      lua_Number x;
      x = agn_tonumber(L, 1);
      lua_pushnumber(L, a*tools_expx2(M_SQRT1_2*(x - b)/c, -1));
      break;
    }
    case LUA_TCOMPLEX: {
      lua_Number x, y, re, im;
#ifndef PROPCMPLX
      agn_Complex z = agn_tocomplex(L, 1);
      x = creal(z);
      y = cimag(z);
#else
      x = agn_complexreal(L, 1); y = agn_compleximag(L, 1);
#endif
      re = tools_cexpx2(0.5*(M_SQRT2*(x - b)/c), M_SQRT1_2*y/c , -1, &im);
#ifndef PROPCMPLX
      agn_createcomplex(L, a*(re + im*I));
#else
      agn_createcomplex(L, a*re, a*im);
#endif
      break;
    }
    default:
      luaL_error(L, "Error in " LUA_QS ": (complex) number expected, got %s.", "calc.gaussian",
        luaL_typename(L, 1));
  }
  return 1;
}


/* Computes the Weierstra function (formula) and returns a number. The precision is given by its fourth argument, eps, which is `Eps` by default. */
static int calc_weier (lua_State *L) {  /* 2.12.0 RC 4 */
  volatile lua_Number a, b, x, r, eps, z, q;
  int i;
  x = agn_checknumber(L, 1);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  eps = agnL_optpositive(L, 4, agn_getepsilon(L));  /* 2.13.0 change */
  if (luai_numiseven(b) || !(0 < a && a < 1)) {  /* b even or a out-of-range? -> infinite loops otherwise */
    lua_pushundefined(L);
    return 1;
  }
  i = r = q = 0;
  while (1) {
    z = tools_intpow(a, i) * cos(tools_intpow(b, i) * x);
    if (fabs(z) < eps) break;
    r = tools_koadd(r, z, &q); i++; /* 2.13.0 change */
  }
  lua_pushnumber(L, r + q);
  return 1;
}


/* Computes the first derivative of the univariate or multivariate function f at real point x, a number.

   If the option eps=h is given, the epsilon value h (a positive number preferably close to zero) is used
   internally for the computation, its default is math.epsilon(x).

   The second, etc. arguments to f may be given right after argument x.

   The return is the imaginary part of f(x + I*h)/h , or `fail` if f did not evaluate to the complex plane.

   This function as at least three times faster than `calc.xpdiff`.

   The idea has been taken from the Euler Math Toolbox, thus its name. See also:
   www.euler-math-toolbox.de by Rene Grothmann, version 2018-11-16, source file util/numerical.e. */
static int calc_eulerdiff (lua_State *L) {  /* 2.14.3 */
  lua_Number x, eps, delta;
  int sided, deriv, i, c, nargs;
#ifndef PROPCMPLX
  agn_Complex z;
#else
  lua_Number im;
#endif
  const char *option = NULL;
  (void)sided; (void)deriv; (void)delta;
  c = 1;
  luaL_checktype(L, 1, LUA_TFUNCTION);
  x = agn_checknumber(L, 2);
  nargs = lua_gettop(L);
  aux_geteps(L, x, nargs, "calc.eulerdiff");
  /* first check whether f is defined at x, ..., otherwise the function would return a finite result. */
  if (isnan(agnL_fncall(L, 1, x, 3, nargs, 1))) {
    lua_pushundefined(L);
    return 1;
  }
  /* now determine derivative << f, x, h -> imag(f(x + I*h)) / h >> */
  lua_pushvalue(L, 1);
#ifndef PROPCMPLX
  agn_createcomplex(L, x + I*eps);
#else
  agn_createcomplex(L, x, eps);
#endif
  for (i=3; i <= nargs; i++, c++) lua_pushvalue(L, i);
  if (lua_pcall(L, c, 1, 0) != 0 || !lua_iscomplex(L, -1)) {
    agn_poptop(L);  /* remove error message string */
    lua_pushfail(L);
    return 1;
  }
#ifndef PROPCMPLX
  z = agn_tocomplex(L, -1);
  agn_poptop(L);
  lua_pushnumber(L, cimag(z)/eps);
#else
  im = agn_compleximag(L, -1);
  agn_poptop(L);
  lua_pushnumber(L, im/eps);
#endif
  return 1;
}


static const luaL_Reg calclib[] = {
  {"Ai", calc_Ai},                         /* added on June 29, 2015 */
  {"Bi", calc_Bi},                         /* added on June 29, 2015 */
  {"cheby", calc_cheby},                   /* added on July 05, 2017 */
  {"Chi", calc_Chi},                       /* added on May 24, 2010 */
  {"Ci", calc_Ci},                         /* added on May 24, 2010 */
  {"clampedspline", calc_clampedspline},   /* added on November 04, 2012 */
  {"clampedsplinecoeffs", calc_clampedsplinecoeffs},  /* added on November 04, 2012 */
  {"dawson", calc_dawson},                 /* added on May 25, 2010 */
  {"diff", calc_diff},                     /* added on May 29, 2010 */
  {"dilog", calc_dilog},                   /* added on May 25, 2010 */
  {"Ei", calc_Ei},                         /* added on May 25, 2010 */
  {"En", calc_En},                         /* added on July 07, 2015 */
  {"eps", calc_eps},                       /* added on February 14, 2016 */
  {"eucliddist", calc_eucliddist},         /* added on January 25, 2016 */
  {"eulerdiff", calc_eulerdiff},           /* added on December 26, 2018 */
  {"fminbr", calc_fminbr},                 /* added on May 29, 2010 */
  {"fprod", calc_fprod},                   /* added on Nov 23, 2009 */
  {"fresnelc", calc_fresnelc},             /* added on May 29, 2010 */
  {"fresnels", calc_fresnels},             /* added on May 29, 2010 */
  {"fsum", calc_fsum},                     /* added on April 20, 2008 */
  {"gaussian", calc_gaussian},             /* added on August 06, 2017 */
  {"gtrap", calc_gtrap},                   /* added on March 04, 2014 */
  {"ibeta", calc_ibeta},                   /* added on July 08, 2015 */
  {"igamma", calc_igamma},                 /* added on July 07, 2015 */
  {"igammac", calc_igammac},               /* added on July 07, 2015 */
  {"intde", calc_intde},                   /* added on May 18, 2010 */
  {"intdei", calc_intdei},                 /* added on May 22, 2010 */
  {"intdeo", calc_intdeo},                 /* added on May 22, 2010 */
  {"interp", calc_interp},                 /* added on October 28, 2012 */
  {"invibeta", calc_invibeta},             /* added on July 08, 2015 */
  {"iscont", calc_iscont},                 /* added on January 21, 2016 */
  {"isdiff", calc_isdiff},                 /* added on May 04, 2017 */
  {"limit", calc_limit},                   /* added on January 21, 2016 */
  {"linterp", calc_linterp},               /* added on January 28, 2014 */
  {"logistic", calc_logistic},             /* added on August 04, 2017 */
  {"nakspline", calc_nakspline},           /* added on October 28, 2012 */
  {"naksplinecoeffs", calc_naksplinecoeffs}, /* added on October 28, 2012 */
  {"neville", calc_neville},               /* added on October 28, 2012 */
  {"newtoncoeffs", calc_newtoncoeffs},     /* added on October 28, 2012 */
  {"polyfit", calc_polyfit},               /* added on January 28, 2014 */
  {"polygen", calc_polygen},               /* added on July 21, 2013 */
  {"polylog", calc_polylog},               /* added on July 07, 2015 */
  {"Psi", calc_Psi},                       /* added on May 25, 2010 */
  {"regulafalsi", calc_regulafalsi},       /* added on May 29, 2010 */
  {"savgol", calc_savgol},                 /* added on July 16, 2017 */
  {"sections", calc_sections},             /* added on May 29, 2010 */
  {"Shi", calc_Shi},                       /* added on May 24, 2010 */
  {"Si", calc_Si},                         /* added on May 24, 2010 */
  {"sigmoid", calc_sigmoid},               /* added on March 04, 2014 */
  {"simaptive", calc_simaptive},           /* added on March 04, 2014 */
  {"smoothstep", calc_smoothstep},         /* added on August 02, 2017 */
  {"softsign", calc_softsign},             /* added on October 18, 2018 */
  {"Ssi", calc_Ssi},                       /* added on May 24, 2010 */
  {"weier", calc_weier},                   /* added on June 16, 2018 */
  {"xpdiff", calc_xpdiff},                 /* added on May 29, 2010 */
  {"zeta", calc_zeta},                     /* added on July 07, 2015 */
  {NULL, NULL}
};


/*
** Open calculus library
*/

LUALIB_API int luaopen_calc (lua_State *L) {
  luaL_newmetatable(L, "calc");
  luaL_register(L, NULL, numarray_arraylib);  /* associate __gc method */
  luaL_register(L, AGENA_CALCLIBNAME, calclib);
  return 1;
}

