/* Character Buffer, initiated January 04, 2020

   Call `charbuf_finish` if you want to use the buffered string and do not know its size, e.g. when using lua_pushstring, after tha last 
   append to include a terminating '\0'. 
   
   Note that packages, luaL_Buffer & Co are faster than the functions defined here since luaL_Buffer uses the stack and no mallocs/reallocs. */

#include <stdlib.h>
#include <string.h>

#include "agena.h"
#include "ldebug.h"
#include "agnhlps.h"

#define charbuf_c
#define LUA_LIB

#include "charbuf.h"

void INLINE charbuf_init (lua_State *L, Charbuf *buf, size_t initsize) {  /* refactored, 2.16.12 */
  size_t optlen, chunks;
  optlen = tools_optstrlen(initsize, &chunks);
  buf->data = (char *)malloc(optlen*sizeof(char));
  if (buf->data == NULL)
    luaG_runerror(L, "Error: buffer allocation error.");
  memset(buf->data + initsize, 0, optlen - initsize);  /* only zero the bytes that will not be filled by characters later on */
  buf->capacity = optlen;
  buf->size = 0;
}


void INLINE charbuf_append (lua_State *L, Charbuf *buf, const char *str, size_t n) {  /* refactored, 2.16.12, rewritten 2.16.13 */
  char *temp;
  size_t chunks;
  if (n == 0) return;
  if (buf->size + n + 1 > buf->capacity) {  /* Agena 1.0.6 fix: +1 for trailing '\0' that is inserted later, otherwise Agena might crash */
    buf->capacity = tools_optstrlen(buf->capacity*2 + n, &chunks);  /* Agena 1.0.6 change to speed up computation */
    temp = (char *)realloc(buf->data, buf->capacity*sizeof(char));
    if (temp == NULL) {
      xfree(buf->data);
      luaG_runerror(L, "Error: buffer re-allocation error.");
    } else
      buf->data = temp;
  }
  /* special treatment of one-char strings is not necessary as memcpy is faster than a single char assignment. */
  memcpy(buf->data + buf->size, str, n);  /* much faster than strncpy, as fast as `for (i=0; i < n; i++) *buffer++ = *string++` */
  buf->size += n;
  return;
}


void INLINE charbuf_appendchar (lua_State *L, Charbuf *buf, int chr) {  /* added 2.16.13 */
  char *temp;
  size_t chunks;
  if (buf->size + 2 > buf->capacity) {  /* Agena 1.0.6 fix: +1 for trailing '\0' that is inserted later, otherwise Agena might crash */
    buf->capacity = tools_optstrlen(buf->capacity*2 + 1, &chunks);  /* Agena 1.0.6 change to speed up computation */
    temp = (char *)realloc(buf->data, buf->capacity*sizeof(char));
    if (temp == NULL) {
      xfree(buf->data);
      luaG_runerror(L, "Error: buffer re-allocation error.");
    } else
      buf->data = temp;
  }
  buf->data[buf->size] = chr;
  buf->size++;
  return;
}
