/*
** $Id: lmathlib.c,v 1.67 2005/08/26 17:36:32 roberto Exp $
** Standard mathematical library
** See Copyright Notice in agena.h
*/

#include <stdlib.h>
#include <math.h>
#include <stdint.h>  /* for UINT32_MAX */
#include <string.h>

#define lmathlib_c
#define LUA_LIB
#define LUA_CORE  /* just to import agnconf.h's luai_num* macros */

#include "agena.h"

#include "agnxlib.h"
#include "agenalib.h"
#include "agnconf.h"
#include "agnt64.h"

#include "agncmpt.h"  /* for trunc and isfinite, and FLT_EVAL_METHOD constant, and off64* types */
#include "lobject.h"
#include "cephes.h"

/* For unknown reasons I am too tired to clarify, we import the *GET_*, *SET_*, *INSERT*, *EXTRACT*,
   etc. Sun Microsystems macros from the following header file instead of agnhlps.h: */
#include "sunpro.h"


static int math_min (lua_State *L) {  /* changed 2.14.4; with just two arguments, fmin is slower */
  int i, n;
  lua_Number d, dmin;
  n = lua_gettop(L);  /* number of arguments */
  dmin = agn_checknumber(L, 1);
  for (i=2; i <= n; i++) {
    d = agn_checknumber(L, i);
    if (d < dmin) dmin = d;
  }
  lua_pushnumber(L, dmin);
  return 1;
}


static int math_max (lua_State *L) {  /* changed 2.14.4; with just two arguments, fmax is slower */
  int i, n;
  lua_Number d, dmax;
  n = lua_gettop(L);  /* number of arguments */
  dmax = agn_checknumber(L, 1);
  for (i=2; i <= n; i++) {
    d = agn_checknumber(L, i);
    if (d > dmax) dmax = d;
  }
  lua_pushnumber(L, dmax);
  return 1;
}


static int math_random (lua_State *L) {
  /* RNG algorithm taken from https://en.wikipedia.org/w/index.php?title=Random_number_generation&diff=778577551&oldid=612091085,
     written by George Marsaglia. 0.32.2, 22.05.2010 */
  int nargs, mode;
  nargs = lua_gettop(L);
  mode = (nargs > 0 && lua_type(L, nargs) == LUA_TBOOLEAN);  /* 2.8.3 fix */
  switch (nargs - mode) {  /* check number of arguments */
    case 0: {  /* no arguments */
      lua_pushnumber(L, tools_random(mode));  /* Number between 0 and 1 */
      break;
    }
    case 1: {  /* only upper limit given */
      lua_Number u;
      u = agn_checknumber(L, 1);  /* 2.10.1 fix */
      if (u < 1)  /* 2.10.1 improvement */
        luaL_error(L, "Error in " LUA_QS ": border must be positive.", "math.random");
      lua_pushnumber(L, tools_randomrange(1, u, mode));  /* 2.14.3 fix, avoid low frequency */
      break;
    }
    case 2: {  /* lower and upper limits given */
      lua_Number l, u;
      l = agn_checknumber(L, 1);  /* 2.10.1 fix */
      u = agn_checknumber(L, 2);  /* 2.10.1 fix */
      if (u < l)  /* 2.10.1 improvement */
        luaL_error(L, "Error in " LUA_QS ": upper border is less than lower border.", "math.random");
      lua_pushnumber(L, tools_randomrange(l, u, mode));  /* 2.14.3 fix, avoid low frequency */
      break;
    }
    default:
      return luaL_error(L, "Error in " LUA_QS ": wrong arguments given.", "math.random");  /* 2.10.1, 2.14.6 changed */
  }
  return 1;
}


LUAI_UINT32 m_w = AGN_RANDOM_MW; /* must not be zero, nor 0x464fffff; for definition, see agnconf.h */
LUAI_UINT32 m_z = AGN_RANDOM_MZ; /* must not be zero, nor 0x9068ffff; for definition, see agnconf.h */

static int math_randomseed (lua_State *L) {  /* changed 0.32.2, 23.05.2010, extended 2.10.1 */
  if (lua_gettop(L) != 0) {
    lua_Number w, z;
    w = agn_checkposint(L, 1);
    z = agn_checkposint(L, 2);
    /* do not check for w = 0x464fffff & z = 0x9068ffff, as being warned by George Marsaglia;
       set seeds to global C domain ... */
    m_w = (LUAI_UINT32)w;
    m_z = (LUAI_UINT32)z;
    /* ... and return them */
  }
  /* if you change the number of results, CHANGE environ_restart, as well ! */
  lua_pushnumber(L, m_w);  /* do NOT use lua_pushinteger as there might be overflows, 2.15.1 fix */
  lua_pushnumber(L, m_z);  /* do NOT use lua_pushinteger as there might be overflows, 2.15.1 fix */
  return 2;
}


static int math_isprime (lua_State *L) {  /* patched 0.27.0; tuned 0.30.3; extended 1.0.6 */
  long long int x;
  lua_Number n;
  n = agn_checkinteger(L, 1);  /* 2.13.0a */
  if (n > 9223372036854775807.0 || n < -9223372036854775808.0)  /* max/min of type LONG LONG INT; 2.14.6 simplification */
    luaL_error(L, "Error in " LUA_QS ": number too big, too small or not an integer.", "math.isprime");
  x = (long long int)n;
  if (x < 4)
    lua_pushboolean(L, x > 1);
  else if (x % 2 == 0)  /* even number ? */
    lua_pushboolean(L, 0);
  else {
    long long int i;
    const long long int imax = (long long int)sqrt(x) + 1;
    for (i = 3; i <= imax; i += 2) {
      if (x % i == 0) {
        lua_pushfalse(L);
        return 1;
      }
    }
    lua_pushtrue(L);
  }
  return 1;
}


static int math_prevprime (lua_State *L) {  /* 0.33.2, fixed and extended 1.0.6 */
  long long int x;
  lua_Number n;
  n = agn_checkinteger(L, 1);  /* 2.13.0 */
  if (n > 9223372036854775807.0 || n < -9223372036854775808.0)  /* max/min of type LONG LONG INT */
    luaL_error(L, "Error in " LUA_QS ": number too big or too small.", "math.prevprime");
  x = (long long int)n;
  if (x < 3)
    lua_pushfail(L);
  else if (x == 3)
    lua_pushinteger(L, 2);
  else if (x < 6)
    lua_pushinteger(L, 3);
  else {
    unsigned long long int n, b, i;
    b = (x % 2 == 0) ? x - 1 : x - 2;
    for (i=b; ; i = i - 2) {
      n = 3;
      while (n * n < i && i % n != 0)
        n += 2;
      if (i % n != 0) break;
    }
    lua_pushnumber(L, i);  /* Agena 1.0.6 */
  }
  return 1;
}


static int math_nextprime (lua_State *L) {  /* 0.33.2, extended 1.0.6 */
  long long int x;
  lua_Number n;
  n = agn_checkinteger(L, 1);  /* 2.13.0 */
  if (n > 9223372036854775807.0 || n < -9223372036854775808.0)  /* max/min of type LONG LONG INT */
    luaL_error(L, "Error in " LUA_QS ": number too big or too small.", "math.nextprime");
  x = (long long int)n;
  if (x < 2)
    lua_pushinteger(L, 2);
  else if (x < 3)  /* patched Agena 1.0.6 */
    lua_pushinteger(L, 3);
  else {
    unsigned long long int n, b, i;
    b = (x % 2 == 0) ? x + 1 : x + 2;
    for (i=b; ; i = i + 2) {
      n = 3;
      while (n * n < i && i % n != 0)
        n += 2;
      if (i % n != 0) break;
    }
    lua_pushnumber(L, i);  /* Agena 1.0.6 */
  }
  return 1;
}


/* constant found in Cephes Math Library Release 2.8, June, 2000 by Stephen L. Moshier
   file sin.c (and fixed it) */
#define P64800   (4.8481368110953599358991410e-6)  /* = (Pi/180)/3600 */

/* converts degrees to radians */
static int math_toradians (lua_State *L) {
  lua_Number m, s;
  m = agnL_optnumber(L, 2, 0);
  s = agnL_optnumber(L, 3, 0);
  lua_pushnumber(L,
    ((agn_checknumber(L, 1)*60.0 + m)*60.0 + s)*P64800
  );
  return 1;
}


/* 2.8.4, see http://stackoverflow.com/questions/13975745/the-fastest-way-to-get-current-quadrant-of-an-angle,
   solution #2 posted by Vincent Mimoun-Prat.
   The function returns the quadrant of an angle given in radians and returns an integer in [1, 4]. */
static int math_quadrant (lua_State *L) {
  lua_Number angle = fmod(agn_checknumber(L, 1)/RADIANS_PER_DEGREE, 360.0);  /* [0..360) if angle is positive, (-360..0] if negative */
  if (angle < 0) angle += 360.0;  /* back to [0..360) */
  lua_pushnumber(L, fmod(trunc(angle/90), 4) + 1);  /* the quadrant, 2.11.0 tuning */
  return 1;
}


/* converts sexagesimal, passed as a triple, to decimal */
static int math_todecimal (lua_State *L) {
  lua_Number h, m, s, sgn;
  h = agn_checknumber(L, 1);
  m = agnL_optnumber(L, 2, 0);
  s = agnL_optnumber(L, 3, 0);
  if (h < 0) {  /* 2.9.8 improvement */
    sgn = -1;   /* 2.2.0 patch */
    h = -h;
  } else
    sgn = 1;
  lua_pushnumber(L, sgn*(h + m/60 + s/3600));
  return 1;
}


/* Converts a number representing a decimal number x into its sexagesimal representation consiting of three numbers:
   degrees, minutes, and seconds. For example: 10.5125 returns 10, 30, 45. */
static int math_tosgesim (lua_State *L) {   /* based on math.dms, 2.9.8, replacing the former - inprecise - Agena 1.12.7 version */
  long long int deg, min, sec, sign;  /* do not use int or long int due to possible overflows */
  lua_Number x = agn_checknumber(L, 1);
  if (x < 0) {
    x = -x;
    sign = -1;
  } else
    sign = 1;
  sec = sun_floor(3600.0*x + 0.5);
  deg = sec / 3600;
  sec = fabs(fmod(sec, 3600.0));
  min = sec / 60;
  sec = fmod(sec, 60);
  lua_pushnumber(L, sign*deg);
  lua_pushnumber(L, min);
  lua_pushnumber(L, sec);
  return 3;
}


/* Converts a number representing a decimal number x into its TI-30 sexagesimal DMS representation and returns a number.
   For example: 10.5125 returns 10.3045, representing 1030'45''.

   Based on http://stackoverflow.com/questions/6862684/converting-from-decimal-degrees-to-degrees-minutes-seconds-tenths,
   solution #4 proposed by user807566.

   Do not use the formula math.dms := proc(x :: number) is local f := frac(x)*60; return int(x) + int(f)/100 + frac(f)/100 end;
   or any version also detecting the sign of x for it is prone to slight rounding errors in loops with fractional step sizes. */
static int math_dms (lua_State *L) {   /* rewritten in C, 2.8.3, 25 percent faster than the former - inprecise - Agena version */
  long long int deg, min, sec, sign;  /* do not use int or long int due to possible overflows */
  lua_Number x = agn_checknumber(L, 1);
  if (x < 0) {
    x = -x;
    sign = -1;
  } else
    sign = 1;
  sec = sun_floor(3600.0*x + 0.5);
  deg = sec / 3600;
  sec = fabs(fmod(sec, 3600.0));
  min = sec / 60;
  sec = fmod(sec, 60);
  lua_pushnumber(L, sign*((lua_Number)deg + (lua_Number)min/100.0 + (lua_Number)sec/10000.0));
  return 1;
}


/* Converts a non-negative integer in the range [0, 255] to its hexadecimal representation, returned as a 2-character string. */
static int math_tohex (lua_State *L) {  /* 2.11.3 */
  int h;
  const char *hex;
  char s[2 + 1];
  hex = "0123456789ABCDEF";
  h = agn_checknonnegint(L, 1);
  if (h > 255) h %= 256;
  s[0] = hex[(h >> 4) & 0xF];
  s[1] = hex[(h) & 0xF];
  s[2] = '\0';
  lua_pushstring(L, s);
  return 1;
}


static int math_nextafter (lua_State *L) {
  lua_Number dir;  /* 2.10.0 improvement */
  dir = agnL_optnumber(L, 2, HUGE_VAL);
  lua_pushnumber(L, sun_nextafter(agn_checknumber(L, 1), dir));
  return 1;
}


/* Converts the number x in the scale [a1, a2] to one in the scale [b1, b2]. The second and third arguments
   must be pairs of numbers. If the third argument is missing, then x is converted to a number in [0, 1].
   The return is a number. */
static int math_norm (lua_State *L) {
  lua_Number x, a1, a2, b1, b2;
  int t3;
  x = agn_checknumber(L, 1);
  agn_pairgetinumbers(L, "math.norm", 2, &a1, &a2);  /* 2.9.8; 2.10.1 does not delete pair from arg list */
  if (x < a1 || x > a2)
    luaL_error(L, "Error in " LUA_QS ": first argument out of range.", "math.norm");
  if (a1 > a2)  /* 2.2.1 */
    luaL_error(L, "Error in " LUA_QS ": left border greater than right border in second argument.", "math.norm");
  t3 = lua_type(L, 3);
  if (t3 == LUA_TNONE) {
    b1 = 0; b2 = 1;
  }
  else {  /* 2.10.1 */
    agn_pairgetinumbers(L, "math.norm", 3, &b1, &b2);  /* 2.9.8 */
    if (b1 > b2)  /* 2.2.1 */
      luaL_error(L, "Error in " LUA_QS ": left border greater than right border in third argument.", "math.norm");
  }
  lua_pushnumber(L, b1 + (x - a1) * (b2 - b1)/(a2 - a1));  /* 2.2.1 fix */
  return 1;
}


/* Returns the largest power of two less than or equal to x, where x is a non-negative integer. If x <= 1, the result is x;
   if x < 0, undefined is returned. See also: math.ceilpow2. */
static int math_floorpow2 (lua_State *L) {  /* 2.16.0 */
  lua_Number x;
  x = agn_checknumber(L, 1);  /* 2.16.0 */
  if (unlikely(x < 0))  /* 2.16.0 optimisation */
    lua_pushundefined(L);
  else if (unlikely(sun_exponent(x) > 30))  /* return fail with x >= 2^31, 2.16.0 optimisation */
    lua_pushfail(L);
  else {
    unsigned int v = x;
    v -= (v & (v - 1)) != 0;  /* if v is not a power of 2, decrement v by 1 */
    v |= v >> 1;
    v |= v >> 2;
    v |= v >> 4;
    v |= v >> 8;
    v |= v >> 16;
    v &= ~(v >> 1);
    lua_pushinteger(L, v);
  }
  return 1;
}


/* Returns smallest exponent to 2 equals or greater than x, i.e. ilog2(x - 1) + 1 */
static int math_ceillog2 (lua_State *L) {  /* 2.3.1 */
  int x = agn_checkinteger(L, 1);
  if (unlikely(x < 1))  /* 2.3.3 improvement, 2.16.0 optimisation */
    lua_pushundefined(L);
  else
    lua_pushinteger(L, (luaO_log2(x - 1) + 1));
  return 1;
}


/* Finds the smallest power of two greater than or equal to x, taken from:
   http://graphics.stanford.edu/~seander/bithacks.html#RoundUpPowerOf2 */
static int math_ceilpow2 (lua_State *L) {  /* 2.3.3 */
  lua_Number x;
  x = agn_checknumber(L, 1);  /* 2.16.0 */
  if (unlikely(x < 0))  /* 2.16.0 optimisation */
    lua_pushundefined(L);
  else if (unlikely(sun_exponent(x) > 30))  /* return fail with x >= 2^31, 2.16.0 optimisation */
    lua_pushfail(L);
  else {
    unsigned int v = x;
    v--;
    v |= v >> 1;
    v |= v >> 2;
    v |= v >> 4;
    v |= v >> 8;
    v |= v >> 16;
    v++;
    v += (v == 0);
    lua_pushnumber(L, v);
  }
  return 1;
}


/* Taken from: http://graphics.stanford.edu/~seander/bithacks.html#InterleaveTableLookup

Interleaves the bits of integers x and y, so that all of the bits of x are in the even
positions and y in the odd; the function can be used to linearising 2D integer coordinates,
combining x and y into a single integer that can be compared easily has the property that
a number is usually close to another if their x and y values are close. */
static const unsigned short MortonTable256[256] = {
  0x0000, 0x0001, 0x0004, 0x0005, 0x0010, 0x0011, 0x0014, 0x0015,
  0x0040, 0x0041, 0x0044, 0x0045, 0x0050, 0x0051, 0x0054, 0x0055,
  0x0100, 0x0101, 0x0104, 0x0105, 0x0110, 0x0111, 0x0114, 0x0115,
  0x0140, 0x0141, 0x0144, 0x0145, 0x0150, 0x0151, 0x0154, 0x0155,
  0x0400, 0x0401, 0x0404, 0x0405, 0x0410, 0x0411, 0x0414, 0x0415,
  0x0440, 0x0441, 0x0444, 0x0445, 0x0450, 0x0451, 0x0454, 0x0455,
  0x0500, 0x0501, 0x0504, 0x0505, 0x0510, 0x0511, 0x0514, 0x0515,
  0x0540, 0x0541, 0x0544, 0x0545, 0x0550, 0x0551, 0x0554, 0x0555,
  0x1000, 0x1001, 0x1004, 0x1005, 0x1010, 0x1011, 0x1014, 0x1015,
  0x1040, 0x1041, 0x1044, 0x1045, 0x1050, 0x1051, 0x1054, 0x1055,
  0x1100, 0x1101, 0x1104, 0x1105, 0x1110, 0x1111, 0x1114, 0x1115,
  0x1140, 0x1141, 0x1144, 0x1145, 0x1150, 0x1151, 0x1154, 0x1155,
  0x1400, 0x1401, 0x1404, 0x1405, 0x1410, 0x1411, 0x1414, 0x1415,
  0x1440, 0x1441, 0x1444, 0x1445, 0x1450, 0x1451, 0x1454, 0x1455,
  0x1500, 0x1501, 0x1504, 0x1505, 0x1510, 0x1511, 0x1514, 0x1515,
  0x1540, 0x1541, 0x1544, 0x1545, 0x1550, 0x1551, 0x1554, 0x1555,
  0x4000, 0x4001, 0x4004, 0x4005, 0x4010, 0x4011, 0x4014, 0x4015,
  0x4040, 0x4041, 0x4044, 0x4045, 0x4050, 0x4051, 0x4054, 0x4055,
  0x4100, 0x4101, 0x4104, 0x4105, 0x4110, 0x4111, 0x4114, 0x4115,
  0x4140, 0x4141, 0x4144, 0x4145, 0x4150, 0x4151, 0x4154, 0x4155,
  0x4400, 0x4401, 0x4404, 0x4405, 0x4410, 0x4411, 0x4414, 0x4415,
  0x4440, 0x4441, 0x4444, 0x4445, 0x4450, 0x4451, 0x4454, 0x4455,
  0x4500, 0x4501, 0x4504, 0x4505, 0x4510, 0x4511, 0x4514, 0x4515,
  0x4540, 0x4541, 0x4544, 0x4545, 0x4550, 0x4551, 0x4554, 0x4555,
  0x5000, 0x5001, 0x5004, 0x5005, 0x5010, 0x5011, 0x5014, 0x5015,
  0x5040, 0x5041, 0x5044, 0x5045, 0x5050, 0x5051, 0x5054, 0x5055,
  0x5100, 0x5101, 0x5104, 0x5105, 0x5110, 0x5111, 0x5114, 0x5115,
  0x5140, 0x5141, 0x5144, 0x5145, 0x5150, 0x5151, 0x5154, 0x5155,
  0x5400, 0x5401, 0x5404, 0x5405, 0x5410, 0x5411, 0x5414, 0x5415,
  0x5440, 0x5441, 0x5444, 0x5445, 0x5450, 0x5451, 0x5454, 0x5455,
  0x5500, 0x5501, 0x5504, 0x5505, 0x5510, 0x5511, 0x5514, 0x5515,
  0x5540, 0x5541, 0x5544, 0x5545, 0x5550, 0x5551, 0x5554, 0x5555
};

static int math_morton (lua_State *L) {  /* 2.3.3 */
  unsigned short x, y;   /* Interleave bits of x and y, so that all of the
    bits of x are in the even positions and y in the odd; unsigned int z = 0;
    z gets the resulting Morton Number. */
  unsigned int z;
  x = agn_checknumber(L, 1);
  y = agn_checknumber(L, 2);
  z = MortonTable256[y >> 8] << 17 |
    MortonTable256[x >> 8]   << 16 |
    MortonTable256[y & 0xFF] <<  1 |
    MortonTable256[x & 0xFF];
  lua_pushnumber(L, z);
  return 1;
}


/* Returns a value equivalent to exp(x) - 1, with x a number. It is computed in a way that is accurate even
   if x is near 0, since exp(~0) and 1 are nearly equal. See also: math.lnplusone.

From: http://en.cppreference.com/w/cpp/numeric/math/expm1

   Parameters
arg 	- 	value of floating-point or Integral type
Return value

If no errors occur earg
-1 is returned.

If a range error due to overflow occurs, +HUGE_VAL, +HUGE_VALF, or +HUGE_VALL is returned.

If a range error occurs due to underflow, the correct result (after rounding) is returned.
Error handling

Errors are reported as specified in math_errhandling

If the implementation supports IEEE floating-point arithmetic (IEC 60559),

    If the argument is 0, it is returned, unmodified
    If the argument is -?, -1 is returned
    If the argument is +?, +? is returned
    If the argument is NaN, NaN is returned

Notes

The functions std::expm1 and std::log1p are useful for financial calculations, for example, when calculating small daily interest rates: (1+x)n
-1 can be expressed as std::expm1(n * std::log1p(x)). These functions also simplify writing accurate inverse hyperbolic functions.

For IEEE-compatible type double, overflow is guaranteed if 709.8 < arg */
static int math_expminusone (lua_State *L) {  /* 2.3.3 */
  lua_pushnumber(L, sun_expm1(agn_checknumber(L, 1)));  /* 2.11.1 tuning */
  return 1;
}


/* Returns a value equivalent to ln(1 + x), with x a number. It is computed in a way that is accurate even
   if x is near zero.

   Example: ln(1.0000000000000001) -> 0, math.lnplusone(0.0000000000000001)
   -> 1e-016. See also: math.expminusone.

From: http://en.cppreference.com/w/cpp/numeric/math/log1p

If no errors occur ln(1+arg) is returned.

If a domain error occurs, an implementation-defined value is returned (NaN where supported)

If a pole error occurs, -HUGE_VAL, -HUGE_VALF, or -HUGE_VALL is returned.

If a range error occurs due to underflow, the correct result (after rounding) is returned.
Error handling

Errors are reported as specified in math_errhandling

Domain error occurs if arg is less than -1.

Pole error may occur if arg is -1.

If the implementation supports IEEE floating-point arithmetic (IEC 60559),

    If the argument is 0, it is returned unmodified
    If the argument is -1, -infinity is returned and FE_DIVBYZERO is raised.
    If the argument is less than -1, NaN is returned and FE_INVALID is raised.
    If the argument is +infinity, +infinity is returned
    If the argument is NaN, NaN is returned

Some Notes:

The functions std::expm1 and std::log1p are useful for financial calculations, for example, when calculating small daily interest rates: (1+x)n
-1 can be expressed as std::expm1(n * std::log1p(x)). These functions also simplify writing accurate inverse hyperbolic functions. */
static int math_lnplusone (lua_State *L) {  /* 2.3.3 */
  lua_Number x = agn_checknumber(L, 1);
  if (x <= -1)
    lua_pushundefined(L);
  else
    lua_pushnumber(L, log1p(x));
  return 1;
}


/* Returns `false` if at least one of its arguments is `undefined`, and `true` otherwise. */

static int math_isordered (lua_State *L) {  /* 2.3.3 */
  lua_pushboolean(L, !isunordered(agn_checknumber(L, 1), agn_checknumber(L, 2)));
  return 1;
}


/* Returns a number with the magnitude of x and the sign of y. It is a plain binding to
   C's copysign function and does not postprocess its result. */
static int math_copysign (lua_State *L) {  /* 2.3.3 */
  lua_pushnumber(L, copysign(agn_checknumber(L, 1), agn_checknumber(L, 2)));
  return 1;
}


static int math_signbit (lua_State *L) {  /* 2.4.2 */
  lua_pushboolean(L, signbit(agn_checknumber(L, 1)) != 0);
  return 1;
}


/* Returns the inverse hyperbolic sine of the number x and returns a number. It works
   in the real domain only. */
static int math_arcsinh (lua_State *L) {  /* 2.14.10 */
  lua_pushnumber(L, sun_asinh(agn_checknumber(L, 1)));  /* 2.14.13, 12 % faster */
  return 1;
}


/* Returns the inverse hyperbolic cosine of the number x and returns a number. It works
   in the real domain only. sun_acosh is a lot slower. */
static int math_arccosh (lua_State *L) {  /* 2.3.3 */
  lua_pushnumber(L, acosh(agn_checknumber(L, 1)));  /* log(x + luai_numsign(x) * sqrt(x*x - 1))) */
  return 1;
}


/* Computes the inverse hyperbolic tangent of the number x and returns a number. It works
   in the real domain only. */
static int math_arctanh (lua_State *L) {  /* 2.14.10 */
  lua_pushnumber(L, atanh(agn_checknumber(L, 1)));
  return 1;
}


/* The function adds x and y using Kahan-Ozawa round-off error prevention and returns two numbers: the sum of x and y plus the updated value of the correction variable. The correction variable q should be 0 at first invocation.
A typical usage should look like:

x, q -> 0;
y := 0.1;
while x < 1 do
   x, q := math.koadd(x, y, q)
od;
print(s, q);

*/
static int math_koadd (lua_State *L) {  /* 2.4.1, Kahan-Ozawa summation, gets argument x and correction value q, return new x and q */
  volatile lua_Number q, s, sold, u, v, w, x, t;
  s = agn_checknumber(L, 1);    /* cumulatative sum */
  x = agn_checknumber(L, 2);    /* value to be added to sum */
  q = agnL_optnumber(L, 3, 0);  /* correction value */
  v = x - q;
  sold = s;
  s += v;
  if (fabs(x) < fabs(q)) {
    t = x;
    x = -q;
    q = t;
  }
  u = (v - x) + q;
  if (fabs(sold) < fabs(v)) {
    t = sold;
    sold = v;
    v = t;
  }
  w = (s - sold) - v;
  q = u + w;
  /* s now contains sumdata */
  lua_pushnumber(L, s);
  lua_pushnumber(L, q);
  return 2;
}


/* Emulates Matlab's `eps` function and returns the relative spacing between |x| and its next larger
   number in the machines floating point system. On x86 machines and with C doubles,
   eps() and eps(1) return 2.2204460492503e-016 = 2 ^(-52), and eps(2) returns 4.4408920985006e-016
   = 2 ^(-51). 2.4.2, extended 2.9.7 to compute a mathematical epsilon that takes into account the
   magnitude of its first argument. */
static int math_eps (lua_State *L) {  /* 2.4.2, five times faster than an Agena implementation */
  lua_Number x = fabs(luaL_optnumber(L, 1, 1));
  lua_pushnumber(L, (lua_gettop(L) == 1) ? sun_nextafter(x, HUGE_VAL) - x : tools_matheps(x));  /* 2.9.7 */
  return 1;
}


/* The following math_precision function is much simpler than math_eps/sun_nextafter, but is not much faster:

   Computes the double precision of a number x away from zero, and returns a positive number. The function is
   similar to `math.eps`, but uses a much simpler algorithm. If no argument is given, then x is 1.

   The function uses code taken from:
   - https://randomascii.wordpress.com/2012/02/13/dont-store-that-in-a-float/
     originally written by Bruce Dawson, and
   - http://stackoverflow.com/questions/15685181/how-to-get-the-sign-mantissa-and-exponent-of-a-floating-point-number
     answered by eran (parts struct) */
static int math_precision (lua_State *L) {
  double x;
  double_cast d;
  x = luaL_optnumber(L, 1, 1);
  if (sizeof(double) != 8)
    luaL_error(L, "Error in " LUA_QS ": system does not support 64-bit C doubles.", "math.precision");
  d.f = x;
  /* lua_pushinteger(L, d.i & ((1LL << 52) - 1));  // mantissa
  lua_pushinteger(L, (d.i >> 52) & 0x7FF);         // exponent
  lua_pushinteger(L, (d.i >> 63) != 0);            // is argument negative ?
  lua_pushinteger(L, d.parts.mantissa);
  lua_pushinteger(L, d.parts.exponent);
  lua_pushinteger(L, d.parts.sign); */
  d.i += 1;
  lua_pushnumber(L, fabs(d.f - x));  /* precision */
  return 1;
}


/* Returns the number of integer digits (without decimal places) in the number x to the base b. By default, b is 10.
   Formerly implemented in Agena as: math.ndigits := << n :: number -> entier(ln(n)/ln(10) + 1) >>;
   See also: math.decompose, math.nthdigit. */
static int math_ndigits (lua_State *L) {  /* 2.7.0, extended 2.11.0 RC1 */
  lua_Number x, b;
  x = agn_checknumber(L, 1);
  b = agnL_optposint(L, 2, 10);
  if (x < 0) x = -x;
  if (x < 1)
    lua_pushnumber(L, 1);
  else
    lua_pushnumber(L, 1 + sun_floor(log(x)/log(b)));
  return 1;
}


/* Splits an integer x to the base b into its digits and returns them in a sequence, with the highest-order digit
   as the first element and the lowest-order digit as the last element. Any sign of x is ignored. By default, the base
   is 10, but you may choose any other positive base.

   Example: b := 256; > math.decompose(15 * b^2 + 7 * b + 1, 256): -> seq(15, 7, 1) */

static int math_decompose (lua_State *L) {  /* 2.7.0 */
  lua_Integer x, base;
  size_t c, i;
  x = agn_checkinteger(L, 1);
  base = agnL_optposint(L, 2, 10);  /* 2.10.0 modificattion */
  agn_createseq(L, 0);  /* we chose a sequence instead of a register for the number of digits in non-base10 numbers
                           cannot always be correctly be determined with C's log function due to round-off errors. */
  c = 0;
  if (x < 0) x = -x;
  if (x == 0) {
    agn_seqsetinumber(L, -1, ++c, 0);
  } else {
    while (x > 0) {
      agn_seqsetinumber(L, -1, ++c, x % base);
      x = x / base;
    }
  }
  /* reverse order of sequence */
  for (i=1; i <= c/2; i++) {
    lua_seqgeti(L, -1, i);
    lua_seqgeti(L, -2, c - i + 1);
    /* swap */
    lua_seqseti(L, -3, i);  /* and pop value */
    lua_seqseti(L, -2, c - i + 1);  /* and pop value */
  }
  return 1;
}


/* math.wrap(x, a, b)
   math.wrap(x, [a])

Conducts a range reduction of the number x to the interval [a, b) and returns a number.
In the second form, if a is not given, a is set to -Pi and b to +Pi. If a is given, a is set to -a and b to +a,
so a should be positive.

See also: math.norm, zx.range. */

#define P1 (4 * 7.8539812564849853515625e-01)
#define P2 (4 * 3.7748947079307981766760e-08)
#define P3 (4 * 2.6951514290790594840552e-15)

static int math_wrap (lua_State *L) {
  lua_Number theta, min, max;
  size_t nargs;
  theta = agn_checknumber(L, 1);
  nargs = lua_gettop(L);
  if (nargs == 1) {  /* 2.10.0: Same as math.wrap(theta, -Pi, Pi), but 35 percent faster, taken from Julia 0.5.0 */
    lua_Number y, r;
    int isneg, sign;
    theta = agn_checknumber(L, 1);
    isneg = signbit(theta);
    sign = isneg ? -1 : 1;
    if (isneg) theta = -theta;
    y = 2*luai_numint(theta/(2*PI));
    r = ((theta - y*P1) - y*P2) - y*P3;
    if (r > PI) r -= (2*PI);
    lua_pushnumber(L, sign*r);
  } else {
    min = agnL_optnumber(L, 2, -PI);
    max = agnL_optnumber(L, 3, PI);
    if (nargs == 2) {
      min = -min; max = -min;
    }
    if (min >= max)
      luaL_error(L, "Error in " LUA_QS ": minimum >= maximum.", "math.wrap");
    lua_pushnumber(L, tools_reducerange(theta, min, max));
  }
  return 1;
}


static int math_fdim (lua_State *L) {
  if (lua_gettop(L) == 2)
    lua_pushnumber(L, fdim(agn_checknumber(L, 1), agn_checknumber(L, 2)));
  else {  /* 2.13.0 unification with former math.fdima */
    lua_Number x, y, a;
    x = agn_checknumber(L, 1);
    y = agn_checknumber(L, 2);
    a = agn_checknumber(L, 3);
    if (isnan(x) || isnan(y))
      lua_pushundefined(L);
    else
      lua_pushnumber(L, x >= y ? x - y : a);
  }
  return 1;
}


static int math_isminuszero (lua_State *L) {  /* 2.8.4 */
  lua_Number x = agn_checknumber(L, 1);
  lua_pushboolean(L, luai_numisminuszero(x));
  return 1;
}


static int math_rint (lua_State *L) {  /* 2.8.6 */
  lua_pushnumber(L, rint(agn_checknumber(L, 1)));
  return 1;
}


/* 2.9.0, returns -1 if its numeric argument x is -infinity, +1 if its numeric argument x is +infinity,
  or 0 if neither.

  There seems to be a bug in MingW/GCC's version of isinf: it also returns +1 for -infinity, so an alternative
  has been implemented which is 3 percent faster than copysign(1, x) * fabs(isinf(x));
  (fpclassify(x) == FP_INFINITE) * copysign(1, x) is not faster;

  lua_Number x = agn_checknumber(L, 1);
  volatile lua_Number t;
  if ((t == x) && ((t - x) != 0.0))  # C89 compliant version taken from: http://www.devx.com/tips/Tip/42853
    lua_pushnumber(L, x < 0.0 ? -1 : 1);
  else
    lua_pushnumber(L, 0);
  is slower ! */
static int math_isinfinity (lua_State *L) {
  lua_Number x = agn_checknumber(L, 1);
  lua_pushnumber(L, (x == -HUGE_VAL || x == HUGE_VAL) * copysign(1, x));
  return 1;
}


static int math_infinite (lua_State *L) {
  lua_Number x = agn_checknumber(L, 1);
  lua_pushboolean(L, x == -HUGE_VAL || x == HUGE_VAL);
  return 1;
}


/* logarithmic factorial, written by John D. Cook and release to the public domain, modification by a walz;
   taken from: http://www.johndcook.com/blog/cpp_log_factorial; 2.10.0 */

static const lua_Number lnfactorials_lt_255[] = {
  0.000000000000000, 0.000000000000000, 0.693147180559945,  1.791759469228055,  3.178053830347946,
  4.787491742782046, 6.579251212010101, 8.525161361065415, 10.604602902745251, 12.801827480081469,
  15.104412573075516, 17.502307845873887, 19.987214495661885, 22.552163853123421, 25.191221182738683,
  27.899271383840894, 30.671860106080675, 33.505073450136891, 36.395445208033053, 39.339884187199495,
  42.335616460753485, 45.380138898476908, 48.471181351835227, 51.606675567764377, 54.784729398112319,
  58.003605222980518, 61.261701761002001, 64.557538627006323, 67.889743137181526, 71.257038967168000,
  74.658236348830158, 78.092223553315307, 81.557959456115029, 85.054467017581516, 88.580827542197682,
  92.136175603687079, 95.719694542143202, 99.330612454787428, 102.968198614513810, 106.631760260643450,
  110.320639714757390, 114.034211781461690, 117.771881399745060, 121.533081515438640, 125.317271149356880,
  129.123933639127240, 132.952575035616290, 136.802722637326350, 140.673923648234250, 144.565743946344900,
  148.477766951773020, 152.409592584497350, 156.360836303078800, 160.331128216630930, 164.320112263195170,
  168.327445448427650, 172.352797139162820, 176.395848406997370, 180.456291417543780, 184.533828861449510,
  188.628173423671600, 192.739047287844900, 196.866181672889980, 201.009316399281570, 205.168199482641200,
  209.342586752536820, 213.532241494563270, 217.736934113954250, 221.956441819130360, 226.190548323727570,
  230.439043565776930, 234.701723442818260, 238.978389561834350, 243.268849002982730, 247.572914096186910,
  251.890402209723190, 256.221135550009480, 260.564940971863220, 264.921649798552780, 269.291097651019810,
  273.673124285693690, 278.067573440366120, 282.474292687630400, 286.893133295426990, 291.323950094270290,
  295.766601350760600, 300.220948647014100, 304.686856765668720, 309.164193580146900, 313.652829949878990,
  318.152639620209300, 322.663499126726210, 327.185287703775200, 331.717887196928470, 336.261181979198450,
  340.815058870798960, 345.379407062266860, 349.954118040770250, 354.539085519440790, 359.134205369575340,
  363.739375555563470, 368.354496072404690, 372.979468885689020, 377.614197873918670, 382.258588773060010,
  386.912549123217560, 391.575988217329610, 396.248817051791490, 400.930948278915760, 405.622296161144900,
  410.322776526937280, 415.032306728249580, 419.750805599544780, 424.478193418257090, 429.214391866651570,
  433.959323995014870, 438.712914186121170, 443.475088120918940, 448.245772745384610, 453.024896238496130,
  457.812387981278110, 462.608178526874890, 467.412199571608080, 472.224383926980520, 477.044665492585580,
  481.872979229887900, 486.709261136839360, 491.553448223298010, 496.405478487217580, 501.265290891579240,
  506.132825342034830, 511.008022665236070, 515.890824587822520, 520.781173716044240, 525.679013515995050,
  530.584288294433580, 535.496943180169520, 540.416924105997740, 545.344177791154950, 550.278651724285620,
  555.220294146894960, 560.169054037273100, 565.124881094874350, 570.087725725134190, 575.057539024710200,
  580.034272767130800, 585.017879388839220, 590.008311975617860, 595.005524249382010, 600.009470555327430,
  605.020105849423770, 610.037385686238740, 615.061266207084940, 620.091704128477430, 625.128656730891070,
  630.172081847810200, 635.221937855059760, 640.278183660408100, 645.340778693435030, 650.409682895655240,
  655.484856710889060, 660.566261075873510, 665.653857411105950, 670.747607611912710, 675.847474039736880,
  680.953419513637530, 686.065407301994010, 691.183401114410800, 696.307365093814040, 701.437263808737160,
  706.573062245787470, 711.714725802289990, 716.862220279103440, 722.015511873601330, 727.174567172815840,
  732.339353146739310, 737.509837141777440, 742.685986874351220, 747.867770424643370, 753.055156230484160,
  758.248113081374300, 763.446610112640200, 768.650616799717000, 773.860102952558460, 779.075038710167410,
  784.295394535245690, 789.521141208958970, 794.752249825813460, 799.988691788643450, 805.230438803703120,
  810.477462875863580, 815.729736303910160, 820.987231675937890, 826.249921864842800, 831.517780023906310,
  836.790779582469900, 842.068894241700490, 847.352097970438420, 852.640365001133090, 857.933669825857460,
  863.231987192405430, 868.535292100464630, 873.843559797865740, 879.156765776907600, 884.474885770751830,
  889.797895749890240, 895.125771918679900, 900.458490711945270, 905.796028791646340, 911.138363043611210,
  916.485470574328820, 921.837328707804890, 927.193914982476710, 932.555207148186240, 937.921183163208070,
  943.291821191335660, 948.667099599019820, 954.046996952560450, 959.431492015349480, 964.820563745165940,
  970.214191291518320, 975.612353993036210, 981.015031374908400, 986.422203146368590, 991.833849198223450,
  997.249949600427840, 1002.670484599700300, 1008.095434617181700, 1013.524780246136200, 1018.958502249690200,
  1024.396581558613400, 1029.838999269135500, 1035.285736640801600, 1040.736775094367400, 1046.192096209724900,
  1051.651681723869200, 1057.115513528895000, 1062.583573670030100, 1068.055844343701400, 1073.532307895632800,
  1079.012946818975000, 1084.497743752465600, 1089.986681478622400, 1095.479742921962700, 1100.976911147256000,
  1106.478169357800900, 1111.983500893733000, 1117.492889230361000, 1123.006317976526100, 1128.523770872990800,
  1134.045231790853000, 1139.570684729984800, 1145.100113817496100, 1150.633503306223700, 1156.170837573242400
};

static int math_lnfact (lua_State *L) {
  lua_Integer n = agn_checknonnegint(L, 1);
  if (n > 254)
    lua_pushnumber(L, lgamma(n + 1));
  else
    lua_pushnumber(L, lnfactorials_lt_255[n]);
  return 1;
}


/* Returns x with its sign flipped if y is negative. For example abs(x) = flipsign(x, x).

   This is a port of the corresponding function of the same name available in Julia 0.5. 2.10.0 */
static int math_flipsign (lua_State *L) {
  lua_Number x, y;
  x = agn_checknumber(L, 1);
  y = agn_checknumber(L, 2);
  lua_pushnumber(L, (signbit(y)) ? -x : x);
  return 1;
}


/* The function returns the largest integer less than or equal to x/y.

   This is a port of the corresponding function of the same name available in Julia 0.5. 2.10.0 */
static int math_fld (lua_State *L) {
  lua_Number x, y;
  x = agn_checknumber(L, 1);
  y = agn_checknumber(L, 2);
  lua_pushnumber(L, luai_numentier(x/y));
  return 1;
}


/* The function returns the smallest integer larger than or equal to x/y.

   This is a port of the corresponding function of the same name available in Julia 0.5. 2.10.0 */
static int math_cld (lua_State *L) {
  lua_Number x, y;
  x = agn_checknumber(L, 1);
  y = agn_checknumber(L, 2);
  lua_pushnumber(L, luai_numentier(x/y) + 1);
  return 1;
}


/* The function computes x^p % m.

   This is a port of the corresponding function powermod available in Julia 0.5. 2.10.0 */
static int PowerModInt (int x, int e, int m) {  /* assumes m is non-zero */
  int f, p;  /* p = power */
  if (m == 1) return 0;
  else if (e == 0) return 1;
  x %= m;  /* reduce r initially */
  if (e < 0) {  /* invert x modulo m with Euclidian algorithm */
    if (x == 0) return 0;  /* 2.12.3 patch */
    x = 1/x % m;
    e = -e;
  }
  /* use repeated squaring method (right-to-left) */
  p = 1;
  f = tools_intpow(2, sun_ilogb(e) + 1);
  while (f > 1) {
    p = (p * p) % m;
    f >>= 1;
    if (f <= e) {
      p = (p * x) % m;
      e -= f;
    }
  }
  return p;
}

static int math_powmod (lua_State *L) {
  lua_Number x, p, m, r;
  x = agn_checknumber(L, 1);
  p = agn_checknumber(L, 2);
  m = agn_checknumber(L, 3);
  if (m == 0) {
    lua_pushundefined(L);
  } else {
    int isintp = tools_isint(p);  /* 2.14.13 change */
    if (tools_isint(x) && isintp && tools_isint(m))
      lua_pushnumber(L, PowerModInt(x, p, m));
    else {
      r = (isintp) ? luai_numipow(x, p) : luai_numpow(x, p);
      lua_pushnumber(L, (r < m) ? r : luai_nummod(r, m));
    }
  }
  return 1;
}


/* Returns the integer square root: the largest integer m such that m*m <= x.

   This is a port of the corresponding function of the same name in Julia 0.5. 2.10.0 */
static int math_isqrt (lua_State *L) {
  lua_Number x;
  lua_Integer s;
  x = agn_checknumber(L, 1);
  if (x == 0)
    lua_pushnumber(L, 0);
  else if (x < 0)
    lua_pushundefined(L);
  else {
    s = luai_numint(sqrt(x));
    s = (s + ((lua_Integer)x)/s) >> 1;
    lua_pushnumber(L, s*s > x ? s - 1 : s);
  }
  return 1;
}


/* Checks whether a given integer x is a power of base 2 and returns `true` or `false`. With
   floats, alwaysnegative x or too lagrge values, returns `fail`.
   see: https://github.com/OpsLabJPL/MarsImagesIOS/blob/master/Mars-Images/Math.m */
static int math_ispow2 (lua_State *L) {  /* 2.10.0 */
  lua_Number x;
  x = agn_checknumber(L, 1);
  /* if (unlikely(tools_isfloat(x) || x < 0 || sun_irregular(x))) */ /* 2.14.13 change, 2.16.0 optimisation ... */
  if (!tools_isregnonnegint(x))  /* 2.16.6 optimisation by 5 % */
    lua_pushfail(L);  /* ... or very large number not correctly representible any more ?  */
  else {
    int64_t y = x;  /* 2.16.0 fix; __builtin_popcount == 1 is slower */
    lua_pushboolean(L, !(!y) & !(y & (y - 1)));  /* = (x != 0) && 2 << (y - 1); tuned 2.16.6 */
  }
  return 1;
}


/* Returns the mantissa m of a number x such that m * 2^math.exponent(x) equals x.
   The result is identical to the one returned by `frexp`. The function is around
   20 percent faster but returns correct results only if your system supports IEEE 754
   floating-point numbers, whereas `frexp` always works regardless of the internal
   representation. See also: frexp, math.exponent.

   We use Sun's `frexp` implementation because it also treats subnormal values, 0,
   undefined and infinity values and is only slightly slower than a bitfield/bitmask-
   based implementation. */
static int math_mantissa (lua_State *L) {
  lua_pushnumber(L, sun_frexp_man(agn_checknumber(L, 1)));
  return 1;
}


/* Taken from: newlib-3.0.0.20180831/newlib/libm/math/s_signif.c
 *
 * ====================================================
 * Copyright (C) 1993 by Sun Microsystems, Inc. All rights reserved.
 *
 * Developed at SunPro, a Sun Microsystems, Inc. business.
 * Permission to use, copy, modify, and distribute this
 * software is freely granted, provided that this notice
 * is preserved.
 * ====================================================
 */

/* Returns the mantissa of number x in normalised form ranging between [1, 2), i.e. math.significand(x) =
   2*math.mantissa(x) = ldexp(x, -ilog2(x)). If x is 0, the return is 0.

   The function just mimics C's significand function.

   See also: `math.exponent`. */
static int math_significand (lua_State *L) {  /* 2.14.13 XXX UNDOCUMENTED */
  lua_Number x = agn_checknumber(L, 1);
  lua_pushnumber(L, sun_ldexp(x, (double)-sun_ilogb(x)));
  return 1;
}


/* Returns the exponent e of a number x such that math.mantissa(x) * 2^e equals x.
   The result is identical to the one returned by `frexp`. The function is around
   20 percent faster but returns correct results only if your system supports IEEE 754
   floating-point numbers, whereas `frexp` always works regardless of the internal
   representation. See also: frexp, math.mantissa.

   We use Sun's `frexp` implementation because it also treats subnormal values, 0,
   undefined and infinity values and is only slightly slower than a bitfield/bitmask-
   based implementation. */
static int math_exponent (lua_State *L) {  /* 2.10.0, changed and tuned 2.11.0 RC1 */
  lua_pushinteger(L, sun_frexp_exp(agn_checknumber(L, 1)));
  return 1;
}


/* Returns the unbiased exponent of number x, i.e. returns math.exponent(x) - 1;
   except for x = 0 where the result is -1023. */
static int math_uexponent (lua_State *L) {  /* 2.13.0, documented with 2.16.12 */
  double x;
  int32_t hx, ux;
  x = agn_checknumber(L, 1);
  GET_HIGH_WORD(hx, x);
  ux = ((((uint32_t)hx) >> 20) & 0x7ff) - 0x3ff;
  lua_pushnumber(L, ux);
  return 1;
}


/* The function is a plain binding to the C `%` modulus operator. Both its arguments must
   be integers. The return is an integer. */
static int math_modulus (lua_State *L) {  /* 2.10.0 */
  int y;
  y = agn_checkinteger(L, 2);
  /* if the second operand is zero, Agena CRASHES, so check it. */
  if (likely(y != 0))
    lua_pushnumber(L, agn_checkinteger(L, 1) % y);
  else
    lua_pushundefined(L);  /* as in Maple */
  return 1;
}


static int math_nearmod (lua_State *L) {  /* Returns the closest value to the given number x divisible by the given modulus m,
  equivalent to round(x/m) * m. 2.14.8, idea taken from VidCoder source file MathUtilities.cs */
  lua_Integer x, m;
  x = agn_checkinteger(L, 1);  /* number */
  m = agn_checkinteger(L, 2);  /* modulus */
  lua_pushnumber(L, (m == 0) ? AGN_NAN : sun_round((lua_Number)x/m) * m);
  return 1;
}


#define PRIME  257

/* See: http://rosettacode.org/wiki/Modular_inverse, C code */
static int mul_inv (int a, int b) {
  int t, nt, r, nr, q, tmp;
  if (b < 0) b = -b;
  if (a < 0) a = b - (-a % b);
  t = 0;  nt = 1;  r = b;  nr = a % b;
  while (nr != 0) {
    q = r/nr;
    tmp = nt;  nt = t - q*nt;  t = tmp;
    tmp = nr;  nr = r - q*nr;  r = tmp;
  }
  if (r > 1) return -1;  /* no inverse */
  if (t < 0) t += b;
  return t;
}


/* Computes the modular (multiplicative) inverse a^(-1) of an integer a (modulo m) such that
   a * a^(-1) = 1 (modulo m). m is 257 by default. If m is not prime, then not every nonzero
   integer a has a modular inverse - in this case the function returns `undefined`. */
static int math_modinv (lua_State *L) {
  LUAI_INT32 k, prime;
  int r;
  k = agn_checkinteger(L, 1);
  prime = agnL_optposint(L, 2, PRIME);
  r = mul_inv(k, prime);
  if (r == -1)
    lua_pushundefined(L);
  else
    lua_pushinteger(L, r);
  return 1;
}


/* Returns the relative spacing between |x| and its next larger number on the machines
   floating point system, taking into account the magnitude of its argument. The function
   works like `math.eps` with the `true` option but is 20 percent faster.

   If |x| < 1, you may choose a constant epsilon value yourself, e.g. Eps. */
static int math_epsilon (lua_State *L) {  /* 2.10.0, code optimised 2.10.1, extended 2.14.2 */
  lua_Number x, eps, ulp;
  int nres = 0;
  if (agn_isnumber(L, 1)) {
    int method;
    x = fabs(agn_tonumber(L, 1));
    method = agnL_optnonnegint(L, 2, 0);
    ulp = sun_nextafter(x, HUGE_VAL) - x;  /* machine epsilon at x */
    switch (method) {
      case 0: {  /* x * sqrt(ulp), for backward/forward differences */
        lua_pushnumber(L, x * sqrt(ulp));
        break;
      }
      case 1: {  /* x * cbrt(ulp), for central differences */
        lua_pushnumber(L, x * cbrt(ulp));
        break;
      }
      case 2: {  /* sqrt(ulp) * (|x| + sqrt(ulp)), for backward/forward differences  */
        eps = sqrt(ulp);
        lua_pushnumber(L, eps*(x + eps));
        break;
      }
      case 3: {  /* cbrt(ulp) * (|x| + cbrt(ulp)), for central differences */
        eps = cbrt(ulp);
        lua_pushnumber(L, eps*(x + eps));
        break;
      }
      default:
        luaL_error(L, "Error in " LUA_QS ": invalid option %d.", "math.epsilon", method);
    }
    nres = 1;
  } else if (lua_isfunction(L, 1)) {
    lua_Number abserr, origh, h;
    int nargs, n;
    x = agn_checknumber(L, 2);
    nargs = lua_gettop(L);
    n = 4;  /* iterations */
    if (lua_ispair(L, nargs)) {
      agn_pairgeti(L, nargs, 2);
      if (!agn_isnumber(L, -1)) {
        agn_poptop(L);
        luaL_error(L, "Error in " LUA_QS ": right-hand side of option must be a number.", "math.epsilon");
      }
      n = agn_tonumber(L, -1);
      agn_poptop(L);
      if (tools_isfloat(n) || n < 1)
        luaL_error(L, "Error in " LUA_QS ": right-hand side of option must be a positive integer.", "math.epsilon");
      nargs--;
    }
    h = agnL_fneps(L, 1, x, n, 3, nargs, &origh, &abserr);
    lua_pushnumber(L, origh);
    lua_pushnumber(L, h);
    lua_pushnumber(L, abserr);
    nres = 3;
  } else
     luaL_error(L, "Error in " LUA_QS ": wrong kind of argument.", "math.epsilon");
  return nres;
}


/* Checks whether the number x is subnormal, i.e. whether internally the leading digit of its mantissa is 0.
   The function returns `true` or `false`. Subnormal numbers are very close to zero, have reduced
   precision and lead to excessive CPU usage, they are in the range [-2.2250738585072009e-308, -4.9406564584124654e-324] and
   [4.9406564584124654e-324, 2.2250738585072009e-308]. 0, undefined and +/-infinity are not subnormal.
   fpclassify(x) == FP_SUBNORMAL does not work with MinGW GCC */
static int math_issubnormal (lua_State *L) {  /* 2.10.0 */
  lua_pushboolean(L, sun_subnormal(agn_checknumber(L, 1)));
  return 1;
}


/* Checks whether its numeric argument x is subnormal and in this case returns -0 or 0, depending on the sign of x; otherwise
   returns its argument x.
   The function is useful to prevent excessive CPU usage in case of arguments very close to zero.
   For more information, see `math.issubnormal`. */
#define chop(x) (sun_subnormal(x) ? ((signbit(x)) ? -0.0 : 0.0) : x)
static int math_zerosubnormal (lua_State *L) {  /* 2.10.0 */
  lua_Number x = agn_checknumber(L, 1);
  lua_pushnumber(L, chop(x));
  return 1;
}


/* Checks whether its numeric argument x is subnormal and in this case normalises it, i.e. returns a non-zero normalised value
   that is close to x; otherwise returns its argument x unaltered.
   It is useful to prevent excessive CPU usage in case of arguments very close to zero.
   For more information, see `math.issubnormal`. */
static int math_normalise (lua_State *L) {  /* 2.10.0, idea taken from Go 1.8, see bottom of file src/math/bits.go */
  lua_Number x;
  x = agn_checknumber(L, 1);
  lua_pushnumber(L, (sun_subnormal(x)) ? x * (1LL << 52LL) : x);  /* = 1 << 52 = 2^52 = 4503599627370496LL */
  return 1;
}


/* Returns the sign s, the mantissa m and the exponent e of the number x, in this order, such that
   s*m*2^e = x. The sign is -1 if x is negative (including -0) and 1 otherwise. The mantissa is a
   non-negative float, the exponent a negative or positive integer or zero.

   If any option is given, then instead of the sign the signbit s is returned: 1 if x is negative or
   -0, and 0 otherwise. In this case x = signum(-s)*m*2^e. */
static int math_frexp (lua_State *L) {  /* 2.10.0 */
  lua_Number x, m;
  int e, s, anyoption;
  x = agn_checknumber(L, 1);
  anyoption = lua_gettop(L) > 1;
  m = sun_xfrexp(x, &e, &s);
  if (!anyoption) s = (s == 0) ? 1 : -1;
  lua_pushnumber(L, s);
  lua_pushnumber(L, m);
  lua_pushnumber(L, e);
  return 3;
}


/* Checks whether a number can be represented exactly on your system.
   See: https://en.wikipedia.org/wiki/Double-precision_floating-point_format#IEEE_754_double-precision_binary_floating-point_format:_binary64 */
static int math_isirregular (lua_State *L) {  /* 2.10.0 */
  LUAI_INT32 hx, lx;
  lua_Number x;
  x = agn_checknumber(L, 1);
  EXTRACT_WORDS(hx, lx, x);
  hx &= 0x7FFFFFFF;     /* absolute value */
  if (hx < 0x43300000)  /* x < 2^52; a number with decimal places can internally be represented as a number with decimal places, but not */
    lua_pushfalse(L);   /* necessarily itself. With n < 52, the spacing between two subsequent representable numbers is the _fraction_ 2^(n-52). */
  else if (hx < 0x43400000)  /* 2^52 <= x < 2^53, representable numbers are exactly the integers; spacing between representable */
    lua_pushfail(L);         /* numbers is exactly 1 */
  else if (hx == 0x43400000 && lx == 0)  /* 2^53 <= x <= 2^53 + 1, representable numbers are exactly the integers; spacing between representable */
    lua_pushfail(L);    /* numbers is exactly 1, 2.16.0 fix */
  else                  /* x > 2^53 (+ 1), an integer mostly cannot be exactly represented; with n > 52, spacing is the _integer_ 2^(n-52) */
    lua_pushtrue(L);
  return 1;
}


/* Returns its argument x - a number - if x is non-negative, otherwise returns 0. By passing any
   non-negative optional number d (the `direction`), the return is the same.

   By passing any non-positive optional number d, returns x if it is negative, otherwise returns 0.

   If x should be `undefined`, you can return any other number by passing the optional argument subs,
   which is `undefined` by default. */
static int math_branch (lua_State *L) {  /* 2.10.0, see also (unused) tools_branch */
  lua_Number x, d, r;
  x = agn_checknumber(L, 1);
  d = agnL_optnumber(L, 2, HUGE_VAL);  /* the `direction` or interval: if d > 0 -> return in [0, infinity], */
  if (isnan(x)) x = agnL_optnumber(L, 3, AGN_NAN);  /* optionally convert x = undefined to any other number, default is `itself` */
  r = signbit(x) == 0;                 /* if d <= 0 -> return in [-infinity, 0] */
  r = (d > 0) ? r*x : (!r)*x;
  if (luai_numisminuszero(r)) r = 0;   /* convert -0 to 0 */
  lua_pushnumber(L, r);
  return 1;
}


/* The falling factorial function computes x*(x - 1)*(x - 2)* ... *(x - n + 1), with x a number and n an integer.
   If n is negative, the rising factorial function is computed. See also: fact. */
static int math_fall (lua_State *L) {  /* 2.10.2 */
  int i, n;
  lua_Number r, x;
  r = 1.0;
  x = agn_checknumber(L, 1);
  n = agn_checkinteger(L, 2);
  if (n > 0) {
    for (i=1; i <= n; i++) {
      r *= x;
      x -= 1.0;
    }
  } else if (n < 0) {  /* rising factorial = Pochhammer function */
    for (i=-1; n <= i; i--) {
      r *= x;
      x += 1.0;
    }
  }
  lua_pushnumber(L, r);
  return 1;
}


#define diff(x)         fabs((x) - sun_round((x)))
static double aux_pochhammer (lua_Number x, lua_Number n, lua_Number limit) {
  lua_Number xplusn = x + n;
  int isnegintn, ismintx;
  isnegintn = tools_isnegint(n);
  if (x == 0 && !isnegintn)
    return (n == 0) ? 1 : 0;
  else if (x != 0.0 && n == 0.0)  /* like in Maple V R 4 */
    return 1;
  /* for !(...) condition see http://functions.wolfram.com/introductions/PDF/Pochhammer.pdf: */
  ismintx = tools_isint(-x);
  if ((xplusn == 0 || tools_isnegint(xplusn) || isnegintn || ((xplusn <= 0) && diff(xplusn) <= tools_matheps(xplusn)) ) &&
            !(ismintx && -x >= 0 && tools_isint(n) && n <= -x))
    return AGN_NAN;
  else if (n < 0 && (xplusn == 0 || tools_isnegint(xplusn)))
    return 1/aux_pochhammer(xplusn, -n, limit);
  else if (isnegintn)  /* put this before next condition, don't put it into conditions above */
    return 1/aux_pochhammer(xplusn, -n, limit);
  else if (tools_isnegint(x) && tools_isfloat(n))
    return 0;
  else if (tools_isposint(n) && n < limit) {  /* use Maple V R4's definition */
    lua_Number r;
    int i;
    r = 1.0;
    for (i=1; i <= n; i++) {
      r *= x;
      x += 1.0;  /* r is the result */
    }
    return r;
  }
  else if (xplusn >= limit && x > 0 && n > 0)  /* avoid overflow */
    return sun_exp(lgamma(xplusn) - lgamma(x));
  else
    return cephes_gamma(xplusn)/cephes_gamma(x);
}

/* Computes the Pochhammer function (rising factorial), where both x and n can be real numbers. It returns a number. Test the function independently with poch2.tst, put test cases into math2.tst ! */
static int math_pochhammer (lua_State *L) {  /* 2.13.0 */
  lua_Number x, n, limit;
  x = agn_checknumber(L, 1);
  n = agn_checknumber(L, 2);
  limit = agnL_optnumber(L, 3, 33);  /* with n <= 32, simple MVR4 multiplication is faster, else (l)gamma versions */
  /* with x + n > 174, we are in infinity */
  lua_pushnumber(L, aux_pochhammer(x, n, limit));
  return 1;
}


/* Computes the rectangular pulse function for number x. a represents the rising edge, and b the falling edge
   of the rectangular pulse function. By default, a = -0.5 and b = +0.5.

   It returns 0 if x < a or x > b; 0.5 if (x = a or x = b) and a <> b, and 1 otherwise.

   See also: heaviside.

   See: https://de.mathworks.com/help/symbolic/rectangularpulse.html?requestedDomain=www.mathworks.com,
        https://en.wikipedia.org/wiki/Rectangular_function */
static int math_rectangular (lua_State *L) {  /* 2.4.2 / 2.11.0 RC1 */
  lua_Number x, a, b, c;
  int pi;
  c = 0.5;
  x = agn_checknumber(L, 1);
  if (lua_gettop(L) == 2 && lua_isboolean(L, 2)) {
    a = -c; b = c; pi = lua_toboolean(L, 2);
  } else {
    a = agnL_optnumber(L, 2, -c);
    b = agnL_optnumber(L, 3, c);
    pi = agnL_optboolean(L, 4, 0);
  }
  if (a > b)
    luaL_error(L, "Error in " LUA_QS ": second argument larger than third one.", "math.rectangular");
  if (x < a || x > b)  /* 2.11.0 RC2 patch */
    lua_pushnumber(L, 0);
  else if ((x == a || x == b) && a != b) {
    if (!pi)  /* 2.11.0 RC2 */
      lua_pushnumber(L, c);
    else  /* box distribution Pi(x): 1 if |x| < 1/2, 0 if |x| > 1/2, undefined if |x| = 0.5 */
      lua_pushnumber(L, AGN_NAN);
  } else
    lua_pushnumber(L, 1);
  return 1;
}


/* Computes the triangular function for the given number x. By default, the function returns 0 if |x| < 0.5,
   and |x| otherwise.

   By passing a left and a right border a, b, the function returns non-zero values in this range, and 0 otherwise,
   so the general formula used by the function is:

   math.triangular(x, a, b) := max(0, 1 - |2*(x - offset)/d|), where d := |b - a| and offset := a + d/2. */
static int math_triangular (lua_State *L) {  /* 2.11.0 RC2 */
  lua_Number x, a, b, c, d, offset;
  c = 0.5;
  x = agn_checknumber(L, 1);
  a = agnL_optnumber(L, 2, -c);
  b = agnL_optnumber(L, 3, c);
  if (a > b)
    luaL_error(L, "Error in " LUA_QS ": second argument larger than third one.", "math.triangular");
  d = fabs(b - a);
  offset = a + 0.5*d;  /* shift on abscissa with respect to zero */
  x -= offset;  /* shift point to new centre */
  /* for formula and alternatives, see: https://en.wikipedia.org/wiki/Triangular_function */
  lua_pushnumber(L, fMax(0, 1 - fabs(x*2/d)));
  return 1;
}


/* Returns 0 if its number argument x is zero or close to zero, and 1 otherwise:
   0 if |x| <= eps
   1 if |x| > eps

   With complex numbers a + I*b, returns
   0 if |a| <= eps and |b| <= eps
   1 if |a| > eps and |b| > eps

   By default, eps is set to the constant Eps.

   See also: heaviside, math.rectangular, math.zeroin. */
static int math_unitise (lua_State *L) {  /* 2.11.0 RC1 */
  int type;
  lua_Number eps;
  type = lua_type(L, 1);
  eps = agnL_optnonnegative(L, 2, agn_getepsilon(L));  /* 2.13.0 change */
  if (type == LUA_TNUMBER) {
    lua_Number x = fabs(agn_tonumber(L, 1));
    lua_pushinteger(L, x > eps);
  } else if (type == LUA_TCOMPLEX) {
    lua_Number a, b;
#ifndef PROPCMPLX
    agn_Complex z = agn_tocomplex(L, 1);
    a = creal(z); b = cimag(z);
#else
    a = agn_complexreal(L, 1); b = agn_compleximag(L, 1);
#endif
    lua_pushinteger(L, fabs(a) > eps && fabs(b) > eps);
  } else
    luaL_error(L, "Error in " LUA_QS ": argument must be a (complex) number.", "math.unitise");
  return 1;
}


/* Returns x clipped to be between a and b. The return is x if a <= x <= b, a if x < a, and b if x > b.
   By default a = -1, b = +1.
   The function is a port of Mathematica's Clip function.
   See http://reference.wolfram.com/language/ref/Clip.html */

static int math_clip (lua_State *L) {  /* 2.11.0 RC2, extended 2.13.0 */
  lua_Number x, a, b;
  x = agn_checknumber(L, 1);
  a = agnL_optnumber(L, 2, -1);
  b = agnL_optnumber(L, 3, 1);
  if (lua_gettop(L) == 2) {  /* ex-math.symtrunc emulation */
    if (a > 0) a = -a;
    b = -a;
  }
  if (a > b)
    luaL_error(L, "Error in " LUA_QS ": second argument larger than third one.", "math.clip");
  if (!lua_isfunction(L, 4)) {
    lua_Number t;
    t = x < a ? a : x;
    x = t > b ? b : t;
  } else {  /* 2.13.0, we have a function */
    if (x < a || x > b) {    /* x not part of [a, b] ? */
      lua_pushvalue(L, 4);   /* push function */
      lua_pushnumber(L, x);  /* push argument */
      lua_call(L, 1, 1);     /* leaves result on stack */
      if (!agn_isnumber(L, -1)) {
        agn_poptop(L);
        luaL_error(L, "Error in " LUA_QS ": function must return a number.", "math.clip");
      }
      x = agn_tonumber(L, -1);
      agn_poptop(L);         /* level stack */
    }  /* else x in [a, b] as just return x unchanged */
  }  /* stack is leveled */
  lua_pushnumber(L, x);
  return 1;
}


/* The Dirac delta function, also known as the impulse function, returns 0 for all numbers x other than 0, and `infinity`
   if x = 0, iff eps is set to zero which is the default.

   If eps is set to any positive value x, returns 1/(2*eps)*exp(-|x|/eps) even if x = 0.

   For formula(s), see: https://www.maplesoft.com/support/help/maple/view.aspx?path=Dirac */
static int math_dirac (lua_State *L) {  /* 2.11.0 RC2 */
  lua_Number x, eps, reps;
  x = agn_checknumber(L, 1);
  if (isnan(x)) {  /* 2.14.13 fix */
    lua_pushundefined(L);
    return 1;
  }
  eps = agnL_optnumber(L, 2, 0);
  if (eps == 0) {  /* ideal case, the default: 0 if x <> 0 and infinity if x = 0 */
    lua_pushnumber(L, (x == 0) ? HUGE_VAL : 0);
  } else {
    reps = 1/eps;
    lua_pushnumber(L, 0.5*reps * exp(-fabs(x)*reps));  /* 0.638 secs, best convergence, max. value is 33554432 */
  }
  /* piecewise(And(-eps/2 < x, x < eps/2), 1/eps, fabs(x) > eps/2, 0); */
  /* lua_pushnumber(L, 1/M_PI * eps/(x*x + eps*eps));   0.197 secs, slow convergence to infinity, max. value is 21361414.861763 */
  /* lua_pushnumber(L, 1/M_PI * sin(x/eps)/x);          0.369 secs, slow convergence to infinity, max. value is 21361414.861763 */
  return 1;
}


/* For the given number x, returns
   - 0 if x is zero (= constant `math.fp_zero`),
   - 1 if x is infinite, i.e. +/-infinity (= constant `math.fp_infinite`),
   - 2 if x is undefined (= constant `math.fp_nan`),
   - 3 if x is normal (= constant `math.fp_normal`),
   - 4 is x is subnormal (= constant `math.fp_subnormal`; fpclassify might not return correct results in this case on some systems).
   The function returns `fail` if it could not determine the type of floating-point number (of C type double). It is a port to C's
   fpclassify.

   See also: ...

   The function has originally been written by Luiz Henrique de Figueiredo, `C99 math functions for Lua` library file
   lmathx.c, on 04 Apr 2010 22:53:32: `This code is hereby placed in the public domain.` */
static int math_fpclassify (lua_State *L) {  /* 2.11.0 RC2 */
  switch (tools_fpclassify(agn_checknumber(L, 1))) {  /* call portable version of fpclassify */
    case FP_ZERO:
      lua_pushinteger(L, 0); break;
    case FP_INFINITE:
      lua_pushinteger(L, 1); break;
    case FP_NAN:
      lua_pushinteger(L, 2); break;
    case FP_NORMAL:
      lua_pushinteger(L, 3); break;
    case FP_SUBNORMAL:  /* in MinGW incorrect results */
      lua_pushinteger(L, 4); break;
    default:
      lua_pushfail(L);
  }
  return 1;
}


/* Shrinks a number x more or less near zero to exactly zero, using one of many methods.

   The default for eps is Eps. The standard method is 0 for hard shrinking.

   In the first form performs hard shrinking by returning
   - 0 if x < eps,
   - and x otherwise.
   In the second form performs soft shrinking by returning:
   - 0 if |x| <= eps,
   - sign(x) * (abs(x) - eps) if |x| > eps.
   eps is Eps by default.
   The function is a port of Mathematica's Chop function.
   See http://reference.wolfram.com/language/ref/Chop.html &
       https://reference.wolfram.com/legacy/v8/ref/Threshold.html */

static lua_Number aux_chop (lua_State *L, lua_Number x, lua_Number dx, int n, int kind) {  /* 2.11.1 */
  switch (kind) {
    case 0:  /* "Hard" */
      return (fabs(x) >= dx)*x;
    case 1:  /* "Soft" */
      return luai_numsignum(x)*(fMax(0, fabs(x) - dx));
    case 2:  /* "Firm" */
      luaL_error(L, "Error in " LUA_QS ": `firm` method not yet supported.", "math.chop");
      return 0;
    case 3:  /* "PiecewiseGarrote" */
      return (fabs(x) >= dx) * (x - dx*dx/x);
    case 4:  /* "SmoothGarrote" */
      return tools_intpow(x, 2*n + 1)/(tools_intpow(x, 2*n) + tools_intpow(dx, 2*n));
    case 5:  /* "Hyperbola" */
      return (fabs(x) >= dx) * luai_numsignum(x)*tools_mhypot(x, dx);
    default:
      luaL_error(L, "Error in " LUA_QS ": wrong option %ls.", "math.chop", kind);
      return 0;
  }
}

static int math_chop (lua_State *L) {  /* 2.11.0 RC2 */
  lua_Number dx;
  int kind, n;
  dx = agnL_optnumber(L, 2, agn_getepsilon(L));
  n = agnL_optinteger(L, 4, 1);
  kind = 0;
  if (lua_gettop(L) == 3) {
    if (lua_isboolean(L, 3))
      kind = lua_toboolean(L, 3);
    else if (agn_isnumber(L, 3))
      kind = lua_tointeger(L, 3);
    else
      luaL_error(L, "Error in " LUA_QS ": third argument must be a number or boolean.", "math.chop");
  }
  switch (lua_type(L, 1)) {
    case LUA_TNUMBER: {
      lua_Number x = agn_checknumber(L, 1);
      lua_pushnumber(L, aux_chop(L, x, dx, n, kind));
      break;
    }
    case LUA_TCOMPLEX: {
      lua_Number a, b;
#ifndef PROPCMPLX
      agn_Complex z = agn_tocomplex(L, 1);
      a = creal(z); b = cimag(z);
#else
      a = agn_complexreal(L, 1); b = agn_compleximag(L, 1);
#endif
      a = aux_chop(L, a, dx, n, kind);
      b = aux_chop(L, b, dx, n, kind);
#ifndef PROPCMPLX
      __real__ z = a;
      __imag__ z = b;
      agn_createcomplex(L, z);
#else
      agn_createcomplex(L, a, b);
#endif
      break;
    }
    default:
      luaL_error(L, "Error in " LUA_QS ": argument must be a (complex) number.", "math.chop");
  }
  return 1;
}


static int math_sincos (lua_State *L) {  /* 2.11.0, 15 percent faster than calling Agena's sin and cos operators */
  lua_Number x, s, c;
  x = agn_checknumber(L, 1);
  tools_sincos(x, &s, &c);
  lua_pushnumber(L, s);
  lua_pushnumber(L, c);
  return 2;
}


static int math_sinhcosh (lua_State *L) {
  lua_Number z, sih, coh;
  z = agn_checknumber(L, 1);
  tools_sinhcosh(z, &sih, &coh);
  lua_pushnumber(L, sih);
  lua_pushnumber(L, coh);
  return 2;
}


static int math_rempio2 (lua_State *L) {  /* 2.11.0, undocumented */
  lua_Number x, y[2];
  int r;
  x = agn_checknumber(L, 1);
  r = sun_rem_pio2(x, y);
  lua_pushinteger(L, r);
  lua_pushnumber(L, y[0]);
  lua_pushnumber(L, y[1]);
  return 3;
}


/* Returns a function that gets a number with each call, adds it to an internal accumulator using Kahan-Ozawa (KO) summation, and
   returns the accumulated sum. The function can be used if high precision numeric addition is needed. The initial value of the
   accumulator is 0. The function automatically takes care of storing and processing the KO correction value - so the user does
   not have to worry about this. See also: math.koadd. */

static int iter_kahan (lua_State *L) {  /* 2.12.0 RC 3, based on math_koadd */
  volatile lua_Number q, s, v, x, t;
  x = (agn_isnumber(L, 1)) ? agn_tonumber(L, 1) : 0;  /* 2.12.0 RC 4 */
  s = lua_tonumber(L, lua_upvalueindex(1));  /* accumulator */
  q = lua_tonumber(L, lua_upvalueindex(2));  /* correction value */
  v = x - q;
  t = s + v;
  q = (t - s) - v;
  s = t;
  lua_pushnumber(L, s);
  lua_replace(L, lua_upvalueindex(1));  /* update accumulator */
  lua_pushnumber(L, q);
  lua_replace(L, lua_upvalueindex(2));  /* update q */
  lua_pushnumber(L, (fabs(s) < AGN_HEPSILON) ? 0 : s);  /* 2.12.1, do not modify s itself for it will corrupt the intermediate sum */
  return 1;  /* return result */
}

static int iter_kahanozawa (lua_State *L) {  /* 2.12.0 RC 3, based on math_koadd */
  volatile lua_Number q, s, sold, u, v, w, x, t;
  x = (agn_isnumber(L, 1)) ? agn_tonumber(L, 1) : 0;  /* 2.12.0 RC 4 */
  s = lua_tonumber(L, lua_upvalueindex(1));  /* accumulator */
  q = lua_tonumber(L, lua_upvalueindex(2));  /* correction value */
  v = x - q;
  sold = s;
  s += v;
  if (fabs(x) < fabs(q)) {
    t = x;
    x = -q;
    q = t;
  }
  u = (v - x) + q;
  if (fabs(sold) < fabs(v)) {
    t = sold;
    sold = v;
    v = t;
  }
  w = (s - sold) - v;
  q = u + w;
  lua_pushnumber(L, s);
  lua_replace(L, lua_upvalueindex(1));  /* update accumulator */
  lua_pushnumber(L, q);
  lua_replace(L, lua_upvalueindex(2));  /* update q */
  lua_pushnumber(L, (fabs(s) < AGN_HEPSILON) ? 0 : s);  /* 2.12.1, do not modify s itself for it will corrupt the intermediate sum */
  return 1;  /* return result */
}

static int iter_kahanbabuska (lua_State *L) {  /* 2.12.0 RC 3, based on stats_KahanBabuskaMean */
  volatile lua_Number s, cs, ccs, t, c, cc, x, s0;
  x   = (agn_isnumber(L, 1)) ? agn_tonumber(L, 1) : 0;  /* 2.12.0 RC 4 */
  s   = agn_tonumber(L, lua_upvalueindex(1));  /* accumulator */
  cs  = agn_tonumber(L, lua_upvalueindex(2));  /* first correction value */
  ccs = agn_tonumber(L, lua_upvalueindex(3));  /* second correction value */
  t = s + x;
  c = (fabs(s) >= fabs(x)) ? (s - t) + x : (x - t) + s;
  s = t;
  t = cs + c;
  cc = (fabs(cs) >= fabs(c)) ? (cs - t) + c : (c - t) + cs;
  cs = t;
  ccs += cc;
  lua_pushnumber(L, s);
  lua_replace(L, lua_upvalueindex(1));  /* update accumulator */
  lua_pushnumber(L, cs);
  lua_replace(L, lua_upvalueindex(2));  /* update cs */
  lua_pushnumber(L, ccs);
  lua_replace(L, lua_upvalueindex(3));  /* update ccs */
  /* the correction values are added to the final sum, not to intermediate sum. */
  s0 = s + cs + ccs;
  lua_pushnumber(L, (fabs(s0) < AGN_HEPSILON) ? 0 : s0);  /* 2.12.1, do not modify s itself for it will corrupt the intermediate sum */
  return 1;
}

/* See: https://en.wikipedia.org/wiki/Kahan_summation_algorithm */
static int iter_neumaier (lua_State *L) {  /* 2.12.0 RC 3, based on stats_KahanBabuskaMean */
  volatile lua_Number s, c, t, x, s0;
  x = (agn_isnumber(L, 1)) ? agn_tonumber(L, 1) : 0;  /* 2.12.0 RC 4 */
  s = agn_tonumber(L, lua_upvalueindex(1));   /* accumulator */
  c = agn_tonumber(L, lua_upvalueindex(2));   /* correction value */
  t = s + x;
  c += (fabs(s) >= fabs(x)) ? (s - t) + x : (x - t) + s;
  s = t;
  lua_pushnumber(L, s);
  lua_replace(L, lua_upvalueindex(1));  /* update accumulator */
  lua_pushnumber(L, c);
  lua_replace(L, lua_upvalueindex(2));  /* update c */
  /* the correction value is added to the final sum, not to intermediate sum. */
  s0 = s + c;
  lua_pushnumber(L, (fabs(s0) < AGN_HEPSILON) ? 0 : s0);  /* 2.12.1, do not modify s itself for it will corrupt the intermediate sum */
  return 1;
}

/* See: https://github.com/JuliaLang/julia/blob/d386e40c17d43b79fc89d3e579fc04547241787c/base/reduce.jl#L366-L371 */
static int iter_kbn (lua_State *L) {  /* 2.12.0 RC 3, based on iter_neumaier */
  volatile lua_Number s, c, t, x, s0;
  int flag;
  x = (agn_isnumber(L, 1)) ? agn_tonumber(L, 1) : 0;  /* 2.12.0 RC 4 */
  s = agn_tonumber(L, lua_upvalueindex(1));      /* accumulator */
  c = agn_tonumber(L, lua_upvalueindex(2));      /* correction value */
  flag = lua_tointeger(L, lua_upvalueindex(3));  /* flag */
  if (flag)
    s = x - c;
  else {
    t = s + x;
    c -= (fabs(s) >= fabs(x)) ? (s - t) + x : (x - t) + s;
    s = t;
  }
  lua_pushnumber(L, s);
  lua_replace(L, lua_upvalueindex(1));  /* update accumulator */
  lua_pushnumber(L, c);
  lua_replace(L, lua_upvalueindex(2));  /* update c */
  if (flag) {
    lua_pushinteger(L, 0);
    lua_replace(L, lua_upvalueindex(3));  /* switch flag */
  }
  /* the correction value is added to the final sum, not to intermediate sum. */
  s0 = s - c;
  lua_pushnumber(L, (fabs(s0) < AGN_HEPSILON) ? 0 : s0);  /* 2.12.1, do not modify s itself for it will corrupt the intermediate sum */
  return 1;
}

static int math_accu (lua_State *L) {  /* 2.12.0 RC 3, around 25 percent faster than an Agena implementation */
  int nargs;
  lua_Number start;
  const char *method;
  nargs = lua_gettop(L);
  start = (agn_isnumber(L, 1)) ? agn_tonumber(L, 1) : 0;
  method = (agn_isstring(L, nargs)) ? agn_tostring(L, nargs) : "neumaier";
  lua_pushnumber(L, start);  /* initial value */
  lua_pushnumber(L, 0);  /* correction value */
  if (tools_streq(method, "kahan"))  /* 2.16.12 tweak */
    lua_pushcclosure(L, &iter_kahan, 2);
  else if (tools_streq(method, "ozawa"))  /* 2.16.12 tweak */
    lua_pushcclosure(L, &iter_kahanozawa, 2);
  else if (tools_streq(method, "neumaier"))  /* 2.16.12 tweak */
    lua_pushcclosure(L, &iter_neumaier, 2);
  else if (tools_streq(method, "babuska")) {  /* 2.16.12 tweak */
    lua_pushnumber(L, 0);  /* ccs correction value */
    lua_pushcclosure(L, &iter_kahanbabuska, 3);
  } else if (tools_streq(method, "kbn")) {  /* 2.16.12 tweak */
    lua_pushinteger(L, 1);  /* flag */
    lua_pushcclosure(L, &iter_kbn, 3);
  } else
    luaL_error(L, "Error in " LUA_QS ": unknown method " LUA_QS ".", "math.accu", method);
  return 1;
}


/* Computes the unit of least precision (ULP), the spacing between floating-point numbers, as a measure
   of accuracy in numeric calculations. */
static int math_ulp (lua_State *L) {  /* 2.12.3 */
  lua_Number x = agn_checknumber(L, 1);
  lua_pushnumber(L, sun_nextafter(x, HUGE_VAL) - x);
  return 1;
}


/* greatest common divisor, 0.9.0 as of Jan 09, 2008; patched 0.26.1, 11.08.2009, improved 2.10.0
   math.gcd := proc(x :: number, y :: number) is
   return when float(x) or float(y) with 1;
   while y <> 0 do
      x, y := y, x % y
   od;
   return x
end; */
static int math_gcd (lua_State *L) {  /* 2.12.3, 85 % faster than the Agena implementation */
  lua_Number x, y;
  x = agn_checknumber(L, 1);
  y = agn_checknumber(L, 2);
  lua_pushnumber(L, tools_gcd(x, y));
  return 1;
}


/* least common multiple, 0.9.0 as of Jan 09, 2008; patched 0.26.1, 11.08.2009, changed 2.10.0
   math.lcm := proc(x :: number, y :: number) is
   return if y = 0 then 0 else abs(x/math.gcd(x, y)*y) fi
end; */
static int math_lcm (lua_State *L) {  /* 2.12.3, 165 % faster than the Agena implementation */
  lua_Number x, y;
  x = agn_checknumber(L, 1);
  y = agn_checknumber(L, 2);
  if (y == 0)
    lua_pushnumber(L, 0);
  else
    lua_pushnumber(L, fabs(x/tools_gcd(x, y)*y));
  return 1;
}


/*
math.fib := proc(n :: nonnegint) is  # modified 2.12.1
   feature reminisce; # creates a read-write remember table
   if n < 2 then
      return n  # exit conditions, changed to Maple V R4's combinat.fibonacci standard with 2.10.2
   elif n < 71 then  # use Golden angle, values precomputed with Maple V Release 4 are:
      # see:  http://stackoverflow.com/questions/6037472/can-a-fibonacci-function-be-written-to-execute-in-o1-time
      # 1.6180339887498948482  = (1 + sqrt(5))/2
      # -0.6180339887498948482 = (1 - sqrt(5))/2
      # 2.2360679774997896964  = sqrt(5)
      # do not use int for in Linux, this causes wrong results
      return math.rint( (1.6180339887498948482**n - (-0.6180339887498948482)**n)/2.2360679774997896964 )  # 2.10.3 change
   elif n < 77 then  # avoid round-off errors in the formula above with n >= 71
      # with remember tables, recursion is much faster than an iterative approach
      return procname(n - 2) + procname(n - 1)
   else  # result is larger than math.lastcontint, thus is not accurate
      return fail
   fi
end; */
static unsigned long long int aux_fib[] = {
                 0ULL,                1ULL,               1ULL,               2ULL,                3ULL,
                 5ULL,                8ULL,              13ULL,              21ULL,               34ULL,
                55ULL,               89ULL,             144ULL,             233ULL,              377ULL,
               610ULL,              987ULL,            1597ULL,            2584ULL,             4181ULL,
              6765ULL,            10946ULL,           17711ULL,           28657ULL,            46368ULL,
             75025ULL,           121393ULL,          196418ULL,          317811ULL,           514229ULL,
            832040ULL,          1346269ULL,         2178309ULL,         3524578ULL,          5702887ULL,
           9227465ULL,         14930352ULL,        24157817ULL,        39088169ULL,         63245986ULL,
         102334155ULL,        165580141ULL,       267914296ULL,       433494437ULL,        701408733ULL,
        1134903170ULL,       1836311903ULL,      2971215073ULL,      4807526976ULL,       7778742049ULL,
       12586269025ULL,      20365011074ULL,     32951280099ULL,     53316291173ULL,      86267571272ULL,
      139583862445ULL,     225851433717ULL,    365435296162ULL,    591286729879ULL,     956722026041ULL,
     1548008755920ULL,    2504730781961ULL,   4052739537881ULL,   6557470319842ULL,   10610209857723ULL,
    17167680177565ULL,   27777890035288ULL,  44945570212853ULL,  72723460248141ULL,  117669030460994ULL,
   190392490709135ULL,  308061521170129ULL, 498454011879264ULL, 806515533049393ULL, 1304969544928657ULL,
  2111485077978050ULL, 3416454622906707ULL };  /* 0 to 76 */

static int math_fib (lua_State *L) {  /* 2.12.5, 40 percent faster than the Agena implementation */
  int n = agn_checknonnegint(L, 1);
  if (n > 76)
    lua_pushfail(L);
  else
    lua_pushnumber(L, aux_fib[n]);
  return 1;
}


/* For any non-negative integer n returns the index i of the Fibonacci number with fib(i) <= n < fib(i + 1)
   math.fibinv := << x :: nonnegint -> if x < 2 then x else entier(log(x * sqrt 5 + 1/2, math.Phi)) fi >>;  # 2.12.5 */
static int math_fibinv (lua_State *L) {  /* 2.12.6, twice as fast as the Agena implementation, activated 2.16.12 */
  int n = agn_checknonnegint(L, 1);
  if (n < 2)
    lua_pushinteger(L, n);
  else
    lua_pushinteger(L, sun_floor(sun_log(n * SQRT5 + 0.5)*INVPHILN));  /* 2.16.12 optimisation */
  return 1;
}


/* Computes either the Stirling number for n, k, of the first or the second kind. kind is 1 by default
   to compute the first kind; you can set it to 2 to evaluate the second kind. 2.13.0 */
static int math_stirnum (lua_State *L) {
  int i, j, n, k, maxj, *a, kind;
  n = agn_checkinteger(L, 1);
  k = agn_checkinteger(L, 2);
  kind = agnL_optposint(L, 3, 1);
  if (kind > 2)
    luaL_error(L, "Error in " LUA_QS ": third argument must be 1 or 2.", "math.stirnum");
  if (tools_isnegint(n) || tools_isnegint(k)) {  /* for both 1st and 2nd kind */
    lua_pushnumber(L, 0);
    return 1;
  }
  if (n < k || n == k) {
    lua_pushinteger(L, (n < k) ? 0 : 1);
    return 1;
  }
  maxj = n - k;
  a = malloc((maxj + 1) * sizeof(int));
  if (kind == 1) {  /* first kind */
    /* Source: https://stackoverflow.com/questions/5133050/dynamic-programming-approach-to-calculating-stirlings-number */
    for (i=0; i <= maxj; i++) a[i] = 0;
    a[0] = 1;
    for (i=1; i <= k; i++) {
      for (j=1; j <= maxj; j++)
        a[j] -= (i + j - 1)*a[j - 1];
    }
  } else {  /* second kind */
    /* Source: https://stackoverflow.com/questions/5133050/dynamic-programming-approach-to-calculating-stirlings-number */
    for (i=0; i <= maxj; i++) a[i] = 1;
    for (i=2; i <= k; i++) {
      for (j=1; j <= maxj; j++)
        a[j] += i*a[j - 1];
    }
  }
  lua_pushinteger(L, a[maxj]);
  xfree(a);
  return 1;
}


/* Checks if a given integer n is a perfect square, i.e. if sqrt(n)^2 = n. Any power of two is a perfect square, for example
   n = 1, 4, 9, 16, 25, 36, 49, etc.  See also: `sqrt`, `square`, `math.ispow2`.  2.16.13 */
int INLINE issquare (lua_Number n) {
  lua_Number s = luai_numint(sqrt(n));
  return s*s == n;
}

static int math_issquare (lua_State *L) {  /* 2.16.13 */
  lua_pushboolean(L, issquare(agn_checkinteger(L, 1)));
  return 1;
}


/* Checks if a given integer n is a perfect cube, i.e. if cbrt(n)^3 = n. See also: `cbrt`, `cube`. 2.16.13 */
static int math_iscube (lua_State *L) {  /* 2.16.13 */
  lua_Number n, c;
  n = agn_checkinteger(L, 1);
  c = luai_numint(cbrt(n));
  lua_pushboolean(L, c*c*c == n);
  return 1;
}


/* Checks whether a non-negative integer is a Fibonacci number. See: https://www.geeksforgeeks.org/check-number-fibonacci-number,
   by Abhay Rathi. See also: `math.fib`, `math.fibinv`.  2.16.13 */
static int math_isfib (lua_State *L) {
  lua_Number n = agn_checknonnegint(L, 1);
  lua_pushboolean(L, issquare(5*n*n + 4) || issquare(5*n*n - 4));  /* or both ! */
  return 1;
}


/* UNDOC Checks whether a number is halfway between two C floats (not doubles), see fmaf.c  file of the ulysses-libc-master package */
static int math_ishalfway (lua_State *L) {  /* 2.16.13 */
  int32_t lx;
  GET_LOW_WORD(lx, agn_checknumber(L, 1));
  lua_pushboolean(L, (lx & 0x1fffffff) == 0x10000000);
  return 1;
}


/* Returns ln(abs(x)) for numeric or complex x. With complex numbers, takes care of underflows. 2.16.13 */
static int math_lnabs (lua_State *L) {
  if (lua_isnumber(L, 1)) {
    lua_Number x = fabs(agn_tonumber(L, 1));
    if (x == 0)
      lua_pushundefined(L);
    else
      lua_pushnumber(L, sun_log(x));
  } else if (lua_iscomplex(L, 1)) {
    lua_Number a, b, max, q;
#ifndef PROPCMPLX
    agn_Complex x;
    x = agn_tocomplex(L, 1);
    a = fabs(creal(x)); b = fabs(cimag(x));
#else
    a = fabs(agn_complexreal(L, 1)), b = fabs(agn_compleximag(L, 1));
#endif
    if (a >= b) {
      max = a;
      q = luai_numdiv(b, a);
    } else {
      max = b;
      q = luai_numdiv(a, b);
    }
    if (isnan(q)) q = 0;
    lua_pushnumber(L, sun_log(max) + 0.5*log1p(q*q));  /* handles underflows if q is close to 0 */
  } else
    luaL_error(L, "Error in " LUA_QS ": either number or complex number expected.", "lnabs");
  return 1;
}


static const luaL_Reg mathlib[] = {
  {"accu", math_accu},                    /* added on May 26, 2018 */
  {"arccosh", math_arccosh},              /* added on December 04, 2014 */
  {"arcsinh", math_arcsinh},              /* added on April 21, 2019 */
  {"arctanh", math_arctanh},              /* added on April 21, 2019 */
  {"branch", math_branch},                /* added on September 25, 2016 */
  {"ceillog2", math_ceillog2},            /* added on November 21, 2014 */
  {"ceilpow2", math_ceilpow2},            /* added on November 27, 2014 */
  {"chop", math_chop},                    /* added on September 17, 2017 */
  {"cld", math_cld},                      /* added on October 19, 2016 */
  {"clip", math_clip},                    /* added on September 04, 2017 */
  {"copysign", math_copysign},            /* added on December 04, 2014 */
  {"decompose", math_decompose},          /* added on June 05, 2015 */
  {"dirac", math_dirac},                  /* added on September 04, 2017 */
  {"dms", math_dms},                      /* added on July 27, 2015 */
  {"eps", math_eps},                      /* added on February 09, 2015 */
  {"epsilon", math_epsilon},              /* added January 30, 2017 */
  {"expminusone", math_expminusone},      /* added on December 02, 2014 */
  {"exponent", math_exponent},            /* added on January 24, 2017 */
  {"fall", math_fall},                    /* added on June 11, 2017 */
  {"fdim", math_fdim},                    /* added on August 03, 2015 */
  {"fib", math_fib},                      /* added on July 26, 2018 */
  {"fibinv", math_fibinv},                /* added on July 31, 2018 */
  {"fld", math_fld},                      /* added on October 19, 2016 */
  {"flipsign", math_flipsign},            /* added on November 21, 2014 */
  {"floorpow2", math_floorpow2},          /* added on August 01, 2019 */
  {"fpclassify", math_fpclassify},        /* added on September 09, 2017 */
  {"frexp", math_frexp},                  /* added on February 08, 2017 */
  {"gcd", math_gcd},                      /* added on July 22, 2018 */
  {"infinite", math_infinite},            /* added on October 25, 2016 */
  {"iscube", math_iscube},                /* added on January 17, 2020 */
  {"isfib", math_isfib},                  /* added on January 17, 2020 */
  {"ishalfway", math_ishalfway},          /* added on January 22, 2020 */
  {"isinfinity", math_isinfinity},        /* added on October 19, 2015 */
  {"isirregular", math_isirregular},      /* added on March 05, 2017 */
  {"isminuszero", math_isminuszero},      /* added on August 12, 2015 */
  {"isordered", math_isordered},          /* added on December 02, 2014 */
  {"ispow2", math_ispow2},                /* added on January 19, 2017 */
  {"isqrt", math_isqrt},                  /* added on October 19, 2016 */
  {"isquare", math_issquare},             /* added on January 17, 2020 */
  {"isprime", math_isprime},              /* added on August 19, 2007 */
  {"issubnormal", math_issubnormal},      /* added on January 31, 2017 */
  {"koadd", math_koadd},                  /* added on February 01, 2015 */
  {"lcm", math_lcm},                      /* added on July 22, 2018 */
  {"lnabs", math_lnabs},                  /* added on January 22, 2020 */
  {"lnfact", math_lnfact},                /* added on October 14, 2016 */
  {"lnplusone", math_lnplusone},          /* added on December 02, 2014 */
  {"mantissa", math_mantissa},            /* added on January 24, 2017 */
  {"max", math_max},
  {"min", math_min},
  {"modinv", math_modinv},                /* added March 05, 2017 */
  {"modulus", math_modulus},              /* added on Jnauary 29, 2017 */
  {"morton", math_morton},                /* added on November 27, 2014 */
  {"ndigits", math_ndigits},              /* added on August 12, 2017 */
  {"nearmod", math_nearmod},              /* added on March 30, 2019 */
  {"nextafter", math_nextafter},          /* added on August 17, 2009 */
  {"nextprime", math_nextprime},          /* added on June 24, 2010 */
  {"norm", math_norm},                    /* added on May 22, 2010 */
  {"normalise", math_normalise},          /* added on March 14, 2017 */
  {"rempio2", math_rempio2},              /* added on October 22, 2017 */
  {"pochhammer", math_pochhammer},        /* added on August 14, 2018 */
  {"powmod", math_powmod},                /* added on October 19, 2016 */
  {"precision", math_precision},          /* added on March 05, 2017 */
  {"prevprime", math_prevprime},          /* added on June 24, 2010 */
  {"quadrant", math_quadrant},            /* added on August 02, 2015 */
  {"random", math_random},
  {"randomseed", math_randomseed},
  {"rectangular", math_rectangular},      /* added on February 08, 2015 / August 06, 2017 */
  {"rint", math_rint},                    /* added on September 28, 2015 */
  {"signbit", math_signbit},              /* added on February 08, 2015 */
  {"significand", math_significand},      /* added on May 04, 2019 */
  {"sincos", math_sincos},                /* added on October 17, 2017 */
  {"sinhcosh", math_sinhcosh},            /* added on November 05, 2017 */
  {"stirnum", math_stirnum},              /* added on August 13, 2018 */
  {"todecimal", math_todecimal},          /* added on August 25, 2013 */
  {"tohex",  math_tohex},                 /* added on November 17, 2017 */
  {"toradians", math_toradians},          /* added on January 21, 2008 */
  {"tosgesim",  math_tosgesim},           /* added on July 26, 2016 */
  {"triangular",  math_triangular},       /* added on September 09, 2017 */
  {"ulp", math_ulp},                      /* added on July 21, 2018 */
  {"uexponent", math_uexponent},          /* added on September 09, 2018 */
  {"unitise",  math_unitise},             /* added on August 30, 2017 */
  {"wrap", math_wrap},                    /* added on November 11, 2015 */
  {"zerosubnormal", math_zerosubnormal},  /* added on February 01, 2017 */
  {NULL, NULL}
};


/*
** Open math library
*/
LUALIB_API int luaopen_math (lua_State *L) {
  luaL_register(L, LUA_MATHLIBNAME, mathlib);
  lua_pushnumber(L, PHI);                  /* added 0.5.0 */
  lua_setfield(L, -2, "Phi");              /* added 0.5.0, Golden ratio */
  lua_pushnumber(L, AGN_LASTCONTINT);      /* added 2.10.0; 2^53 = 9007199254740992 */
  lua_setfield(L, -2, "lastcontint");      /* Largest integer i so that i-1 <> i but i = i+1; needed by math.convertbase */
  lua_pushnumber(L, sun_pow(2, -1022));    /* added 2.10.0 */
  lua_setfield(L, -2, "smallestnormal");   /* smallest normal number 2.2250738585072009e-308 = 2^-1022 on x86 CPUs*/
  /* for math.smallest see library.agn */
  lua_pushinteger(L, 0);                   /* added 2.11.0 RC2 */
  lua_setfield(L, -2, "fp_zero");          /* for FP_ZERO */
  lua_pushinteger(L, 1);                   /* added 2.11.0 RC2 */
  lua_setfield(L, -2, "fp_infinite");      /* for FP_INFINITE */
  lua_pushinteger(L, 2);                   /* added 2.11.0 RC2 */
  lua_setfield(L, -2, "fp_nan");           /* for FP_NAN */
  lua_pushinteger(L, 3);                   /* added 2.11.0 RC2 */
  lua_setfield(L, -2, "fp_normal");        /* for FP_NORMAL */
  lua_pushinteger(L, 4);                   /* added 2.11.0 RC2 */
  lua_setfield(L, -2, "fp_subnormal");     /* for FP_SUBNORMAL */
  return 1;
}

