/*
** $Id: lreglib.c,v 1.38 2005/10/23 17:38:15 roberto Exp $
** Library for Register Manipulation
** See Copyright Notice in agena.h
*/


#include <stdlib.h>
#include <math.h>
#include <string.h>

#define stack_c
#define LUA_LIB

#include "agena.h"

#include "agnxlib.h"
#include "agenalib.h"
#include "agnhlps.h"  /* for introsort */
#include "lstate.h"


#define aux_checkstackindex(L, idx, procname) { \
  cell = agn_checkinteger(L, 1); \
  if (cell < 0) \
    cell = stacktop(L) + cell; \
  if (cell < LUA_CELLSTACKBOTTOM || cell >= stacktop(L)) \
    luaL_error(L, "Error in " LUA_QS ": stack position is out-of-range or empty stack.", procname); \
}

#define aux_checkstack(L, nargs, procname) { \
  if (stacktop(L) + (nargs) - 1 >= stackmax(L)) { \
    size_t newsize = stackmax(L) << 1; \
    if (isnumstack(L)) { \
      L->cells[L->currentstack] = realloc(L->cells[L->currentstack], newsize * sizeof(lua_Number)); \
      if (L->cells[L->currentstack] == NULL) \
        luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", procname); \
    } else { \
      L->charcells[L->currentstack - LUA_NUMSTACKS] = \
        realloc(L->charcells[L->currentstack - LUA_NUMSTACKS], newsize * sizeof(unsigned char)); \
      if (L->charcells[L->currentstack - LUA_NUMSTACKS] == NULL) \
        luaL_error(L, "Error in " LUA_QS ": memory allocation failed.", procname); \
    } \
    stackmax(L) = newsize; \
  } \
}

static int stack_pushvalued (lua_State *L) {  /* 2.9.4, push value residing at the given relative stack position to the top of the stack */
  int cell;
  aux_checkstack(L, 1, "stack.pushvalued");
  aux_checkstackindex(L, 1, "stack.pushvalued");
  if (isnumstack(L))  /* 2.12.7 */
    cells(L, (stacktop(L)++)) = cells(L, cell);
  else
    charcell(L, (stacktop(L)++)) = charcell(L, cell);
  return 0;
}


static int stack_removed (lua_State *L) {  /* 2.9.4, removes the value residing at the given relative stack position,
  returns it, and moves all elements to close the space */
  int cell, i;
  aux_checkstackindex(L, 1, "stack.removed");
  if (isnumstack(L)) {  /* 2.12.7 */
    lua_Number v = cells(L, cell);
    for (i=cell + 1; i < stacktop(L); i++)
      cells(L, i - 1) = cells(L, i);
    stacktop(L)--;
    lua_pushnumber(L, v);
  } else {  /* 2.12.7 */
    char v = charcell(L, cell);
    for (i=cell + 1; i < stacktop(L); i++)
      charcell(L, i - 1) = charcell(L, i);
    stacktop(L)--;
    lua_pushchar(L, v);
  }
  return 1;
}


static int stack_insertd (lua_State *L) {  /* 2.9.4, inserts an element to the given stack position shifting up other elements to open space */
  int cell, i;
  aux_checkstack(L, 1, "stack.insertd");
  aux_checkstackindex(L, 1, "stack.insertd");
  if (isnumstack(L)) {
    for (i=(stacktop(L)++); i > cell; i--)
      cells(L, i) = cells(L, i - 1);
    cells(L, cell) = agn_checknumber(L, 2);
  } else {  /* 2.12.7 */
    for (i=(stacktop(L)++); i > cell; i--)
      charcell(L, i) = charcell(L, i - 1);
    if (agn_isstring(L, 2))
      charcell(L, cell) = agn_tostring(L, 2)[0];
    else if (agn_isnumber(L, 2)) {
      unsigned char n = lua_tointeger(L, 2) + '0';  /* integers 0 to 9 */
      charcell(L, cell) = n;
    } else
      luaL_error(L, "Error in " LUA_QS ": expected a number or string, got %s.", "stack.insertd", lua_typename(L, 2));
  }
  return 0;
}


static int stack_dumpd (lua_State *L) {  /* 2.9.4, returns all or a given number of values in the built-in stack
  in a new sequence and pops them from the stack. The number of pre-allocated slots is not changed, see shrinkd. */
  int i, n, csf;
  if (L->currentstack == 0)  /* 2.14.2 */
    luaL_error(L, "Error in " LUA_QS ": cannot operate on stack 0.", "stack.dumpd");
  csf = stacktop(L);
  if (csf == 0)
    lua_pushnil(L);
  else {
    n = luaL_optinteger(L, 1, csf - LUA_CELLSTACKBOTTOM);  /* 2.9.5 */
    if (n < 0 || n > csf - LUA_CELLSTACKBOTTOM)
      luaL_error(L, "Error in " LUA_QS ": invalid argument.", "stack.dumpd");
    if (isnumstack(L)) {
      int c = 0;
      agn_createseq(L, n);
      for (i=csf - n; i < csf; i++)
        agn_seqsetinumber(L, -1, ++c, cells(L, i));
    } else  /* 2.12.7 */
      lua_pushlstring(L, getcharcell(L, csf - n), n);
    stacktop(L) -= n;
  }
  return 1;
}


static int stack_sized (lua_State *L) {  /* 2.9.4, returns the current number of elements in the built-in current stack
  along with the current maximum number of slots allocated by the system. See also: allotted. */
  lua_pushinteger(L, stacktop(L));
  lua_pushinteger(L, stackmax(L));
  return 2;
}


static int stack_shrinkd (lua_State *L) {  /* 2.9.4; if possible, shrinks the number of pre-allocated but not assigned slots in the
  built-in stack. The function returns the new number of pre-allocated slots but does not pop any elements. The function is useful
  to reduce memory consumption if a lot of numbers have been removed from the current stack. */
  size_t p, csf, cn, i, nargs, oldcellnumber;
  nargs = lua_gettop(L);
  cn = L->currentstack;
  if (cn == 0)  /* 2.14.2 */
    luaL_error(L, "Error in " LUA_QS ": cannot operate on stack 0.", "stack.shrinkd");
  if (nargs == 0) {
    csf = stacktop(L);
    /* GNU C Lib: "If the new size is the same as the old size, realloc is guaranteed to change nothing ..." */
    if (csf == 0)  /* 2.15.4 fix */
      p = LUA_CELLSTACKSIZE;
    else
      p = 2 << ( ( (csf < LUA_CELLSTACKSIZE) ? ceillog2(LUA_CELLSTACKSIZE) : ceillog2(csf) ) - 1 );
    if (isnumstack(L))
      L->cells[cn] = realloc(L->cells[cn], p * sizeof(lua_Number));
    else
      L->charcells[cn - LUA_NUMSTACKS] = realloc(L->charcells[cn - LUA_NUMSTACKS], p * sizeof(unsigned char));
    stackmax(L) = p;
    lua_pushinteger(L, p);
    i = 1;
  } else {  /* 2.9.7: shrink all stacks */
    oldcellnumber = L->currentstack;
    for (i=0; i < LUA_NUMSTACKS + LUA_CHARSTACKS; i++) {
      L->currentstack = cn = i;
      csf = stacktop(L);
      if (csf == 0)  /* 2.15.4 fix */
        p = LUA_CELLSTACKSIZE;
      else
        p = 2 << ( ( (csf < LUA_CELLSTACKSIZE) ? ceillog2(LUA_CELLSTACKSIZE) : ceillog2(csf) ) - 1 );
      if (isnumstack(L))
        L->cells[cn] = realloc(L->cells[cn], p * sizeof(lua_Number));
      else
        L->charcells[cn - LUA_NUMSTACKS] = realloc(L->charcells[cn - LUA_NUMSTACKS], p * sizeof(unsigned char));
      stackmax(L) = p;
      lua_pushinteger(L, p);
    }
    i--;
    L->currentstack = oldcellnumber;
  }
  return i;
}


/* Reverses the positions of the elements in the built-in stack. The function returns nothing. */
static int stack_reversed (lua_State *L) {  /* 2.9.4 */
  size_t i, offset, csf;
  int n;  /* 2.9.7 fix */
  if (L->currentstack == 0)  /* 2.14.2 */
    luaL_error(L, "Error in " LUA_QS ": cannot operate on stack 0.", "stack.reversed");
  csf = stacktop(L);
  n = luaL_optinteger(L, 1, csf - LUA_CELLSTACKBOTTOM);  /* 2.9.6 */
  if (n < 0 || n > csf - LUA_CELLSTACKBOTTOM)
    luaL_error(L, "Error in " LUA_QS ": invalid argument.", "stack.reversed");
  offset = csf - n;
  n = csf;
  if (isnumstack(L)) {  /* 2.12.7 */
    lua_Number t;
    for (i=0 + offset; i < (n + offset)/2; i++) {
      t = cells(L, i);
      cells(L, i) = cells(L, n - i - 1 + offset);
      cells(L, n - i - 1 + offset) = t;
    }
  } else {  /* 2.12.7 */
    unsigned char t;
    for (i=0 + offset; i < (n + offset)/2; i++) {
      t = charcell(L, i);
      charcell(L, i) = charcell(L, n - i - 1 + offset);
      charcell(L, n - i - 1 + offset) = t;
    }
  }
  return 0;
}


/* Moves all the elements in the built-in stack n places from the bottom to the top if n is positive,
   and n places from the top to the bottom if n is negative. The function returns nothing. */
static int stack_rotated (lua_State *L) {  /* 2.9.4 */
  size_t s, i, j;
  int n;
  lua_Number t;
  unsigned char t1;
  if (L->currentstack == 0)  /* 2.14.2 */
    luaL_error(L, "Error in " LUA_QS ": cannot operate on stack 0.", "stack.rotated");
  s = stacktop(L);
  n = agnL_optinteger(L, 1, 1);
  if (s < 2) return 0;
  if (n < 0) {
    if (isnumstack(L)) {  /* 2.12.7 */
      for (j=0; j < -n; j++) {
        t = cells(L, 0);
        for (i=0; i < s - 1; i++) cells(L, i) = cells(L, i + 1);
        cells(L, s - 1) = t;
      }
    } else {  /* 2.12.7 */
      for (j=0; j < -n; j++) {
        t1 = charcell(L, 0);
        for (i=0; i < s - 1; i++) charcell(L, i) = charcell(L, i + 1);
        charcell(L, s - 1) = t1;
      }
    }
  } else if (n > 0) {  /* 2.12.7 */
    if (isnumstack(L)) {
      for (j=0; j < n; j++) {
        t = cells(L, s - 1);
        for (i=s - 1; i > 0; i--) cells(L, i) = cells(L, i - 1);
        cells(L, 0) = t;
      }
    } else {
      for (j=0; j < n; j++) {
        t1 = charcell(L, s - 1);
        for (i=s - 1; i > 0; i--) charcell(L, i) = charcell(L, i - 1);
        charcell(L, 0) = t1;
      }
    }
  }
  return 0;
}


/* Sorts the last given number of elements pushed onto the current stack in ascending order using the fast Introsort
   algorithm. If no argument is passed, all elements in the current stack are processed. The function returns nothing. */
static int stack_sorted (lua_State *L) {  /* 2.9.7 */
  size_t csf;
  int n;
  csf = stacktop(L);
  n = luaL_optinteger(L, 1, csf - LUA_CELLSTACKBOTTOM);
  if (!isnumstack(L))
    luaL_error(L, "Error in " LUA_QS ": cannot sort character stacks.", "stack.sorted");
  if (L->currentstack == 0)  /* 2.14.2 */
    luaL_error(L, "Error in " LUA_QS ": cannot operate on stack 0.", "stack.sortedd");
  if (n < 0 || n > csf - LUA_CELLSTACKBOTTOM)
    luaL_error(L, "Error in " LUA_QS ": invalid argument.", "stack.sorted");
  tools_dintrosort(L->cells[L->currentstack], csf - n, csf - 1, 0, 2*log2(csf));
  return 0;
}


static int stack_resetd (lua_State *L) {  /* 2.9.4, Clears the entire stack so that it becomes empty. The functions
  does not return anything and should be used if succeedingly working with stacks in a session. The number of
  pre-allocated slots is not changed, however, see shrinkd. */
  size_t nargs;
  nargs = lua_gettop(L);
  if (L->currentstack == 0)
    luaL_error(L, "Error in " LUA_QS ": cannot operator on stack 0.", "stack.resetd");
  if (nargs == 0) {
    stacktop(L) = LUA_CELLSTACKBOTTOM;
  } else {
    size_t i;
    int cnr, oldcnr;
    oldcnr = L->currentstack;
    for (i=1; i <= nargs; i++) {
      cnr = agn_checkinteger(L, i);
      if (cnr < 0 || cnr >= LUA_NUMSTACKS + LUA_CHARSTACKS)
        luaL_error(L, "Error in " LUA_QS ": invalid stack number %d.", "stack.resetd", cnr);
      L->currentstack = cnr;  /* 2.14.2 */
      stacktop(L) = LUA_CELLSTACKBOTTOM;
    }
    L->currentstack = oldcnr;
  }
  return 0;
}


static int stack_selected (lua_State *L) {  /* 2.9.7, returns the current stack number. */
  lua_pushinteger(L, L->currentstack);  /* changed 2.14.2 */
  return 1;
}


static int stack_choosed (lua_State *L) {  /* 2.9.7, returns the number of the stack with the least number of stored values. */
  int i, min, minchar, x, smallest, smallestchar, oldstack;
  min = minchar = smallest = smallestchar = MAX_INT;
  oldstack = L->currentstack;
  for (i=LUA_NUMSTACKS + LUA_CHARSTACKS - 1; i >= LUA_NUMSTACKS; i--) {
    L->currentstack = i;
    x = stacktop(L);
    if (x <= minchar) { minchar = x; smallestchar = i; }  /* always return the smallest stack number */
  }
  for (i=LUA_NUMSTACKS - 1; i > 0 ; i--) {  /* changed 2.14.2, do not check stack 0 */
    L->currentstack = i;
    x = stacktop(L);
    if (x <= min) { min = x; smallest = i; }  /* always return the smallest stack number */
  }
  L->currentstack = oldstack;
  lua_pushinteger(L, smallest);      /* changed 2.14.2 */
  lua_pushinteger(L, smallestchar);  /* changed 2.14.2 */
  return 2;
}


/* Determines the current stack number p, and then switches to the number stack q with the least number of elements if
   any number x or no argument is given, or to the character stack q with the least number of elements if any string x is
   given. The function returns p and q, in this order. You can use p later to switch back to the former stack with the
   `switchd` statement. */
static int stack_switchto (lua_State *L) {  /* 2.14.4, switch to smallest stack. */
  int x, i, min, smallest, oldstack;
  min = smallest = MAX_INT;
  oldstack = L->currentstack;
  if (lua_isnoneornil(L, 1) || lua_isnumber(L, 1)) {  /* switch to smallest number stack */
    for (i=LUA_NUMSTACKS - 1; i > 0 ; i--) {
      L->currentstack = i;
      x = stacktop(L);
      if (x <= min) { min = x; smallest = i; }  /* always return the smallest stack number */
    }
  } else if (lua_isstring(L, 1)) {  /* goto char stack */
    for (i=LUA_NUMSTACKS + LUA_CHARSTACKS - 1; i >= LUA_NUMSTACKS; i--) {
      L->currentstack = i;
      x = stacktop(L);
      if (x <= min) { min = x; smallest = i; }  /* always return the smallest stack number */
    }
  } else
    luaL_error(L, "Error in " LUA_QS ": invalid argument.", "stack.switchto");
  L->currentstack = smallest;
  lua_pushinteger(L, oldstack);
  lua_pushinteger(L, L->currentstack);
  return 2;
}


/* Pushes all characters in the string str onto a character stack. The function returns the size of `str`. */
static int stack_pushstringd (lua_State *L) {  /* 2.12.7 */
  size_t l;
  const char *str = agn_checklstring(L, 1, &l);
  if (isnumstack(L))
    luaL_error(L, "Error in " LUA_QS ": need a character stack, got stack #%d.", "stack.pushstringd", L->currentstack);
  if (l != 0) {
    aux_checkstack(L, l, "stack.pushstringd");
    while (*str) {
      charcell(L, (stacktop(L)++)) = *str++;
    }
  }
  lua_pushinteger(L, l);
  return 1;
}


static int stack_replaced (lua_State *L) {  /* 2.12.7 */
  int cell;
  aux_checkstackindex(L, 1, "stack.replaced");
  if (isnumstack(L)) {
    lua_Number x = agn_checknumber(L, 2);
    lua_pushnumber(L, cells(L, cell));
    cells(L, cell) = x;
  } else {  /* 2.12.7 */
    lua_pushlstring(L, getcharcell(L, cell), 1);
    if (agn_isstring(L, 2))
      charcell(L, cell) = agn_tostring(L, 2)[0];
    else if (agn_isnumber(L, 2)) {
      unsigned char n = lua_tointeger(L, 2) + '0';  /* integers 0 to 9 */
      charcell(L, cell) = n;
    } else
      luaL_error(L, "Error in " LUA_QS ": expected a number or string, got %s.", "stack.replaced", lua_typename(L, 2));
  }
  return 1;
}


static int stack_explored (lua_State *L) {  /* 2.12.7 */
  int i, n;
  n = stacktop(L);
  if (n == 0)
    lua_pushnil(L);
  else {
    if (isnumstack(L)) {
      agn_createseq(L, n);
      for (i=0; i < n; i++)
        agn_seqsetinumber(L, -1, i + 1, cells(L, i));
    } else
      lua_pushlstring(L, getcharcell(L, 0), n);
  }
  return 1;
}


static int stack_attribd (lua_State *L) {  /* 2.15.4 */
  lua_createtable(L, 0, 0);
  lua_rawsetstringnumber(L, -1, "currentstack", L->currentstack);
  lua_rawsetstringnumber(L, -1, "stacktop", stacktop(L));
  lua_rawsetstringnumber(L, -1, "stackmax", stackmax(L));
  lua_rawsetstringnumber(L, -1, "defaultsize", LUA_CELLSTACKSIZE);
  lua_rawsetstringboolean(L, -1, "isnumstack", isnumstack(L));
  return 1;
}


/* }====================================================== */

static const luaL_Reg stack_funcs[] = {
  {"attribd", stack_attribd},               /* added July 01, 2019 */
  {"choosed", stack_choosed},               /* added January 13, 2016 */
  {"dumpd", stack_dumpd},                   /* added November 29, 2015 */
  {"explored", stack_explored},             /* added August 11, 2018 */
  {"insertd", stack_insertd},               /* added November 29, 2015 */
  {"pushvalued", stack_pushvalued},         /* added November 29, 2015 */
  {"pushstringd", stack_pushstringd},       /* added August 11, 2018 */
  {"removed", stack_removed},               /* added November 29, 2015 */
  {"replaced", stack_replaced},             /* added August 11, 2018 */
  {"resetd", stack_resetd},                 /* added November 30, 2015 */
  {"reversed", stack_reversed},             /* added November 30, 2015 */
  {"rotated", stack_rotated},               /* added November 30, 2015 */
  {"selected", stack_selected},             /* added January 11, 2016 */
  {"shrinkd", stack_shrinkd},               /* added November 30, 2015 */
  {"sized", stack_sized},                   /* added November 30, 2015 */
  {"sorted", stack_sorted},                 /* added January 03, 2016 */
  {"switchto", stack_switchto},             /* added January 04, 2019 */
  {NULL, NULL}
};


LUALIB_API int luaopen_stack (lua_State *L) {
  luaL_register(L, AGENA_STACKLIBNAME, stack_funcs);
  return 1;
}

